package net.sf.saxon;
import net.sf.saxon.event.CommentStripper;
import net.sf.saxon.event.ContentEmitter;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.style.StyleNodeFactory;
import net.sf.saxon.style.StylesheetStripper;
import net.sf.saxon.tree.DocumentImpl;
import net.sf.saxon.tree.TreeBuilder;

import javax.xml.transform.Templates;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.sax.TemplatesHandler;


/**
  * <b>TemplatesHandlerImpl</b> implements the javax.xml.transform.sax.TemplatesHandler
  * interface. It acts as a ContentHandler which receives a stream of
  * SAX events representing a stylesheet, and returns a Templates object that
  * represents the compiled form of this stylesheet.
  * @author Michael H. Kay (michael.h.kay@ntlworld.com)
  */

public class TemplatesHandlerImpl extends ContentEmitter implements TemplatesHandler {

    private Configuration config;
    private TreeBuilder builder;
    private StyleNodeFactory nodeFactory;
    private Templates templates;
    private String systemId;

    /**
    * Create a TemplatesHandlerImpl and initialise variables. The constructor is protected, because
    * the Filter should be created using newTemplatesHandler() in the SAXTransformerFactory
    * class
    */

    protected TemplatesHandlerImpl(Configuration config) {

        this.config = config;

        // Allocate a namepool for elements and attributes used in the stylesheet

        NamePool localNamePool = new NamePool();
	    setNamePool(localNamePool);

        //setIsStyleSheet(true);

        nodeFactory = new StyleNodeFactory(localNamePool, config.isAllowExternalFunctions());

        StylesheetStripper styleStripper = new StylesheetStripper();
        styleStripper.setStylesheetRules(localNamePool);

        builder = new TreeBuilder();
        builder.setNamePool(localNamePool);
        builder.setNodeFactory(nodeFactory);
        builder.setLineNumbering(true);

        styleStripper.setUnderlyingReceiver(builder);

        CommentStripper commentStripper = new CommentStripper();
        commentStripper.setUnderlyingReceiver(styleStripper);

        this.setReceiver(commentStripper);

    }

    /**
    * Get the Templates object to used for a transformation
    */

    public Templates getTemplates() {
        if (templates==null) {
            DocumentImpl doc = (DocumentImpl)builder.getCurrentDocument();
            if (doc==null) {
                return null;
            }
            PreparedStyleSheet sheet = new PreparedStyleSheet(config);
            try {
                sheet.setStyleSheetDocument(doc, nodeFactory);
                templates = sheet;
            } catch (TransformerConfigurationException tce) {
                // don't know why we aren't allowed to just throw it!
                throw new UnsupportedOperationException(tce.getMessage());
            }
        }

        return templates;
    }

    /**
    * Set the SystemId of the document
    */

    public void setSystemId(String url) {
        systemId = url;
        builder.setSystemId(url);
    }

    /**
    * Get the systemId of the document
    */

    public String getSystemId() {
        return systemId;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): None
//
