package net.sf.saxon.dom;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.type.Type;
import net.sf.saxon.type.SchemaType;

import org.w3c.dom.*;

/**
  * The root node of an XPath tree. (Or equivalently, the tree itself).<P>
  * This class should have been named Root; it is used not only for the root of a document,
  * but also for the root of a result tree fragment, which is not constrained to contain a
  * single top-level element.
  * @author <A HREF="mailto:michael.h.kay@ntlworld.com>Michael H. Kay</A>
  */

public class DocumentWrapper extends NodeWrapper implements DocumentInfo {

    protected NamePool namePool;
    protected String baseURI;
    protected int documentNumber;

    public DocumentWrapper(Document doc, String baseURI) {
        super(doc, null, 0);
        node = doc;
        nodeKind = Type.DOCUMENT;
        this.baseURI = baseURI;
        docWrapper = this;
        namePool = NamePool.getDefaultNamePool();
    }

    /**
     * Create a wrapper for a node in this document
     */

    public NodeWrapper wrap(Node node) {
        if (node == docWrapper) {
            return this;
        }
        return makeWrapper(node, this);
    }

	/**
	* Set the name pool used for all names in this document
	*/

	public void setNamePool(NamePool pool) {
	    namePool = pool;
		documentNumber = pool.allocateDocumentNumber(this);
	}

	/**
	* Get the name pool used for the names in this document
	*/

	public NamePool getNamePool() {
	    return namePool;
	}

	/**
	* Get the unique document number
	*/

	public int getDocumentNumber() {
	    return documentNumber;
	}

    /**
    * Get the element with a given ID, if any
    * @param id the required ID value
    * @return a NodeInfo representing the element with the given ID, or null if there
     * is no such element. This implementation does not necessarily conform to the
     * rule that if an invalid document contains two elements with the same ID, the one
     * that comes last should be returned.
    */

    public NodeInfo selectID(String id) {
        Node el = ((Document)node).getElementById(id);
        if (el==null) {
            return null;
        }
        return wrap(el);
    }

    /**
    * Determine whether this is the same node as another node. <br />
    * Note: a.isSameNode(b) if and only if generateId(a)==generateId(b)
    * @return true if this Node object and the supplied Node object represent the
    * same node in the tree.
    */

    public boolean isSameNode(NodeInfo other) {
        if (!(other instanceof DocumentWrapper)) {
            return false;
        }
        return node == ((DocumentWrapper)other).node;
    }

    /**
    * Get the unparsed entity with a given name
    * @param name the name of the entity
    * @return null: JDOM does not provide access to unparsed entities
    */

    public String[] getUnparsedEntity(String name) {
        return null;
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
