package net.sf.saxon.event;
import net.sf.saxon.om.NamePool;
import org.xml.sax.Locator;
import javax.xml.transform.Result;
import javax.xml.transform.TransformerException;

/**
  * Receiver: This interface represents a recipient of XML tree-walking events. It is
  * based on SAX2's ContentHandler, but adapted to handle additional events, and
  * to use Saxon's name pool. Namespaces and Attributes are handled by separate events
  * following the startElement event. Schema types can be defined for elements and attributes.
  *
  * This interface is designed as an eventual replacement for the old Emitter class, as well as
  * the Outputter interface. The name change from "Emitter" to "Receiver" reflects that the essence
  * of the interface is that all implementations receive notification of these events; they do very
  * different things with the events once recieved.
  */

public interface Receiver extends Result {

	/**
	* Set the namePool in which all name codes can be found
	*/

	public void setNamePool(NamePool namePool);

	/**
	* Get the namePool in which all name codes can be found
	*/

	public NamePool getNamePool();

	/**
	* Set the System ID of the destination tree
	*/

	public void setSystemId(String systemId);

    /**
    * Set locator, to identify position in the document. A new locator may be supplied at any time.
    * It is not necessary to supply any locator. If a locator is supplied, it may be used
    * by the Receiver to obtain information about the location of the current event in a source
    * XML document.
    */

    public void setDocumentLocator(Locator locator) throws TransformerException;

    /**
    * Notify the start of the document
    */

    public void startDocument() throws TransformerException;

    /**
    * Notify an unparsed entity URI.
    * @param name The name of the unparsed entity
    * @param systemID The system identifier of the unparsed entity
    * @param publicID The public identifier of the unparsed entity
    */

    public void setUnparsedEntity(String name, String systemID, String publicID) throws TransformerException;

    /**
    * Notify the start of an element
    * @param nameCode integer code identifying the name of the element within the name pool.
    * @param typeCode integer code identifying the element's type within the name pool.
    * @param properties: for future use. Should be set to zero.
    */

    public void startElement(int nameCode, int typeCode, int properties) throws TransformerException;

    /**
    * Notify a namespace. Namespaces are notified <b>after</b> the startElement event, and before
    * any children for the element. The namespaces that are reported are only required
    * to include those that are different from the parent element; however, duplicates may be reported.
    * A namespace must not conflict with any namespaces already used for element or attribute names.
    * @param namespaceCode an integer: the top half is a prefix code, the bottom half a URI code.
    * These may be translated into an actual prefix and URI using the name pool. A prefix code of
    * zero represents the empty prefix (that is, the default namespace). A URI code of zero represents
    * a URI of "", that is, a namespace undeclaration.
    * @throws IllegalStateException: attempt to output a namespace when there is no open element
    * start tag
    */

    public void namespace(int namespaceCode, int properties) throws TransformerException;

    /**
    * Notify an attribute. Attributes are notified after the startElement event, and before any
    * children. Namespaces and attributes may be intermingled.
    * @param nameCode The name of the attribute, as held in the name pool
    * @param typeCode The type of the attribute, as held in the name pool
    * @param properties Bit significant value. The following bits are defined:
    *        <dt>DISABLE_ESCAPING</dt>    <dd>Disable escaping for this attribute</dd>
    *        <dt>NO_SPECIAL_CHARACTERS</dt>      <dd>Attribute value contains no special characters</dd>
    * @throws IllegalStateException: attempt to output an attribute when there is no open element
    * start tag
    */

    public void attribute(int nameCode, int typeCode, CharSequence value, int properties) throws TransformerException;

    /**
    * Notify the start of the content, that is, the completion of all attributes and namespaces.
    * Note that the initial receiver of output from XSLT instructions will not receive this event,
    * it has to detect it itself. Note that this event is reported for every element even if it has
    * no attributes, no namespaces, and no content.
    */

    public void startContent() throws TransformerException;

    /**
    * Notify the end of an element. The receiver must maintain a stack if it needs to know which
    * element is ending.
    */

    public void endElement() throws TransformerException;

    /**
    * Notify character data. Note that some receivers may require the character data to be
    * sent in a single event, but in general this is not a requirement.
    * @param chars The characters
    * @param properties Bit significant value. The following bits are defined:
    *        <dt>DISABLE_ESCAPING</dt>           <dd>Disable escaping for this text node</dd>
    *        <dt>USE_CDATA</dt>                  <dd>Output as a CDATA section</dd>
    *        <dt>NO_SPECIAL_CHARACTERS</dt>      <dd>Value contains no special characters</dd>
    *        <dt>WHITESPACE</dt>                 <dd>Text is all whitespace</dd>
    */

    public void characters(CharSequence chars, int properties) throws TransformerException;

    /**
    * Output a processing instruction
    * @param name The PI name. This must be a legal name (it will not be checked).
    * @param data The data portion of the processing instruction
    * @param properties Additional information about the PI. The following bits are
    * defined:
    *        <dt>CHECKED</dt>    <dd>Data is known to be legal (e.g. doesn't contain "?>")</dd>
    * @throws IllegalArgumentException: the content is invalid for an XML processing instruction
    */

    public void processingInstruction(String name, CharSequence data, int properties) throws TransformerException;

    /**
    * Notify a comment. Comments are only notified if they are outside the DTD.
    * @param content The content of the comment
    * @param properties Additional information about the comment. The following bits are
    * defined:
    *        <dt>CHECKED</dt>    <dd>Comment is known to be legal (e.g. doesn't contain "--")</dd>
    * @throws IllegalArgumentException: the content is invalid for an XML comment
    */

    public void comment (CharSequence content, int properties) throws TransformerException;

    /**
    * Notify the end of the document
    */

    public void endDocument() throws TransformerException;



}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
