package net.sf.saxon.event;
import net.sf.saxon.om.Axis;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.Orphan;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.tinytree.TinyBuilder;
import net.sf.saxon.tinytree.TinyDocumentImpl;
import net.sf.saxon.type.Type;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.EmptySequence;
import net.sf.saxon.value.SequenceExtent;
import net.sf.saxon.value.Value;

import javax.xml.transform.TransformerException;
import java.util.ArrayList;
import java.util.List;


/**
 * This outputter is used when writing a sequence of atomic values and nodes, that
 * is, when xsl:variable is used with content and an "as" attribute. The outputter
 * builds the sequence and provides access to it. (It isn't really an outputter at all,
 * it doesn't pass the events to anyone, it merely constructs the sequence in memory
 * and provides access to it). Note that the event sequence can include calls such as
 * startElement and endElement that require trees to be built. If nodes such as attributes
 * and text nodes are received while an element is being constructed, the nodes are added
 * to the tree. Otherwise, "orphan" nodes (nodes with no parent) are created and added
 * directly to the sequence.
 *
 * <p>This class is not used to build temporary trees. For that, the ComplexContentOutputter
 * is used.</p>
 *
 * <p>The SequenceOutputter no longer does type checking. This function is now performed,
 * where necessary, by a SequenceChecker that filters the event stream before it reaches
 * the SequenceOutputter.</p>
 *
 * @author <A HREF="mailto:michael.h.kay@ntlworld.com>Michael H. Kay</A>
 */

public final class SequenceOutputter extends Outputter {

    private List list;
    private NamePool namePool;
    private String systemId;
    private Receiver tree = null;
    private TinyBuilder builder = null;
    private int level = 0;
    private boolean inStartTag = false;

    /**
    * Create a new SequenceOutputter
    */

	public SequenceOutputter() {
	    //System.err.println("new SequenceOutputter");
	    this.list = new ArrayList();
	}

    /**
    * Set the name pool. This method must be called before startDocument() is called.
    * @param pool the name pool. This name pool must contain all the name codes that are
    * passed across this interface.
    */

    public void setNamePool(NamePool pool) {
        namePool = pool;
    }

    /**
    * Get the name pool
    * @return the Name Pool that was supplied using the setNamePool() method
    */

    public NamePool getNamePool() {
        return namePool;
    }

    /**
    * Set the system ID
    * @param systemId the URI used to identify the tree being passed across this interface
    */

    public void setSystemId(String systemId) {
        this.systemId = systemId;
    }

    /**
    * Get the system ID
    * @return the system ID that was supplied using the setSystemId() method
    */

    public String getSystemId() {
        return systemId;
    }

    /**
    * Notify the start of the document. This event is notified once, before any other events.
    * This implementation does nothing.
    */

    public final void startDocument() throws TransformerException {
    }

    /**
    * Output an element start tag.
    * @param nameCode The element name code - a code held in the Name Pool
    * @param typeCode Integer code identifying the type of this element. Zero identifies the default
    * type, that is xs:anyType
    * @param properties bit-significant flags indicating any special information
    */

    public void startElement(int nameCode, int typeCode, int properties) throws TransformerException {

        if (inStartTag) {
            startContent();
        }

        if (tree==null) {
            builder = new TinyBuilder();
            builder.setNamePool(namePool);
            //builder.setSizeParameters(treeSizeParameters);

            NamespaceReducer reducer = new NamespaceReducer();
            reducer.setUnderlyingReceiver(builder);
            reducer.setNamePool(namePool);

            ComplexContentOutputter complex = new ComplexContentOutputter();
            complex.setNamePool(namePool);
            complex.setReceiver(reducer);
            tree = complex;

            tree.setSystemId(systemId);
            tree.setNamePool(namePool);
            tree.startDocument();
            tree.startElement(nameCode, typeCode, properties);
        } else {
            tree.startElement(nameCode, typeCode, properties);
        }
        level++;
        //System.err.println("level = " + level);
        inStartTag = true;
        previousAtomic = false;
    }

    /**
    * Output an element end tag.
    */

    public void endElement() throws TransformerException {
        if (inStartTag) startContent();
        level--;
        //System.err.println("level = " + level);
        if (level==0) {
            tree.endElement();
            tree.endDocument();
            DocumentInfo doc = builder.getCurrentDocument();
            tree = null;
            builder = null;
            SequenceIterator iter = doc.iterateAxis(Axis.CHILD);
            //if (iter.hasNext()) {
                // always true
            NodeInfo element = (NodeInfo)iter.next();
            // mark the element as the effective root of the tree
            ((TinyDocumentImpl)doc).setRootNode(element);
            // add the constructed element to the result sequence
            append(element);
            //}
        } else {
            tree.endElement();
        }
        previousAtomic = false;
    }

    /**
    * Output a namespace declaration. <br>
    * This is added to a list of pending namespaces for the current start tag.
    * If there is already another declaration of the same prefix, this one is
    * ignored.
    * Note that unlike SAX2 startPrefixMapping(), this call is made AFTER writing the start tag.
    * @param nscode The namespace code
    * @param properties Allows special properties to be passed if required
    * @throws TransformerException if there is no start tag to write to (created using writeStartTag),
    * or if character content has been written since the start tag was written.
    */

    public void namespace(int nscode, int properties)
    throws TransformerException {
        if (tree == null) {
            Orphan o = new Orphan(namePool);
            o.setNodeKind(Type.NAMESPACE);
            o.setNameCode(namePool.allocate("", "", namePool.getPrefixFromNamespaceCode(nscode)));
            o.setStringValue(namePool.getURIFromNamespaceCode(nscode));
            append(o);
        } else {
            tree.namespace(nscode, properties);
        }
        previousAtomic = false;
    }

    /**
    * Output an attribute value. <br>
    * @param nameCode An integer code representing the name of the attribute, as held in the Name Pool
    * @param typeCode Integer code identifying the type annotation of the attribute; zero represents
    * the default type (xs:untypedAtomic)
    * @param value The value of the attribute
    * @param properties Bit significant flags for passing extra information to the serializer, e.g.
    * to disable escaping
    * @throws TransformerException if there is no start tag to write to (created using writeStartTag),
    * or if character content has been written since the start tag was written.
    */

    public void attribute(int nameCode, int typeCode, CharSequence value, int properties)
    throws TransformerException {
        //System.err.println("Attribute " + value + " tree=" + tree + " namePool =" + namePool );
        if (tree == null) {
            Orphan o = new Orphan(namePool);
            o.setNodeKind(Type.ATTRIBUTE);
            o.setNameCode(nameCode);
            o.setStringValue(value);
            o.setTypeAnnotation(typeCode);
            append(o);
        } else {
            tree.attribute(nameCode, typeCode, value, properties);
        }
        previousAtomic = false;
    }

    /**
    * The startContent() event is notified after all namespaces and attributes of an element
    * have been notified, and before any child nodes are notified.
    * @throws TransformerException for any failure
    */

    public void startContent() throws TransformerException {
        inStartTag = false;
        tree.startContent();
        previousAtomic = false;
    }

    /**
    * Produce text content output. <BR>
    * @param s The String to be output
    * @param properties bit-significant flags for extra information, e.g. disable-output-escaping
    * @throws TransformerException for any failure
    */

    public void characters(CharSequence s, int properties) throws TransformerException {
        //System.err.println("Characters " + s + " tree=" + tree);
        if (s==null) return;
        if (s.length() == 0) return;
        if (inStartTag) {
            startContent();
        }
        if (tree == null) {
            Orphan o = new Orphan(namePool);
            o.setNodeKind(Type.TEXT);
            o.setStringValue(s);
            append(o);
        } else {
            tree.characters(s, properties);
        }
        previousAtomic = false;
    }

    /**
    * Write a comment.
    */

    public void comment(CharSequence comment, int properties) throws TransformerException {
        if (inStartTag) startContent();
        if (tree == null) {
            Orphan o = new Orphan(namePool);
            o.setNodeKind(Type.COMMENT);
            o.setStringValue(comment);
            append(o);
        } else {
            tree.comment(comment, properties);
        }
        previousAtomic = false;
    }

    /**
    * Write a processing instruction
    * No-op in this implementation
    */

    public void processingInstruction(String target, CharSequence data, int properties) throws TransformerException {
        if (inStartTag) startContent();
        if (tree == null) {
            Orphan o = new Orphan(namePool);
            o.setNameCode(namePool.allocate("", "", target));
            o.setNodeKind(Type.PROCESSING_INSTRUCTION);
            o.setStringValue(data);
            append(o);
        } else {
            tree.processingInstruction(target, data, properties);
        }
        previousAtomic = false;
    }

    /**
    * Close the output
    */

    public void endDocument() throws TransformerException {
        previousAtomic = false;
    }

    /**
    * Append an item to the sequence, performing any necessary type-checking and conversion
    */

    public void append(Item item) throws TransformerException {

        // If an atomic value is written to a tree, and the previous item was also
        // an atomic value, then add a single space to separate them

        if (tree!=null && previousAtomic && item instanceof AtomicValue) {
            tree.characters(" ", 0);
        }
        previousAtomic = (item instanceof AtomicValue);

        if (tree==null) {
            list.add(item);
        } else {
            if (item instanceof AtomicValue) {
                tree.characters(item.getStringValue(), 0);
            } else {
                ((NodeInfo)item).copy(tree, NodeInfo.ALL_NAMESPACES, true);
            }
        }
    }

    /**
    * Get the sequence that has been built
    */

    public Value getSequence() {
        switch (list.size()) {
            case 0:
                return EmptySequence.getInstance();
            case 1:
                Item item = (Item)list.get(0);
                return Value.asValue(item);
            default:
                return new SequenceExtent(list);
        }
    }

    /**
     * Get the first item in the sequence that has been built
     */

    public Item getFirstItem() {
        if (list.size() == 0) {
            return null;
        } else {
            return (Item)list.get(0);
        }
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
