package net.sf.saxon.event;
import net.sf.saxon.charcode.UnicodeCharacterSet;
import net.sf.saxon.om.XMLChar;
import org.xml.sax.Locator;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.TransformerException;
import java.util.Stack;

/**
  * XMLEmitter is an Emitter that generates XML output
  * to a specified destination.
  */

public class XMLEmitter extends Emitter
{
    protected boolean empty = true;
    protected boolean escaping = true;
    protected boolean openStartTag = false;
    protected boolean declarationIsWritten = false;
    protected int elementCode;

    protected boolean preferHex = false;
    protected boolean undeclareNamespaces = false;

    // The element stack holds the display names (lexical QNames) of elements that
    // have been started but not finished. It is used to obtain the element name
    // for the end tag.

    protected Stack elementStack = new Stack();

    // Namecodes in the range 0..1024 are common. So for these codes,
    // we maintain a direct lookup from the namecode to the display name
    // that bypasses reference to the namepool

    private String[] nameLookup = new String[1024];


    static boolean[] specialInText;         // lookup table for special characters in text
    static boolean[] specialInAtt;          // lookup table for special characters in attributes
        // create look-up table for ASCII characters that need special treatment

    static {
        specialInText = new boolean[128];
        for (int i=0; i<=127; i++) specialInText[i] = false;
        specialInText[(char)0] = true;
            // used to switch escaping on and off for mapped characters
        specialInText['\r'] = true;
        specialInText['<'] = true;
        specialInText['>'] = true;
        specialInText['&'] = true;

        specialInAtt = new boolean[128];
        for (int i=0; i<=127; i++) specialInAtt[i] = false;
        specialInAtt[(char)0] = true;
            // used to switch escaping on and off for mapped characters
        specialInAtt['\r'] = true;
        specialInAtt['\n'] = true;
        specialInAtt['\t'] = true;
        specialInAtt['<'] = true;
        specialInAtt['>'] = true;
        specialInAtt['&'] = true;
        specialInAtt['\"'] = true;
    }

    /**
    * Set Document Locator. Provided merely to satisfy the interface.
    */

    public void setDocumentLocator(Locator locator) {}


    /**
    * Start of the document. Nothing is done at this stage: the opening of the output
     * file is deferred until some content is written to it.
    */

    public void startDocument () throws TransformerException {}

    /**
     * Do the real work of starting the document. This happens when the first
     * content is written.
     * @throws TransformerException
     */

    protected void openDocument () throws TransformerException
    {
        if (writer==null) {
            makeWriter();
        }
        if (characterSet==null) {
            characterSet = UnicodeCharacterSet.getInstance();
        }
        writeDeclaration();
        String rep = outputProperties.getProperty(SaxonOutputKeys.CHARACTER_REPRESENTATION);
        if (rep!=null) {
        	preferHex = (rep.trim().equalsIgnoreCase("hex"));
        }
        rep = outputProperties.getProperty(SaxonOutputKeys.UNDECLARE_NAMESPACES);
        if (rep!=null) {
        	undeclareNamespaces = (rep.trim().equalsIgnoreCase("yes"));
        }
    }

    /**
    * Output the XML declaration
    */

    public void writeDeclaration() throws TransformerException {
        if (declarationIsWritten) return;
        declarationIsWritten = true;
        try {
            String byteOrderMark = outputProperties.getProperty(SaxonOutputKeys.BYTE_ORDER_MARK);

            if ("yes".equals(byteOrderMark)) {
                writer.write('\uFEFF');
            }

            String omit = outputProperties.getProperty(OutputKeys.OMIT_XML_DECLARATION);
            if (omit==null) {
                omit = "no";
            }

            String version = outputProperties.getProperty(OutputKeys.VERSION);
            if (version==null) version = "1.0";

            String encoding = outputProperties.getProperty(OutputKeys.ENCODING);

            if (encoding==null || encoding.equalsIgnoreCase("utf8")) {
                encoding = "UTF-8";
            }

            // If user requests an encoding other than UTF-8, UTF-16, or ASCII,
            // then force an XML declaration to be output, otherwise the
            // result will not be well-formed XML

            if (omit.equals("yes") &&
                  !(encoding.equalsIgnoreCase("UTF-8")) ||
                    encoding.equalsIgnoreCase("utf-16") ||
                    encoding.equalsIgnoreCase("us-ascii") ||
                    encoding.equalsIgnoreCase("ascii")) {
                omit = "no";
            }

            String standalone = outputProperties.getProperty(OutputKeys.STANDALONE);

            if (omit.equals("no")) {
                writer.write("<?xml version=\"" + version + "\" " +
                              "encoding=\"" + encoding + "\"" +
                              (standalone!=null ? (" standalone=\"" + standalone + "\"") : "") +
                              "?>");
                    // no longer write a newline character: it's wrong if the output is an
                    // external general parsed entity
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Output the document type declaration
    */

    protected void writeDocType(String type, String systemId, String publicId) throws TransformerException {
        try {
            if (declarationIsWritten) {
                writer.write("\n");
            }
            writer.write("<!DOCTYPE " + type + "\n");
            if (systemId!=null && publicId==null) {
                writer.write("  SYSTEM \"" + systemId + "\">\n");
            } else if (systemId==null && publicId!=null) {     // handles the HTML case
                writer.write("  PUBLIC \"" + publicId + "\">\n");
            } else {
                writer.write("  PUBLIC \"" + publicId + "\" \"" + systemId + "\">\n");
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * End of the document. Close the output stream.
    */

    public void endDocument () throws TransformerException
    {
        try {
            if (writer != null) {
                writer.flush();
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Start of an element. Output the start tag, escaping special characters.
    */

    public void startElement (int nameCode, int typeCode, int properties) throws TransformerException
    {
        if (empty) {
            openDocument();
        }
        String displayName = null;

        // See if we've seen this name before
        if (nameCode < 1024) {
            displayName = nameLookup[nameCode];
        }

        // Otherwise, look it up in the namepool and check that it's encodable
        if (displayName == null) {
    	    displayName = namePool.getDisplayName(nameCode);
            if (nameCode < 1024) {
                nameLookup[nameCode] = displayName;
            }
            int badchar = testCharacters(displayName);
            if (badchar!=0) {
                throw new TransformerException("Element name contains a character (decimal + " +
                                                badchar + ") not available in the selected encoding");
            }
        }

        elementStack.push(displayName);
        elementCode = nameCode;

        try {
            if (empty) {
                String systemId = outputProperties.getProperty(OutputKeys.DOCTYPE_SYSTEM);
                String publicId = outputProperties.getProperty(OutputKeys.DOCTYPE_PUBLIC);
                if (systemId!=null) {
                    writeDocType(displayName, systemId, publicId);
                }
                empty = false;
            }
            if (openStartTag) {
                closeStartTag(displayName, false);
            }
            writer.write('<');
            writer.write(displayName);
            openStartTag = true;

        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    public void namespace(int namespaceCode, int properties) throws TransformerException {
        try {
            String nsprefix = namePool.getPrefixFromNamespaceCode(namespaceCode);
            String nsuri = namePool.getURIFromNamespaceCode(namespaceCode);

            if (nsprefix.equals("")) {
                writer.write(' ');
                writeAttribute(elementCode, "xmlns", nsuri, 0);
            } else {
                int badchar = testCharacters(nsprefix);
                if (badchar!=0) {
                    throw new TransformerException("Namespace prefix contains a character (decimal + " +
                                                    badchar + ") not available in the selected encoding");
                }
                if (undeclareNamespaces || !nsuri.equals("")) {
                    writer.write(' ');
                    writeAttribute(elementCode, "xmlns:" + nsprefix, nsuri, 0);
                }
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    public void attribute(int nameCode, int typeCode, CharSequence value, int properties)
    throws TransformerException {
        String displayName = null;

        // See if we've seen this name before
        if (nameCode < 1024) {
            displayName = nameLookup[nameCode];
        }

        // Otherwise, look it up in the namepool and check that it's encodable
        if (displayName == null) {
    	    displayName = namePool.getDisplayName(nameCode);
            if (nameCode < 1024) {
                nameLookup[nameCode] = displayName;
            }
            int badchar = testCharacters(displayName);
            if (badchar!=0) {
                throw new TransformerException("Attribute name contains a character (decimal + " +
                                                badchar + ") not available in the selected encoding");
            }
        }
        try {
            writer.write(' ');
            writeAttribute(
                elementCode,
                displayName,
                value,
                properties );

        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    public void startContent() throws TransformerException {
        // don't add ">" to the start tag until we know whether the element has content
    }

    public void closeStartTag(String displayName, boolean emptyTag) throws TransformerException {
        try {
            if (openStartTag) {
                if (emptyTag) {
                    writer.write(emptyElementTagCloser(displayName));
                } else {
                    writer.write('>');
                }
                openStartTag = false;
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Close an empty element tag. (This is overridden in XHTMLEmitter).
    */

    protected String emptyElementTagCloser(String displayName) {
        return "/>";
    }

    /**
    * Write attribute name=value pair.
     * @param elCode The element name is not used in this version of the
    * method, but is used in the HTML subclass.
     * @param attname The attribute name, which has already been validated to ensure
     * it can be written in this encoding
     * @param value The value of the attribute
     * @param properties Any special properties of the attribute
    */

    protected void writeAttribute(int elCode, String attname, CharSequence value, int properties) throws TransformerException {
        try {
            String val = value.toString();
            writer.write(attname);
            if ((properties & ReceiverOptions.NO_SPECIAL_CHARS) != 0) {
                writer.write('=');
                writer.write('"');
                writer.write(val);
                writer.write('"');
            } else if ((properties & ReceiverOptions.USE_NULL_MARKERS) != 0) {
                // null (0) characters will be used before and after any section of
                // the value where escaping is to be disabled
                writer.write('=');
                char delimiter = (val.indexOf('"') >= 0 ? '\'' : '"');
                writer.write(delimiter);
                writeEscape(value, true);
                writer.write(delimiter);
            } else {
                writer.write("=\"");
                writeEscape(value, true);
                writer.write('\"');
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }


    /**
    * Test that all characters in a name are supported in the target encoding.
     * @return zero if all the characters are available, or the value of the
     * first offending character if not
    */

    protected int testCharacters(CharSequence chars) throws TransformerException {
        for (int i=0; i<chars.length(); i++) {
            char c = chars.charAt(i);
            if (c > 127) {
                if (XMLChar.isHighSurrogate(c)) {
                    int cc = XMLChar.supplemental(c, chars.charAt(++i));
                    if (!characterSet.inCharset(cc)) {
                        return cc;
                    }
                } else if (!characterSet.inCharset(c)) {
                    return c;
                }
            }
        }
        return 0;
    }

    /**
    * End of an element.
    */

    public void endElement () throws TransformerException
    {
        String displayName = (String)elementStack.pop();
        try {
            if (openStartTag) {
                closeStartTag(displayName, true);
            } else {
                writer.write("</");
                writer.write(displayName);
                writer.write('>');
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Character data.
    */

    public void characters (CharSequence chars, int properties) throws TransformerException
    {
        if (empty) {
            openDocument();
        }
        try {
            if (openStartTag) {
                closeStartTag(null, false);
            }
            // System.err.println("Output characters [" + new String(ch, start, length) + "]");
            if ((properties & ReceiverOptions.DISABLE_ESCAPING) != 0) {
                if (testCharacters(chars) == 0) {
                    writer.write(chars.toString());
                } else {
                    // Recoverable error: using disable output escaping with characters
                    // that are not available in the target encoding
                    // TODO: issue a RecoverableError warning
                    writeEscape(chars, false);
                }
            } else {
                writeEscape(chars, false);
            }
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }


    /**
    * Handle a processing instruction.
    */

    public void processingInstruction (String target, CharSequence data, int properties)
        throws TransformerException  {
        if (empty) {
            openDocument();
        }
        try {
            if (openStartTag) {
                closeStartTag(null, false);
            }
            writer.write("<?" + target + (data.length()>0 ? ' ' + data.toString() : "") + "?>");
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Write contents of array to current writer, after escaping special characters.
    * This method converts the XML special characters (such as < and &) into their
    * predefined entities.
    * @param chars The character sequence containing the string
    * @param inAttribute  Set to true if the text is in an attribute value
    */

    protected void writeEscape(CharSequence chars, boolean inAttribute)
    throws java.io.IOException {
        int segstart = 0;
        boolean disabled = false;
        boolean[] specialChars = (inAttribute ? specialInAtt : specialInText);

        while (segstart < chars.length()) {
            int i = segstart;

            // find a maximal sequence of "ordinary" characters
            while (i < chars.length()) {
                char c = chars.charAt(i);
                if (c < 128) {
                    if (specialChars[c]) {
                        break;
                    } else {
                        i++;
                    }
                } else if (XMLChar.isHighSurrogate(c)) {
                    break;
                } else if (!characterSet.inCharset(c)) {
                    break;
                } else {
                    i++;
                }
            }

            // write out this sequence
            writer.write(chars.subSequence(segstart, i).toString());

            // exit if this was the whole string
            if (i >= chars.length()) {
                return;
            }

            char c = chars.charAt(i);
            if (c==0) {
                // used to switch escaping on and off
                disabled = !disabled;
            } else if (disabled) {
                writer.write(c);
            } else if (c>127) {
                if (XMLChar.isHighSurrogate(c)) {
                    int charval = XMLChar.supplemental(c, chars.charAt(++i));
                    if (characterSet.inCharset(charval)) {
                        writer.write(chars.subSequence(segstart+i-1, segstart+i+1).toString());
                    } else {
                        outputCharacterReference(charval);
                    }
                } else {
                    // process characters not available in the current encoding
                    outputCharacterReference(c);
                }

            } else {

                // process special ASCII characters

                if (c=='<') {
                    writer.write("&lt;");
                } else if (c=='>') {
                    writer.write("&gt;");
                } else if (c=='&') {
                    writer.write("&amp;");
                } else if (c=='\"') {
                    writer.write("&#34;");
                } else if (c=='\n') {
                    writer.write("&#xA;");
                } else if (c=='\r') {
                    writer.write("&#xD;");
                } else if (c=='\t') {
                    writer.write("&#x9;");
                }
            }
            segstart = ++i;
        }
    }

	/**
	* Output a decimal or hexadecimal character reference
	*/

    private char[] charref = new char[10];
    protected void outputCharacterReference(int charval) throws java.io.IOException {
		if (preferHex) {
	        int o = 0;
	        charref[o++]='&';
	        charref[o++]='#';
			charref[o++]='x';
	        String code = Integer.toHexString(charval);
	        int len = code.length();
	        for (int k=0; k<len; k++) {
	            charref[o++]=code.charAt(k);
	        }
	        charref[o++]=';';
	        writer.write(charref, 0, o);
		} else {
	        int o = 0;
	        charref[o++]='&';
	        charref[o++]='#';
	        String code = Integer.toString(charval);
	        int len = code.length();
	        for (int k=0; k<len; k++) {
	            charref[o++]=code.charAt(k);
	        }
	        charref[o++]=';';
	        writer.write(charref, 0, o);
	    }
    }

    /**
    * Handle a comment.
    */

    public void comment (CharSequence chars, int properties) throws TransformerException
    {
        if (empty) {
            openDocument();
        }
        try {
            if (openStartTag) {
                closeStartTag(null, false);
            }
            writer.write("<!--");
            writer.write(chars.toString());
            writer.write("-->");
        } catch (java.io.IOException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Set the URI for an unparsed entity in the document.
    */

    public void setUnparsedEntity(String name, String uri) throws TransformerException {
	}


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
