package net.sf.saxon.expr;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Value;
import net.sf.saxon.xpath.XPathException;

/**
* Assignation is an abstract superclass for the kinds of expression
* that declare range variables: for, some, and every.
*/

public abstract class Assignation extends ComputedExpression implements Binding {

    protected int slotNumber = -1;     // slot number for range variable
    protected Expression sequence;     // the expression over which the variable ranges
    protected Expression action;       // the action performed for each value of the variable
    protected String variableName;

    protected transient RangeVariableDeclaration declaration;

    /**
     * Set the reference to the variable declaration
     */

    public void setVariableDeclaration (RangeVariableDeclaration decl) {
        declaration = decl;
        variableName = decl.getVariableName();
    }

    /**
     * Add the "return" or "satisfies" expression, and fix up all references to the
     * range variable that occur within that expression
     * @param action the expression that occurs after the "return" keyword of a "for"
     * expression, the "satisfies" keyword of "some/every", or the ":=" operator of
     * a "let" expression.
     *
     * <p>This method must be called <b>after</b> calling setVariableDeclaration()</p>
     */

    public void setAction(Expression action) {
        this.action = action;
        declaration.fixupReferences(this);
    }

    /**
     * Set the "sequence" expression - the one to which the variable is bound
     */

    public void setSequence(Expression sequence) {
        this.sequence = sequence;
    }

    /**
    * Set the slot number for the range variable
    */

    public void setSlotNumber(int nr) {
        slotNumber = nr;
    }

    /**
    * Simplify the expression
    */

    public Expression simplify() throws XPathException {
        sequence = sequence.simplify();
        action = action.simplify();
        return this;
    }

    /**
    * Type-check the expression
    */

//    public Expression analyze(StaticContext env) throws XPathException {
//
//        if (declaration==null) {
//            // we've already done the type checking, no need to do it again
//            return this;
//        }
//
//        // The order of events is critical here. First we ensure that the type of the
//        // sequence expression is established. This is used to establish the type of the variable,
//        // which in turn is required when type-checking the action part.
//
//        sequence = sequence.analyze(env);
//
//        // "some" and "every" have no ordering constraints
//
//        if (this instanceof QuantifiedExpression) {
//            sequence = ExpressionTool.unsorted(sequence, false);
//        }
//
//        if (this instanceof LetExpression) {
//            RoleLocator role = new RoleLocator(RoleLocator.VARIABLE, getVariableName(), 0);
//            sequence = TypeChecker.staticTypeCheck(
//                            sequence, declaration.getRequiredType(), false, role);
//            int actualItemType = sequence.getItemType();
//            declaration.refineTypeInformation(actualItemType,
//                            sequence.getCardinality(),
//                            (sequence instanceof Value ? (Value)sequence : null),
//                            sequence.getSpecialProperties());
//        } else {
//            SequenceType decl = declaration.getRequiredType();
//            SequenceType sequenceType =
//                                new SequenceType(decl.getPrimaryType(),
//                                                    decl.getContentType(),
//                                                    StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE);
//            RoleLocator role = new RoleLocator(RoleLocator.VARIABLE, getVariableName(), 0);
//            sequence = TypeChecker.staticTypeCheck(
//                                    sequence, sequenceType, false, role);
//            int actualItemType = sequence.getItemType();
//            declaration.refineTypeInformation(actualItemType,
//                    StaticProperty.CARDINALITY_EXACTLY_ONE,
//                    null,
//                    sequence.getSpecialProperties());
//        }
//
//
//        //declaration.fixupReferences(this);
//        declaration = null;     // let the garbage collector take it
//
//        //if (references != null) {
//        //    for (Iterator iter = references.iterator(); iter.hasNext();) {
//        //        BindingReference ref = (BindingReference)iter.next();
//        //        ref.setStaticType(getRequiredType(), null, sequence.getProperties());
//        //        ref.fixup(this);
//        //    }
//        //    references = null; // to let the garbage collector have it
//        //}
//
//        action = action.analyze(env);
//
//        if (!(this instanceof LetExpression)) {
//            PromotionOffer offer = new PromotionOffer();
//            offer.containingExpression = this;
//            offer.action = PromotionOffer.RANGE_INDEPENDENT;
//            offer.binding = this;
//            action = action.promote(offer);
//            if (offer.containingExpression instanceof LetExpression) {
//                offer.containingExpression = offer.containingExpression.analyze(env);
//            }
//            return offer.containingExpression;
//        }
//
//        return this;
//    }

    /**
    * Promote this expression if possible
    */

    public Expression promote(PromotionOffer offer) throws XPathException {
        Expression exp = offer.accept(this);
        if (exp != null) {
            return exp;
        } else {
            sequence = sequence.promote(offer);
            if (offer.action == PromotionOffer.INLINE_VARIABLE_REFERENCES ||
                    offer.action == PromotionOffer.UNORDERED) {
                // Don't pass on other requests. We could pass them on, but only after augmenting
                // them to say we are interested in subexpressions that don't depend on either the
                // outer context or the inner context.
                action = action.promote(offer);
            }
            return this;
        }
    }

    /**
    * Get the immediate subexpressions of this expression
    */

    public Expression[] getSubExpressions() {
        Expression[] exp = new Expression[2];
        exp[0] = sequence;
        exp[1] = action;
        return exp;
    }

    // Following methods implement the VariableDeclaration interface, in relation to the range
    // variable

//    public boolean isGlobal() {
//        return false;
//    }

//    public int getVariableFingerprint() {
//        return nameCode & 0xfffff;
//    }

    /**
    * Get data type of the range variable (part of the Binding interface)
    */

    public SequenceType getRequiredType() {
        return new SequenceType(sequence.getItemType(), StaticProperty.EXACTLY_ONE);
    }

    /**
    * Get the display name of the range variable, for diagnostics only
    */

    public String getVariableName() {
        if (variableName == null) {
            return "zz:var" + hashCode();
        } else {
            return variableName;
        }
    }

    /**
    * Get the value of the range variable
    */

    public Value evaluateVariable(XPathContext context) throws XPathException {
        //return context.getController().getBindery().getLocalVariable(slotNumber);
        return context.evaluateLocalVariable(slotNumber);
    }

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
