package net.sf.saxon.expr;
import net.sf.saxon.om.Axis;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.pattern.NameTest;
import net.sf.saxon.pattern.NodeKindTest;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Type;
import net.sf.saxon.xpath.XPathException;

/**
* An expression that represents a reference to a named attribute
*/

// TODO: We are no longer using this class. See the commented-out code in
// AxisExpression. It probably no longer gives any performance benefits.
// By using a standard AxisIterator to find an attribute, we now avoid
// creating the attribute node if its untyped value is all that's required.

public class AttributeReference extends SingleNodeExpression {

	private int fingerprint;
    private NameTest test;

    /**
    * Constructor
    */

    public AttributeReference(int fingerprint) {
        this.fingerprint = fingerprint;
        this.test = new NameTest(Type.ATTRIBUTE, fingerprint);
    }

    /**
    * Determine the data type of the items returned by this expression
    * @return the type attribute()
    */

    public ItemType getItemType() {
        return NodeKindTest.ATTRIBUTE;
    }

    /**
     * Get the attribute name test
     */

    public NameTest getNameTest() {
        return test;
    }

    /**
    * getProperties(): return properties of this expression, for optimization
    */

    public int computeSpecialProperties() {
        return super.computeSpecialProperties() | StaticProperty.ATTRIBUTE_NS_NODESET;
    }

    /**
    * Is this expression the same as another expression?
    */

    public boolean equals(Object other) {
        return (other instanceof AttributeReference &&
            fingerprint == ((AttributeReference)other).fingerprint);
    }

    /**
    * get HashCode for comparing two expressions
    */

    public int hashCode() {
        return ("AttributeReference " + fingerprint).hashCode();
    }

    /**
    * Get the parent node
    */

    private NodeInfo getElementNode(XPathContext context) throws XPathException {
        Item item = context.getContextItem();
        if (item==null) {
            dynamicError("The context item is not set");
        }
        if (item instanceof NodeInfo) {
            return (NodeInfo)item;
        } else {
            return null;
        }
    }

    /**
    * Return the relevant attribute node
    * @param context the evaluation context
    */

    public NodeInfo getNode(XPathContext context) throws XPathException {
        NodeInfo element = getElementNode(context);
        return getAttributeNode(element);
    }

    private NodeInfo getAttributeNode(NodeInfo element) throws XPathException {
        if (element==null) {
            return null;
        }
        if (element.getNodeKind()==Type.ELEMENT) {
            SequenceIterator enum = element.iterateAxis(Axis.ATTRIBUTE, test);
            return (NodeInfo)enum.next();
        }
        return null;
    }

    /**
    * Evaluate as a boolean. Returns true if there are any nodes
    * selected by the Expression
    * @param context The context in which the expression is to be evaluated
    * @return true if there are any nodes selected by the Expression
    */

    // For some reason, this code has significantly worse performance than the
    // implementation in the superclass.

    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
        NodeInfo node = getElementNode(context);
        if (node==null) {
            return false;
        }
        if (node.getNodeKind()==Type.ELEMENT) {
            return node.getAttributeValue(fingerprint)!=null;
        }
        return false;
    }

    /**
    * Evaluate as a string. Returns the string value of the attribute if it exists
    * @param context The context in which the expression is to be evaluated
    * @return the string value of the attribute if present, or "" otherwise
    */

    public String evaluateAsString(XPathContext context) throws XPathException {
        NodeInfo node = getElementNode(context);
        if (node==null) {
            return "";
        }
        if (node.getNodeKind()==Type.ELEMENT) {
            String s = node.getAttributeValue(fingerprint);
            if (s==null) return "";
            return s;
        }
        return "";
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "@" + fingerprint);
    }

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
