package net.sf.saxon.expr;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.type.AtomicType;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Type;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.xpath.XPathException;

/**
* Cast Expression: implements "cast as data-type ( expression )". It also allows an internal
* cast, which has the same semantics as a user-requested cast, but maps an empty sequence to
* an empty sequence.
*/

public final class CastExpression extends ComputedExpression  {

    private Expression source;
    private AtomicType targetType;
    private boolean allowEmpty = false;

    public CastExpression(Expression source, AtomicType target) throws XPathException {
        this.source = source;
        this.targetType = target;
    }

    public CastExpression(Expression source, AtomicType target, boolean allowEmpty) throws XPathException {
        this.source = source;
        this.allowEmpty = allowEmpty;
        targetType = target;
    }

    /**
    * Simplify the expression
    * @return the simplified expression
    */

    public Expression simplify() throws XPathException {
        source = source.simplify();
        if (source instanceof AtomicValue) {
            if (Type.isSubType(targetType, Type.QNAME_TYPE)) {
                return this;
                // can't do it without a static context
            }
            return (AtomicValue)evaluateItem(null);
        }
        return this;
    }

    /**
    * Type-check the expression
    */

    public Expression analyze(StaticContext env) throws XPathException {
        source = source.analyze(env);
        SequenceType atomicType =
                new SequenceType(Type.ATOMIC_TYPE, getCardinality());

        RoleLocator role = new RoleLocator(RoleLocator.TYPE_OP, "cast as", 0);
        source = TypeChecker.staticTypeCheck(source, atomicType, false, role);

        if (Type.isSubType(source.getItemType(), targetType)) {
            return source;
            // TODO: wrong, should change the type label?
        }
        if (Type.isSubType(targetType, Type.QNAME_TYPE)) {
            return new CastAsQName(source).analyze(env);
        }
        if (source instanceof AtomicValue) {
            return (AtomicValue)evaluateItem(null);
        }
        return this;
    }

    /**
    * Promote this expression if possible
    */

    public Expression promote(PromotionOffer offer) throws XPathException {
        source = source.promote(offer);
        return this;
    }

    /**
    * Get the immediate subexpressions of this expression
    */

    public Expression[] getSubExpressions() {
        Expression[] exp = new Expression[1];
        exp[0] = source;
        return exp;
    }


    /**
    * Get the static cardinality of the expression
    */

    public int computeCardinality() {
        return (allowEmpty ? StaticProperty.ALLOWS_ZERO_OR_ONE : StaticProperty.EXACTLY_ONE);
    }

    /**
    * Get the static type of the expression
    */

    public ItemType getItemType() {
        return targetType;
    }


    /**
    * Evaluate the expression
    */

    public Item evaluateItem(XPathContext context) throws XPathException {
        AtomicValue value = (AtomicValue)source.evaluateItem(context);
        if (value==null) {
            if (allowEmpty) {
                throw new XPathException.Dynamic("Cast does not allow an empty sequence");
            } else {
                return null;
            }
        }
        try {
            return value.convert(targetType);
        } catch (XPathException err) {
            // add location information
            dynamicError(err.getMessage());
            return null;
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "cast as " +
                targetType.toString());
        source.display(level+1, pool);
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
