package net.sf.saxon.expr;

import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.value.Value;
import net.sf.saxon.value.Cardinality;
import net.sf.saxon.value.Closure;
import net.sf.saxon.value.EmptySequence;
import net.sf.saxon.value.SequenceExtent;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.SingletonIterator;
import net.sf.saxon.om.Item;

import javax.xml.transform.SourceLocator;

/**
 * This class, ExpressionTool, contains a number of useful static methods
 * for manipulating expressions. Most importantly, it provides the factory
 * method make() for constructing a new expression
 */

public class ExpressionTool {

    /**
     * Parse an expression. This performs the basic analysis of the expression against the
     * grammar, it binds variable references and function calls to variable definitions and
     * function definitions, and it performs context-independent expression rewriting for
     * optimization purposes.
     *
     * @exception XPathException if the expression contains a static error
     * @param expression The expression (as a character string)
     * @param env An object giving information about the compile-time
     *     context of the expression
     * @param terminator The token that marks the end of this expression; typically
     * Tokenizer.EOF, but may for example be a right curly brace
     * @return an object of type Expression
     */

    public static Expression make(String expression, StaticContext env, int start, int terminator) throws XPathException {
        try {
            // System.err.println("Parse " + expression);
            ExpressionParser parser = new ExpressionParser();
            if (terminator == -1) {
                terminator = Tokenizer.EOF;
            }
            Expression exp = parser.parse(expression, start, terminator, env);
            // System.err.println("Before simplify:"); exp.display(5);
            exp = exp.simplify();
            // System.err.println("After simplify:"); exp.display(0);
            //if (exp instanceof Value) {
            //    return exp;
            //} else {
            //    return new TopLevelExpression(exp);
            //}
            return exp;
        } catch (XPathException.Dynamic err) {
            // Don't report any dynamic errors now: defer them until runtime
            return new ErrorExpression(err);
        }
    }

    /**
     * Copy location information (currently just the line number) from one expression
     * to another
     */

    public static void copyLocationInfo(Expression from, Expression to) {
        if (from instanceof ComputedExpression && to instanceof ComputedExpression) {
            ((ComputedExpression)to).setLineNumber(((ComputedExpression)from).getLineNumber());
        }
    }

    /**
     * Get location information for an expression in the form of a SourceLocator
     */

    public static SourceLocator getLocator(Expression exp) {
        if (exp instanceof ComputedExpression) {
            ExpressionLocation loc = new ExpressionLocation();
            loc.setLineNumber(((ComputedExpression)exp).getLineNumber());
            return loc;
        } else {
            return null;
        }
    }

    /**
     * Remove unwanted sorting from an expression, at compile time
     */

    public static Expression unsorted(Expression exp, boolean eliminateDuplicates)
    throws XPathException {
        PromotionOffer offer = new PromotionOffer();
        offer.action = PromotionOffer.UNORDERED;
        offer.mustEliminateDuplicates = eliminateDuplicates;
        return exp.promote(offer);
    }

    /**
     * Do lazy evaluation of an expression. This will return a value, which may optionally
     * be a SequenceIntent, which is a wrapper around an iterator over the value of the expression.
     *
     * @param context the run-time evaluation context for the expression. If
     *     the expression is not evaluated immediately, then parts of the
     *     context on which the expression depends need to be saved as part of
     *      the Closure
     * @exception XPathException if any error occurs in evaluating the
     *     expression
     * @return a value: either the actual value obtained by evaluating the
     *     expression, or a Closure containing all the information needed to
     *     evaluate it later
     */

    public static Value lazyEvaluate(Expression exp, XPathContext context) throws XPathException {
        if (exp instanceof VariableReference) {
            // we always dereference the variable reference; this will often
            // do lazy evaluation of the expression to which the variable is bound
            return eagerEvaluate(exp, context);
        } else if (context == null) {
            // this means we are evaluating a value rather than an expression
            return eagerEvaluate(exp, context);
        } else if (!Cardinality.allowsMany(exp.getCardinality())) {
            // singleton expressions are always evaluated eagerly
            return eagerEvaluate(exp, context);
        } else if ((exp.getDependencies() &
                (   StaticProperty.DEPENDS_ON_POSITION |
                    StaticProperty.DEPENDS_ON_LAST |
                    StaticProperty.DEPENDS_ON_CURRENT_ITEM |
                    StaticProperty.DEPENDS_ON_CURRENT_GROUP)) != 0) {
            // we can't save these values in the closure, so we evaluate
            // the expression now if they are needed
            return eagerEvaluate(exp, context);
        } else {
            // create a Closure, a wrapper for the expression and its context
            return Closure.make(exp, context);
        }

    }

    /**
     * Evaluate the expression now; lazy evaluation is not permitted in this case
     *
     * @param context the run-time evaluation context
     * @exception XPathException if any dynamic error occurs evaluating the
     *     expression
     * @return the result of evaluating the expression
     */

    public static Value eagerEvaluate(Expression exp, XPathContext context) throws XPathException {
        if (exp instanceof Value && !(exp instanceof Closure)) {
            return (Value)exp;
        }
        SequenceIterator iterator = exp.iterate(context);
        if (iterator instanceof EmptyIterator) {
            return EmptySequence.getInstance();
        } else if (iterator instanceof SingletonIterator) {
            Item item = ((SingletonIterator)iterator).getValue();
            return Value.asValue(item);
        }
        SequenceExtent extent = new SequenceExtent(iterator);
        int len = extent.getLength();
        if (len==0) {
            return EmptySequence.getInstance();
        } else if (len==1) {
            return Value.asValue(extent.itemAt(0));
        } else {
            return extent;
        }

    }

    public static boolean markTailFunctionCalls(Expression exp) {
        if (exp instanceof ComputedExpression) {
            return ((ComputedExpression)exp).markTailFunctionCalls();
        } else {
            return false;
        }
    }

    /**
     * Construct indent string, for diagnostic output
     *
     * @param level the indentation level (the number of spaces to return)
     * @return a string of "level*2" spaces
     */

    public static String indent(int level) {
        String s = "";
        for (int i=0; i<level; i++) {
            s += "  ";
        }
        return s;
    }

    /**
    * Allocate slot numbers to range variables
    */

    public static int allocateSlots(Expression exp, int nextFree) {
        if (exp instanceof Assignation) {
            ((Assignation)exp).setSlotNumber(nextFree++);
        }
        Expression[] children = exp.getSubExpressions();
        for (int i=0; i<children.length; i++) {
            nextFree = allocateSlots(children[i], nextFree);
        }
        return nextFree;

        // Note, we allocate a distinct slot to each range variable, even if the
        // scopes don't overlap. This isn't strictly necessary, but might help
        // debugging.
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
