package net.sf.saxon.expr;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.value.*;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.AnyItemType;

/**
* Variable reference: a reference to a variable. This may be an XSLT-defined variable, a range
* variable defined within the XPath expression, or a variable defined in some other static context.
*/

public class VariableReference extends ComputedExpression implements BindingReference {

    Binding binding = null;     // This will be null until fixup() is called
    SequenceType staticType = null;
    Value constantValue = null;
    transient String displayName = null;

    /**
    * Constructor
    * @param declaration the variable declaration to which this variable refers
    */

    public VariableReference(VariableDeclaration declaration) throws XPathException {

        // Register this variable reference with the variable declaration. When the variable declaration
        // is compiled, the declaration will call the fixup() method of the variable reference. Note
        // that the object does not retain a pointer to the variable declaration, which would cause the
        // stylesheet to be locked in memory.

        // System.err.println("Register reference " + this + " with declaration " + declaration + " name=" + declaration.getVariableName());
        declaration.registerReference(this);
        displayName = declaration.getVariableName();
    }

    /**
    * Simplify the expression. Does nothing.
    */

    public Expression simplify() {
        return this;
    }

    /**
    * Set static type. This is a callback from the variable declaration object. As well
    * as supplying the static type, it may also supply a compile-time value for the variable.
    * As well as the type information, other static properties of the value are supplied:
    * for example, whether the value is an ordered node-set.
    */

    public void setStaticType(SequenceType type, Value value, int properties) {
        // System.err.println(this + " Set static type = " + type);
        staticType = type;
        constantValue = value;
        staticProperties = properties | type.getCardinality();
    }

    /**
    * Type-check the expression. At this stage details of the static type must be known.
    * If the variable has a compile-time value, this is substituted for the variable reference
    */

    public Expression analyze(StaticContext env) throws XPathException {
        if (constantValue != null) {
            return constantValue;
        }
        if (staticType==null) {
            throw new IllegalStateException("Variable $" + displayName + " has not been fixed up");
        } else {
            return this;
        }
    }

    /**
    * Fix up this variable reference to a Binding object, which enables the value of the variable
    * to be located at run-time.
    */

    public void fixup(Binding binding) {
        // System.err.println("Binding for " + this + " is " + binding.getVariableName());
        this.binding = binding;
    }

    /**
    * Determine the data type of the expression, if possible
    * @return the type of the variable, if this can be determined statically;
    * otherwise Type.ITEM (meaning not known in advance)
    */

    public ItemType getItemType() {
        if (staticType==null) {
            return AnyItemType.getInstance();
        } else {
            return staticType.getPrimaryType();
        }
    }

    /**
    * Get the static cardinality
    */

    public int computeCardinality() {
        if (staticType==null) {
            return StaticProperty.ALLOWS_ZERO_OR_MORE;
        } else {
            return staticType.getCardinality();
        }
    }

    /**
    * Test if this expression is the same as another expression.
    * (Note, we only compare expressions that
    * have the same static and dynamic context).
    */

    public boolean equals(Object other) {
        return (other instanceof VariableReference &&
                binding == ((VariableReference)other).binding &&
                binding != null);
    }

    /**
    * get HashCode for comparing two expressions
    */

    public int hashCode() {
        return binding.hashCode();
    }


    public int getIntrinsicDependencies() {
        return 0;
    }

    /**
    * Promote this expression if possible
    */

    public Expression promote(PromotionOffer offer) throws XPathException {
        if (offer.action == PromotionOffer.INLINE_VARIABLE_REFERENCES) {
            Expression exp = offer.accept(this);
            if (exp != null) {
                return exp;
            }
        }
        return this;
    }

    /**
    * Get the value of this variable in a given context.
    * @param c the XPathContext which contains the relevant variable bindings
    * @return the value of the variable, if it is defined
    * @throws XPathException if the variable is undefined
    */

    public SequenceIterator iterate(XPathContext c) throws XPathException {
        // System.err.println("Evaluate variable " + binding.getVariableName() + ", context = " + c);
        Value val = evaluateVariable(c);
        return val.iterate(c);

    }

    public Item evaluateItem(XPathContext c) throws XPathException {
        Value actual = evaluateVariable(c);
        return Value.asItem(actual, c);
    }

    private Value evaluateVariable(XPathContext c) throws XPathException {

        if (binding==null) {
            // System.err.println("No binding for " + this);
            throw new IllegalStateException("Variable $" + displayName + " has not been fixed up");
        }

        return binding.evaluateVariable(c);
    }

    /**
    * Get the object bound to the variable
    */

    public Binding getBinding() {
        return binding;
    }

    /**
    * Get the static properties of this expression (other than its type). The result is
    * bit-signficant. These properties are used for optimizations. In general, if
    * property bit is set, it is true, but if it is unset, the value is unknown.
    */
//
//    public int getProperties() {
//        return staticProperties;
//    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        if (displayName != null) {
            System.err.println(ExpressionTool.indent(level) + "$" + displayName);
        } else {
            System.err.println(ExpressionTool.indent(level) + "$(unbound variable)");
        }
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s):
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//
