package net.sf.saxon.functions;
import net.sf.saxon.Controller;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.MappingFunction;
import net.sf.saxon.expr.MappingIterator;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NamespaceResolver;
import net.sf.saxon.sort.DocumentOrderIterator;
import net.sf.saxon.sort.LocalOrderComparer;
import net.sf.saxon.style.ExpressionContext;
import net.sf.saxon.trans.KeyManager;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.Cardinality;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.xpath.XPathException;


public class Key extends SystemFunction implements MappingFunction, XSLTFunction {

    private class KeyContextInfo {
        public DocumentInfo document;
        public Controller controller;
        public int keyFingerprint;
    }

    private NamespaceResolver nsContext = null;
    private int keyFingerprint = -1;

    /**
    * Simplify: add a third implicit argument, the context document
    */

    public Expression simplify() throws XPathException {
        Key f = (Key)super.simplify();
        f.addContextDocumentArgument(2, "key+");
        return f;
    }

    public void checkArguments(StaticContext env) throws XPathException {
        super.checkArguments(env);
        argument[1] = ExpressionTool.unsorted(argument[1], false);
        if (argument[0] instanceof StringValue) {
            // common case, key name is supplied as a constant
            keyFingerprint = ((ExpressionContext)env).getFingerprint(((StringValue)argument[0]).getStringValue(), false);
            if (keyFingerprint==-1) {
                dynamicError("Key " + ((StringValue)argument[0]).getStringValue() + " is not defined");
            }
        } else {
            // we need to save the namespace context
            nsContext = ((ExpressionContext)env).getNamespaceContext();
        }
    }

    /**
    * Get the static properties of this expression (other than its type). The result is
    * bit-signficant. These properties are used for optimizations. In general, if
    * property bit is set, it is true, but if it is unset, the value is unknown.
    */

    public int computeSpecialProperties() {
        return StaticProperty.CONTEXT_DOCUMENT_NODESET | StaticProperty.ORDERED_NODESET;
    }

    /**
    * preEvaluate: this method suppresses compile-time evaluation by doing nothing
    */

    public Expression preEvaluate(StaticContext env) {
        return this;
    }

    /**
    * Enumerate the results of the expression
    */

    public SequenceIterator iterate(XPathContext context) throws XPathException {

        Controller controller = context.getController();

        Item arg2 = argument[2].evaluateItem(context);
        if (!(arg2 instanceof DocumentInfo)) {
            dynamicError("In the key() function," +
                            " the context node must be in a tree whose root is a document node");
            return null;
        }
        DocumentInfo doc = (DocumentInfo)arg2;

        int fprint = keyFingerprint;
        if (fprint == -1) {
            String givenkeyname = argument[0].evaluateItem(context).getStringValue();
            fprint = nsContext.getFingerprint(givenkeyname, false, context.getController().getNamePool());
            if (fprint==-1) {
                dynamicError("Key '" + givenkeyname + "' has not been defined");
                return null;
            }
        }

        // If the second argument is a singleton, we evaluate the function
        // directly; otherwise we recurse to evaluate it once for each Item
        // in the sequence.

        Expression expression = argument[1];
        if (!Cardinality.allowsMany(expression.getCardinality())) {
            AtomicValue keyValue = (AtomicValue)argument[1].evaluateItem(context);
            if (keyValue == null) {
                return EmptyIterator.getInstance();
            }
            KeyManager keyManager = controller.getKeyManager();
            return keyManager.selectByKey(fprint, doc, keyValue, controller);

        } else {
            KeyContextInfo info = new KeyContextInfo();
            info.document = doc;
            info.controller = controller;
            info.keyFingerprint = fprint;

            SequenceIterator keys = argument[1].iterate(context);
            SequenceIterator allValues = new MappingIterator(keys, this, null, info);
            return new DocumentOrderIterator(allValues, LocalOrderComparer.getInstance());
        }
    }



    /**
    * Implement the MappingFunction interface
    */

    public Object map(Item item, XPathContext context, Object info) throws XPathException {
        KeyContextInfo k = (KeyContextInfo)info;
        KeyManager keyManager = k.controller.getKeyManager();
        AtomicValue keyValue;
        if (item instanceof AtomicValue) {
            keyValue = (AtomicValue)item;
        } else {
            keyValue = new StringValue(item.getStringValue());
        }
        return keyManager.selectByKey(
                k.keyFingerprint, k.document, keyValue, k.controller);

        // TODO: this won't handle sequence-valued nodes. For that we need two levels
        // of iteration, one over the nodes, and one over the items in the value of each node.
        // (Or more generally, an atomizing iterator)
    }

}





//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
