package net.sf.saxon.instruct;

import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.PromotionOffer;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.SingletonIterator;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.xpath.XPathException;

import java.util.ArrayList;
import java.util.List;

/**
 * This class is an abstract superclass for Instructions that also behave as Expressions.
 * Currently this is restricted to those instructions used when an XQuery query is compiled.
 * Eventually it is intended that all instructions should support the Expression interface.
 * Note that the Expression interfaces (e.g. iterate()) may only be used on an ExprInstruction
 * if all its child instructions are also instances of ExprInstruction.
 */

public abstract class ExprInstruction extends Instruction implements Expression {

    public Expression simplify() throws XPathException {
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    children[c] = (ExprInstruction)((Expression)children[c]).simplify();
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be Expressions");
                }
            }
        }
        return this;
    }

    public ItemType getItemType() {
        return getResultType().getPrimaryType();
    }

    public int getCardinality() {
        return getResultType().getCardinality();
    }

    public int getDependencies() {
        int dep = 0;
        Expression[] sub = getSubExpressions();
        for (int i=0; i<sub.length; i++) {
            dep |= sub[i].getDependencies();
        }
        return dep;
    }

    public int getSpecialProperties() {
        return StaticProperty.CREATES_NEW_NODES;
        // TODO: we can do better than this!
    }

    protected abstract void promoteInst(PromotionOffer offer) throws XPathException;

    public Expression promote(PromotionOffer offer) throws XPathException {
        Expression exp = offer.accept(this);
        if (exp!=null) {
            return exp;
        } else {
            promoteInst(offer);
            return this;
        }
    }

    protected abstract void getXPathExpressions(List list);

    public Expression[] getSubExpressions() {
        List list = new ArrayList();
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    Expression[] array = ((Expression)children[c]).getSubExpressions();
                    for (int i=0; i<array.length; i++) {
                        list.add(array[i]);
                    }
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be Expressions");
                }
            }
        }
        getXPathExpressions(list);
        Expression[] array = new Expression[list.size()];
        for (int i=0; i<list.size(); i++) {
            array[i] = (Expression)list.get(i);
        }
        return array;
    }

    public Item evaluateItem(XPathContext context) throws XPathException {
        return iterate(context).next();
    }

    public final String evaluateAsString(XPathContext context) throws XPathException {
        Item item = evaluateItem(context);
        if (item==null) {
            return "";
        } else {
            return item.getStringValue();
        }
    }

    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
        SequenceIterator iter = iterate(context);
        Item item = iter.next();
        if (item == null) {
            return false;
        }
        if (item instanceof NodeInfo) {
            return true;
        } else {
            //TODO: appears to be wrong for a sequence of several atomic values...
            return ((AtomicValue)item).effectiveBooleanValue(context);
        }
    }

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        Item item = evaluateItem(context);
        return SingletonIterator.makeIterator(item);
    }

}
