package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Orphan;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.event.SequenceOutputter;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.TypeChecker;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.RoleLocator;

import javax.xml.transform.TransformerException;
import java.util.List;

/**
 * Common superclass for XSLT instructions whose content template produces a text
 * value: xsl:attribute, xsl:comment, xsl:processing-instruction, xsl:namespace,
 * and xsl:text
 */

public abstract class SimpleNodeConstructor extends ExprInstruction {

    protected Expression select = null;
    protected Expression separator = null;

    public final void setSelect(Expression select) {
        this.select = select;
    }

    public final void setSeparator(Expression separator) {
        this.separator = separator;
    }

    public abstract void typeCheck(StaticContext env) throws XPathException;

    /**
     * The analyze() method is called in XQuery, where node constructors
     * are implemented as Expressions. In this case the required type for the
     * select expression is a single string.
     * @param env The static context for the query
     * @return the rewritten expression
     * @throws XPathException if any static errors are found in this expression
     * or any of its children
     */

    public Expression analyze(StaticContext env) throws XPathException {
        typeCheck(env);
        if (separator != null) {
            separator = separator.analyze(env);
            RoleLocator role =
                new RoleLocator(RoleLocator.INSTRUCTION, "xsl:" + getInstructionName() + "/separator", 0);
            separator = TypeChecker.staticTypeCheck(separator, SequenceType.SINGLE_STRING, false, role);
        }
        if (select != null) {
            select = select.analyze(env);
            RoleLocator role =
                new RoleLocator(RoleLocator.INSTRUCTION, "xsl:" + getInstructionName() + "/content", 0);
            select = TypeChecker.staticTypeCheck(select, SequenceType.SINGLE_STRING, false, role);
        } else {
            if (children != null) {
                for (int c=0; c<children.length; c++) {
                    if (children[c] instanceof ExprInstruction) {
                        children[c] = (ExprInstruction)((ExprInstruction)children[c]).analyze(env);
                    } else {
                        throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
                    }
                }
            }
        }
        return this;
    }


    public void getXPathExpressions(List list) {
        if (select != null) {
            list.add(select);
        }
        if (separator != null && !(separator instanceof StringValue)) {
            list.add(separator);
        }
    }

    /**
    * Expand the stylesheet elements subordinate to this one, returning the result
    * as a string. The expansion must not generate any element or attribute nodes.
    * @param context The dynamic context for the transformation
    */

    public CharSequence expandChildren(XPathContext context) throws TransformerException {

        String sep = " ";
        if (separator != null) {
            sep = separator.evaluateAsString(context);
        }

        if (select != null) {
            if (select instanceof StringValue) {
                return ((StringValue)select).getStringValue();
            } else {
                return flatten(select.iterate(context), sep);
            }

        } else {
            Controller controller = context.getController();
            SequenceReceiver old = controller.getReceiver();
            //StringBuffer buffer = new StringBuffer();
            //controller.changeToTextOutputDestination(buffer);
            SequenceOutputter seq = new SequenceOutputter();
            seq.setNamePool(controller.getNamePool());
            controller.changeToSequenceOutputDestination(seq);
            // process the child elements in the stylesheet
            processChildren(context);
            controller.resetOutputDestination(old);
            return flatten(seq.getSequence().iterate(context), sep);

        }
    }

    /**
     * Flatten the value of the sequence to a character string
     */

    private StringBuffer flatten(SequenceIterator content, String separator)
    throws XPathException {
        StringBuffer buffer = new StringBuffer();
        boolean first = true;
        while(true) {
            Item item = content.next();
            if (item == null) {
                break;
            }
            if (first) {
                first = false;
            } else {
                buffer.append(separator);
            }
            buffer.append(item.getStringValue());
        }
        return buffer;
    }

    /**
     * Evaluate as an expression. We rely on the fact that when these instructions
     * are generated by XQuery, there will always be a valueExpression to evaluate
     * the content
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        String content = (select==null ?
                    "" :
                    select.evaluateAsString(context));
        Orphan o = new Orphan(context.getController().getNamePool());
        o.setNodeKind((short)getItemType().getPrimitiveType());
        o.setStringValue(content);
        try {
            o.setNameCode(evaluateNameCode(context));
        } catch (XPathException err) {
            throw err;
        } catch (TransformerException err) {
            throw new XPathException.Dynamic(err);
        }
        return o;
    }

    protected int evaluateNameCode(XPathContext context)
    throws XPathException, TransformerException {
        return -1;
    }

    /**
     * Display this instruction as an expression, for diagnostics
     */

    public void display(int level, NamePool pool) {
        if (select != null) {
            select.display(level, pool);
        } else if (children.length==0) {
            System.err.println(ExpressionTool.indent(level) + "empty content");
        } else {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    ((Expression)children[c]).display(level+1, pool);
                } else {
                    System.err.println(ExpressionTool.indent(level+1) + children[c].getInstructionName());
                }
            }
        }
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
