package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.trace.InstructionInfo;
import net.sf.saxon.trace.TraceListener;

import javax.xml.transform.TransformerException;

/**
* An xsl:template element in the style sheet.
*/

public class Template extends Instruction {

    // The body of the template is represented by a generated xsl:sequence element,
    // which is responsible for any type checking that's needed.

    private Sequence body;
    private int precedence;
    private int minImportPrecedence;
    private boolean needsStackFrame;

    public Template () {}

    public void init (  Sequence body,
                        boolean needsStackFrame,
                        int precedence,
                        int minImportPrecedence) {
        this.body = body;
        this.needsStackFrame = needsStackFrame;
        this.precedence = precedence;
        this.minImportPrecedence = minImportPrecedence;
    }

    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("template");
    }

    public int getPrecedence() {
        return precedence;
    }

    public int getMinImportPrecedence() {
        return minImportPrecedence;
    }

    public boolean needsStackFrame() {
        return this.needsStackFrame;
    }

    /**
    * Process a node in the source document. This is called when the template
    * is invoked using xsl:apply-templates.
    */

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        Controller controller = context.getController();
        if (body==null) {
            // fast path for an empty template
            return null;
        }
        Template savedTemplate = controller.getCurrentTemplate();
        controller.setCurrentTemplate(this);

        TailCall tc;
    	if (controller.isTracing()) { // TODO: trace tail recursion
            tc = traceExpand(controller);
        } else {
            tc = expand(controller);
        }

        controller.setCurrentTemplate(savedTemplate);
        return tc;
    }

    /**
    * Expand the template, with tracing. Called when the template is invoked either
    * by xsl:apply-templates or from xsl:call-template
    */

    protected TailCall traceExpand(Controller controller) throws TransformerException {
	    TraceListener listener = controller.getTraceListener();
	    InstructionDetails info = getInstructionDetails(controller);
    	listener.enter(info);
        TailCall tc = expand(controller);
  		listener.leave(info);
  		return tc;
    }

    /**
    * Expand the template. Called when the template is invoked either
    * by xsl:apply-templates or from xsl:call-template
    */

    protected TailCall expand(Controller controller) throws TransformerException {
        if (body==null) {
            // fast path for an empty template
            return null;
        }
        return body.processLeavingTail(controller.newXPathContext());
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s):
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//
