package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.ReceiverOptions;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.PromotionOffer;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.Orphan;
import net.sf.saxon.pattern.NodeKindTest;
import net.sf.saxon.type.Type;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerException;

/**
* An xsl:value-of element in the stylesheet. <br>
* The xsl:value-of element takes attributes:<ul>
* <li>a mandatory attribute select="expression".
* This must be a valid String expression</li>
* <li>an optional disable-output-escaping attribute, value "yes" or "no"</li>
* <li>an optional separator attribute. This is handled at compile-time: if the separator attribute
* is present, the select expression passed in here will be a call to the string-join() function.</li>
* </ul>
*/

public final class ValueOf extends SimpleNodeConstructor {

    // TODO: merge this class with Text

    private int options;

    public ValueOf( Expression select,
                    boolean disable ) {
        this.select = select;
        this.options = (disable ? ReceiverOptions.DISABLE_ESCAPING : 0);
    }

    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("value-of");
    }

    /**
    * Offer promotion for subexpressions. The offer will be accepted if the subexpression
    * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
    * By default the offer is not accepted - this is appropriate in the case of simple expressions
    * such as constant values and variable references where promotion would give no performance
    * advantage. This method is always called at compile time.
    *
    * @param offer details of the offer, for example the offer to move
    *     expressions that don't depend on the context to an outer level in
    *     the containing expression
    * @exception XPathException if any error is detected
    */

    public void promoteInst(PromotionOffer offer) throws XPathException {
        // TODO: move this up to SimpleNodeConstructor
        if (select != null) {
            select = select.promote(offer);
        }
        if (separator != null) {
            separator = separator.promote(offer);
        }
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof ExprInstruction) {
                    Expression p = ((ExprInstruction)children[c]).promote(offer);
                    if (p instanceof Instruction) {
                        children[c] = (Instruction)p;
                    } else {
                        children[c] = new Sequence(p, null);
                    }
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
                }
            }
        }
    }


    public SequenceType getResultType() {
         return new SequenceType(NodeKindTest.TEXT, StaticProperty.EXACTLY_ONE);
    }

    public void typeCheck(StaticContext env) throws XPathException {

    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        Receiver out = context.getController().getReceiver();
        out.characters(expandChildren(context), options);
        return null;
    }

    public Item evaluateItem(XPathContext context) throws XPathException {
        try {
            CharSequence val = expandChildren(context);
            Orphan o = new Orphan(context.getController().getNamePool());
            o.setNodeKind(Type.TEXT);
            o.setStringValue(val);
            return o;
        } catch (TransformerException err) {
            throw new XPathException.Dynamic(err);
        }
    }

    /**
     * Display this instruction as an expression, for diagnostics
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "value-of");
        select.display(level+1, pool);
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
