package net.sf.saxon.query;

import net.sf.saxon.event.DocumentSender;
import net.sf.saxon.event.NamespaceReducer;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.ResultWrapper;
import net.sf.saxon.om.*;
import net.sf.saxon.tinytree.TinyBuilder;
import net.sf.saxon.type.Type;

import javax.xml.transform.Result;
import javax.xml.transform.TransformerException;
import java.util.Properties;

/**
 * This class takes the result sequence produced by a query, and wraps it as
 * an XML document
 */
public class QueryResult {

    private static String RESULT_NS = "http://saxon.sf.net/xquery-results";

    /**
     * Take the results of a query (or any other SequenceIterator) and create
     * an XML document containing copies of all items in the sequence, suitably wrapped
     * @param iterator  The values to be wrapped
     * @param pool      NamePool to use for the constructed document
     * @return the document containing the wrapped results
     * @throws TransformerException
     */
    public static DocumentInfo wrap(SequenceIterator iterator, NamePool pool) throws TransformerException {

        TinyBuilder builder = new TinyBuilder();

        NamespaceReducer reducer = new NamespaceReducer();
        reducer.setUnderlyingReceiver(builder);
        Receiver tree = reducer;

        tree.setNamePool(pool);
        builder.setNamePool(pool);
        tree.startDocument();

        int resultSequence = pool.allocate("result", RESULT_NS, "sequence");
        int resultDocument = pool.allocate("result", RESULT_NS, "document");
        int resultElement = pool.allocate("result", RESULT_NS, "element");
        int resultAttribute = pool.allocate("result", RESULT_NS, "attribute");
        int resultText = pool.allocate("result", RESULT_NS, "text");
        int resultComment = pool.allocate("result", RESULT_NS, "comment");
        int resultPI = pool.allocate("result", RESULT_NS, "processing-instruction");
        int resultNamespace = pool.allocate("result", RESULT_NS, "namespace");
        int resultAtomicValue = pool.allocate("result", RESULT_NS, "atomic-value");
        int resultType = pool.allocate("", RESULT_NS, "type");


        tree.startElement(resultSequence, 0, 0);
        tree.namespace(pool.allocateNamespaceCode("result", RESULT_NS), 0);
        tree.startContent();

        while (true) {
            Item next = iterator.next();
            if (next == null) {
                break;
            }
            if (next instanceof NodeInfo) {
                switch (((NodeInfo)next).getNodeKind()) {
                    case Type.DOCUMENT:
                        tree.startElement(resultDocument, 0, 0);
                        tree.startContent();
                        ((DocumentInfo)next).copy(tree, NodeInfo.ALL_NAMESPACES, true);
                        tree.endElement();
                        break;
                    case Type.ELEMENT:
                        tree.startElement(resultElement, 0, 0);
                        tree.startContent();
                        ((NodeInfo)next).copy(tree, NodeInfo.ALL_NAMESPACES, true);
                        tree.endElement();
                        break;
                    case Type.ATTRIBUTE:
                        tree.startElement(resultAttribute, 0, 0);
                        ((NodeInfo)next).copy(tree, NodeInfo.ALL_NAMESPACES, true);
                        tree.startContent();
                        tree.endElement();
                        break;
                    case Type.TEXT:
                        tree.startElement(resultText, 0, 0);
                        tree.startContent();
                        ((NodeInfo)next).copy(tree, NodeInfo.ALL_NAMESPACES, true);
                        tree.endElement();
                        break;
                    case Type.COMMENT:
                        tree.startElement(resultComment, 0, 0);
                        tree.startContent();
                        ((NodeInfo)next).copy(tree, NodeInfo.ALL_NAMESPACES, true);
                        tree.endElement();
                        break;
                    case Type.PROCESSING_INSTRUCTION:
                        tree.startElement(resultPI, 0, 0);
                        tree.startContent();
                        ((NodeInfo)next).copy(tree, NodeInfo.ALL_NAMESPACES, true);
                        tree.endElement();
                        break;
                    case Type.NAMESPACE:
                        tree.startElement(resultNamespace, 0, 0);
                        ((NodeInfo)next).copy(tree, NodeInfo.ALL_NAMESPACES, true);
                        tree.startContent();
                        tree.endElement();
                        break;
                }
            } else {
                tree.startElement(resultAtomicValue, 0, 0);
                tree.attribute(resultType, 0, Type.displayTypeName(next), 0);
                tree.startContent();
                tree.characters(next.getStringValue(), 0);
                tree.endElement();
            }

        }
        tree.endElement();
        tree.endDocument();
        return builder.getCurrentDocument();
    }

    /**
     * Serialize a document containing wrapped query results (or any other document, in fact)
     * as XML.
     * @param node                      The document or element to be serialized
     * @param destination               The Result object to contain the serialized form
     * @param outputProperties          Serialization options
     * @throws TransformerException     If serialization fails
     */

    public static void serialize(NodeInfo node, Result destination, Properties outputProperties)
    throws TransformerException {
        int type = node.getNodeKind();
        if (type==Type.DOCUMENT || type==Type.ELEMENT) {
            NamePool pool = node.getNamePool();
            DocumentSender sender = new DocumentSender(node);
            Receiver receiver =
                    ResultWrapper.getReceiver(destination,
                                              pool,
                                              outputProperties,
                                              null);
            NamespaceReducer reducer = new NamespaceReducer();
            reducer.setUnderlyingReceiver(receiver);
            reducer.setNamePool(pool);
            sender.send(reducer);
        } else {
            throw new TransformerException("Node to be serialized must be a Document or Element node");
        }
    }

//    public static void main(String[] params) throws Exception {
//        StaticQueryContext env = new StaticQueryContext(new Configuration());
//        QueryProcessor qp = new QueryProcessor(env);
//        XQueryExpression exp = qp.compileQuery("<a><b/></a>");
//        SequenceIterator iter = exp.iterator(new DynamicQueryContext());
//        NodeInfo node = (NodeInfo)iter.next();
//        serialize(node, new SAXResult(new ExampleContentHandler()), new Properties());
//
//    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
