package net.sf.saxon.sort;

import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.Controller;

/**
 * A SortedGroupIterator is a modified SortedIterator. It sorts a sequence of groups,
 * and is itself a GroupIterator. The modifications retain extra information about
 * the items being sorted. The items are each the leading item of a group, and as well
 * as the item itself, the iterator preserves information about the group: specifically,
 * an iterator over the items in the group, and the value of the grouping key (if any).
 */

public class SortedGroupIterator extends SortedIterator implements GroupIterator {

    public SortedGroupIterator(XPathContext context, GroupIterator base,
                               FixedSortKeyDefinition[] sortKeys) throws XPathException {
        super(context, base, sortKeys);
        // add two items to each tuple, for the iterator over the items in the group,
        // and the grouping key, respectively.
        recordSize += 2;
    }

    /**
     * Override the method that builds the array of values and sort keys.
     * @throws XPathException
     */

    protected void buildArray() throws XPathException {
        int allocated = 100;
        nodeKeys = new Object[allocated * recordSize];
        count = 0;

        SequenceIterator savedIterator = context.getCurrentIterator();
        context.setCurrentIterator(base);
        Controller controller = context.getController();
        SequenceIterator savedXSLTIterator = controller.getCurrentIterator();
        GroupIterator savedGroupIterator = controller.getCurrentGroupIterator();
        controller.setCurrentIterator(base);
        controller.setCurrentGroupIterator((GroupIterator)base);
                // this provides the context for evaluating the sort key

        // initialise the array with data

        while (true) {
            Item item = base.next();
            if (item == null) {
                break;
            }
            if (count==allocated) {
                allocated *= 2;
                Object[] nk2 = new Object[allocated * recordSize];
                System.arraycopy(nodeKeys, 0, nk2, 0, count * recordSize);
                nodeKeys = nk2;
            }
            int k = count*recordSize;
            nodeKeys[k] = item;
            for (int n=0; n<sortkeys.length; n++) {
                nodeKeys[k+n+1] = sortkeys[n].getSortKey().evaluateItem(context);
            }
            nodeKeys[k+sortkeys.length+1] = new Integer(count);
            // extra code added to superclass
            nodeKeys[k+sortkeys.length+2] = ((GroupIterator)base).getCurrentGroupingKey();
            nodeKeys[k+sortkeys.length+3] = ((GroupIterator)base).iterateCurrentGroup();
            count++;
        }
        context.setCurrentIterator(savedIterator);
        controller.setCurrentIterator(savedXSLTIterator);
        controller.setCurrentGroupIterator(savedGroupIterator);
    }

    public AtomicValue getCurrentGroupingKey() {
        return (AtomicValue)nodeKeys[(index-1)*recordSize+sortkeys.length+2];
    }

    public SequenceIterator iterateCurrentGroup() throws XPathException {
        SequenceIterator iter =
                (SequenceIterator)nodeKeys[(index-1)*recordSize+sortkeys.length+3];
        return iter.getAnother();
    }
}
