package net.sf.saxon.sql;
import net.sf.saxon.Controller;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.style.StyleElement;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.ObjectValue;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.instruct.TailCall;
import net.sf.saxon.instruct.InstructionDetails;

import org.w3c.dom.Node;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
/**
* An sql:insert element in the stylesheet.
*/

public class SQLInsert extends StyleElement {

    Expression connection;
    String table;

    /**
    * Determine whether this node is an instruction.
    * @return true - it is an instruction
    */

    public boolean isInstruction() {
        return true;
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body (this is done only so that
    * it can contain xsl:fallback)
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    public void prepareAttributes() throws TransformerConfigurationException {

		table = getAttribute("table");
		if (table==null) {
            reportAbsence("table");
        }
        String connectAtt = getAttribute("connection");
        if (connectAtt==null) {
            reportAbsence("connection");
        } else {
            connection = makeExpression(connectAtt);
        }
    }

    public void validate() throws TransformerConfigurationException {
        checkWithinTemplate();
        connection = typeCheck("connection", connection);
    }

    public Instruction compile() throws TransformerConfigurationException {

        // Collect names of columns to be added

        StringBuffer statement = new StringBuffer();
        statement.append("INSERT INTO " + table + " (");

        Node child = getFirstChild();
		int cols = 0;
		while (child!=null) {
		    if (child instanceof SQLColumn) {
    			if (cols++ > 0)	statement.append(',');
    			String colname = ((SQLColumn)child).getColumnName();
    			statement.append(colname);
		    }
			child = child.getNextSibling();
		}
        statement.append(") VALUES (");

        // Add "?" marks for the variable parameters

		for(int i=0; i<cols; i++) {
			if (i!=0)
			    statement.append(',');
			statement.append('?');
		};

		statement.append(')');

        Instruction inst = new InsertInstruction(connection, statement.toString(), getColumnInstructions());
        return inst;
    }

    public List getColumnInstructions() throws TransformerConfigurationException {
        List list = new ArrayList();
        Node child = getFirstChild();
		while (child!=null) {
		    if (child instanceof SQLColumn) {
    			list.add(((SQLColumn)child).compile());
		    }
			child = child.getNextSibling();
		}
		return list;
    }

    private static class InsertInstruction extends Instruction {

        Expression connectExpression;
        String statement;
        List columnInstructions;

        public InsertInstruction(Expression connection, String statement, List columnInstructions) {
            this.connectExpression = connection;
            this.statement = statement;
            this.columnInstructions = columnInstructions;
        }

        public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
            details.setInstructionName("insert");
            details.setProperty("instruction-namespace", "java:/net.sf.saxon.sql.SQLElementFactory");
        }

        public TailCall processLeavingTail(XPathContext context) throws TransformerException {

            // Prepare the SQL statement (only do this once)

            Controller controller = context.getController();
            Item conn = connectExpression.evaluateItem(context);
            if (!(conn instanceof ObjectValue && ((ObjectValue)conn).getObject() instanceof Connection) ) {
                throw new TransformerException("Value of connection expression is not a JDBC Connection");
            }
            Connection connection = (Connection)((ObjectValue)conn).getObject();
            PreparedStatement ps = null;

            try {
              	ps=connection.prepareStatement(statement);

                // Add the actual column values to be inserted

                Iterator col = columnInstructions.iterator();
                int i=1;
                while (col.hasNext()) {
                    AtomicValue v = (AtomicValue)((SQLColumn.ColumnInstruction)col.next()).getSelectValue(context);

         			// TODO: the values are all strings. There is no way of adding to a numeric column
           		    String val = v.getStringValue();

           		    // another hack: setString() doesn't seem to like single-character string values
           		    if (val.length()==1) val += " ";
           		    //System.err.println("Set statement parameter " + i + " to " + val);
           			ps.setObject(i++, val);

    	        }

    			ps.executeUpdate();
    			if (!connection.getAutoCommit()) {
                    connection.commit();
                }

    	    } catch (SQLException ex) {
    			throw styleError("(SQL) " + ex.getMessage(), controller);
            } finally {
               if (ps != null) {
                   try {
                       ps.close();
                   } catch (SQLException ignore) {}
               }
            }

            return null;
        }

    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
