package net.sf.saxon.style;
import net.sf.saxon.PreparedStyleSheet;
import net.sf.saxon.type.SchemaType;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.instruct.AttributeSet;
import net.sf.saxon.instruct.FixedAttribute;
import net.sf.saxon.instruct.FixedElement;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NamespaceConstant;
import net.sf.saxon.om.Validation;
import net.sf.saxon.tree.DocumentImpl;
import net.sf.saxon.tree.TreeBuilder;
import net.sf.saxon.value.StringValue;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;


/**
* This class represents a literal result element in the style sheet
* (typically an HTML element to be output). <br>
* It is also used to represent unknown top-level elements, which are ignored.
*/

public class LiteralResultElement extends StyleElement {

    private int resultNameCode;
    private int[] attributeNames;
    private Expression[] attributeValues;
    private boolean[] attributeChecked;
    private int numberOfAttributes;
    private boolean toplevel;
    private int[] namespaceCodes;
    private AttributeSet[] attributeSets;
    private SchemaType schemaType = null;
    private int validation = Validation.STRIP;

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    /**
    * Process the attribute list
    */

    public void prepareAttributes() throws TransformerConfigurationException {

        // Process the values of all attributes. At this stage we deal with attribute
        // values (especially AVTs), but we do not apply namespace aliasing to the
        // attribute names.

        int num = attributeList.getLength();

        if (num == 0) {
            numberOfAttributes = 0;
        } else {
            NamePool namePool = getNamePool();
            attributeNames = new int[num];
            attributeValues = new Expression[num];
            attributeChecked = new boolean[num];
            numberOfAttributes = 0;

            for (int i=0; i<num; i++) {

                int anameCode = attributeList.getNameCode(i);
                short attURIcode = namePool.getURICode(anameCode);

                if (attURIcode==NamespaceConstant.XSLT_CODE) {
                    int fp = anameCode & 0xfffff;

                    if (fp == StandardNames.XSL_USE_ATTRIBUTE_SETS) {
                        // deal with this later
                    } else if (fp == StandardNames.XSL_EXTENSION_ELEMENT_PREFIXES) {
                    	// already dealt with
                    } else if (fp == StandardNames.XSL_EXCLUDE_RESULT_PREFIXES) {
                    	// already dealt with
                    } else if (fp == StandardNames.XSL_VERSION) {
                        // already dealt with
                    } else if (fp == StandardNames.XSL_XPATH_DEFAULT_NAMESPACE) {
                        // already dealt with
                    } else if (fp == StandardNames.XSL_TYPE) {
                        // deal with this later
                    } else if (fp == StandardNames.XSL_VALIDATION) {
                        // deal with this later
                    } else {
                        compileError("Unknown XSL attribute " + namePool.getDisplayName(anameCode));
                    }
                } else {
                    attributeNames[numberOfAttributes] = anameCode;
                    Expression exp = makeAttributeValueTemplate(attributeList.getValue(i));
                    attributeValues[numberOfAttributes] = exp;

                    // if we can be sure the attribute value contains no special XML/HTML characters,
                    // we can save the trouble of checking it each time it is output.
                    // Note that the check includes characters that need to be escaped in a URL if the
                    // output method turns out to be HTML (we don't know the method at compile time).

                    attributeChecked[numberOfAttributes] = false;
                    boolean special = false;
                    if (exp instanceof StringValue) {
                        String val = ((StringValue)exp).getStringValue();
                        for (int k=0; k<val.length(); k++) {
                            char c = val.charAt(k);
                            if ((int)c<33 || (int)c>126 ||
                                     c=='<' || c=='>' || c=='&' || c=='\"') {
                                special = true;
                                break;
                             }
                        }
                        attributeChecked[numberOfAttributes] = !special;
                    }
                    numberOfAttributes++;
                }
            }

            // now shorten the arrays if necessary. This is necessary if there are [xsl:]-prefixed
            // attributes that weren't copied into the arrays.

            if (numberOfAttributes < attributeNames.length) {

                int[] attributeNames2 = new int[numberOfAttributes];
                System.arraycopy(attributeNames, 0, attributeNames2, 0, numberOfAttributes);
                attributeNames = attributeNames2;

                Expression[] attributeValues2 = new Expression[numberOfAttributes];
                System.arraycopy(attributeValues, 0, attributeValues2, 0, numberOfAttributes);
                attributeValues = attributeValues2;

                boolean[] attributeChecked2 = new boolean[numberOfAttributes];
                System.arraycopy(attributeChecked, 0, attributeChecked2, 0, numberOfAttributes);
                attributeChecked = attributeChecked2;
            }
        }
    }

    /**
    * Validate that this node is OK
    */

    public void validate() throws TransformerConfigurationException {

        toplevel = (getParentNode() instanceof XSLStyleSheet);

        resultNameCode = getNameCode();

        NamePool namePool = getNamePool();
        short elementURICode = namePool.getURICode(resultNameCode);

        if (toplevel) {
            // A top-level element can never be a "real" literal result element,
            // but this class gets used for unknown elements found at the top level

            if (elementURICode == 0) {
                compileError("Top level elements must have a non-null namespace URI");
            }
        } else {

            // Build the list of output namespace nodes

            // Up to 5.3.1 we listed the namespace nodes associated with this element that were not also
            // associated with an ancestor literal result element (because those will already
            // have been output). Unfortunately this isn't true if the namespace was present on an outer
            // LRE, and was excluded at that level using exclude-result-prefixes, and is now used in an
            // inner element: bug 5.3.1/006

            // At 6.0 we'll try a different optimisation: if this LRE has a parent that is also
            // an LRE, and if this LRE has no namespace declarations or xsl:exclude-result-prefixes
            // attribute or xsl:extension-element-prefixes attribute of its own, and if this element
            // name is in the same namespace as its parent, and if there are no attributes in a
            // non-null namespace, then we don't need to output any namespace declarations to the
            // result.

            boolean optimizeNS = false;
            if ((getParent() instanceof LiteralResultElement) &&
                    (namespaceList==null || namespaceList.length==0) &&
                    ( elementURICode ==
                    	 namePool.getURICode(getParent().getFingerprint()))
                    ) {
                optimizeNS = true;
            }
            if (optimizeNS) {
                for (int a=0; a<attributeList.getLength(); a++ ) {
                    if (((attributeList.getNameCode(a)>>20)&0xff) != 0) {	// prefix != ""
                        optimizeNS = false;
                        break;
                    }
                }
            }

            if (optimizeNS) {
            	namespaceCodes = new int[0];
            	// TODO: is this correct when namespace aliases are in use?
            } else {
                namespaceCodes = getNamespaceCodes();
	        }

            // apply any aliases required to create the list of output namespaces

            XSLStyleSheet sheet = getPrincipalStyleSheet();

            if (sheet.hasNamespaceAliases()) {
                for (int i=0; i<namespaceCodes.length; i++) {
                	// System.err.println("Examining namespace " + namespaceCodes[i]);
                	short scode = (short)(namespaceCodes[i]&0xffff);
                    int ncode = sheet.getNamespaceAlias(scode);
                    if (ncode != -1 && (ncode & 0xffff) != scode) {
                        // apply the namespace alias. Change in 7.3: use the prefix associated
                        // with the new namespace, not the old prefix.
                        namespaceCodes[i] = ncode;
                    }
                }

                // determine if there is an alias for the namespace of the element name

                int ercode = sheet.getNamespaceAlias(elementURICode);
                if ((ercode & 0xffff) != elementURICode) {
                	elementURICode = (short)(ercode & 0xffff);
                    resultNameCode = namePool.allocate(namePool.getPrefixFromNamespaceCode(ercode),
                                                       namePool.getURIFromNamespaceCode(ercode),
                                                       getLocalPart());
                }
            }

            // deal with special attributes

            String useAttSets = getAttributeValue(StandardNames.XSL_USE_ATTRIBUTE_SETS);
            if (useAttSets != null) {
                attributeSets = getAttributeSets(useAttSets, null);
            }

            String type = getAttributeValue(StandardNames.XSL_TYPE);
            if (type != null) {
                if (!getConfiguration().isSchemaAware()) {
                    compileError("The xsl:type attribute is available only with a schema-aware XSLT processor");
                }
                schemaType = getSchemaType(type);
            }

            String validate = getAttributeValue(StandardNames.XSL_VALIDATION);
            if (validate != null) {
                validation = Validation.getCode(validate);
                if (validation != Validation.STRIP && !getConfiguration().isSchemaAware()) {
                    compileError("To perform validation, a schema-aware XSLT processor is needed");
                }
                if (validation == Validation.INVALID) {
                    compileError("Invalid value for xsl:validation. " +
                                 "Permitted values are (strict, lax, preserve, strip)");
                }
            } else {
                validation = getContainingStyleSheet().getDefaultValidation();
            }

            // establish the names to be used for all the output attributes;
            // also type-check the AVT expressions

            short attributeURIs[] = new short[numberOfAttributes];
            if (numberOfAttributes > 0) {

                for (int i=0; i<numberOfAttributes; i++) {

                    int anameCode = attributeNames[i];
                    int alias = anameCode;
                    short attURIcode = namePool.getURICode(anameCode);

                    if (attURIcode!=0) {	// attribute has a namespace prefix
                        int newNSCode = sheet.getNamespaceAlias(attURIcode);
                        if ((newNSCode & 0xffff) != attURIcode) {
                	        attURIcode = (short)(newNSCode & 0xffff);
                            alias = namePool.allocate( namePool.getPrefixFromNamespaceCode(newNSCode),
                                                       namePool.getURIFromNamespaceCode(newNSCode),
                                                       attributeList.getLocalName(i));
                        }
                    }

    	            attributeNames[i] = translate(alias);
    	            attributeURIs[i] = attURIcode;
  	                attributeValues[i] = typeCheck(namePool.getDisplayName(alias), attributeValues[i]);
                }
            }

            // remove any namespaces that are on the exclude-result-prefixes list, unless it is
            // the namespace of the element or an attribute

            int numberExcluded = 0;
            for (int n=0; n<namespaceCodes.length; n++) {
            	short uricode = (short)(namespaceCodes[n] & 0xffff);
                if (isExcludedNamespace(uricode) && !sheet.isAliasResultNamespace(uricode)) {
                    // exclude it from the output namespace list
                    namespaceCodes[n] = -1;
                    numberExcluded++;
                }
            }

            // Now translate the list of namespace codes to use the target namepool,
            // getting rid of the -1 entries while we go.

            int count = namespaceCodes.length - numberExcluded;
            NamePool oldPool = getNamePool();
            NamePool newPool = getTargetNamePool();
            int[] newNamespaceCodes = new int[count];
            count = 0;
            for (int i=0; i<namespaceCodes.length; i++) {
                if (namespaceCodes[i] != -1) {
                    String prefix = oldPool.getPrefixFromNamespaceCode(namespaceCodes[i]);
                    String uri = oldPool.getURIFromNamespaceCode(namespaceCodes[i]);
                    newNamespaceCodes[count++] = newPool.allocateNamespaceCode(prefix, uri);
                }
            }
            namespaceCodes = newNamespaceCodes;
        }
    }

    /**
    * Validate the children of this node, recursively. Overridden for top-level
    * data elements.
    */

    protected void validateChildren() throws TransformerConfigurationException {
        if (!toplevel) {
            super.validateChildren();
        }
    }

    /**
    * Translate a namecode in the stylesheet namepool to a namecode in the target namepool
    */

    private int translate(int oldNameCode) {
        NamePool oldPool = getNamePool();
        NamePool newPool = getTargetNamePool();
        String prefix = oldPool.getPrefix(oldNameCode);
        String uri = oldPool.getURI(oldNameCode);
        String localName = oldPool.getLocalName(oldNameCode);
        return newPool.allocate(prefix, uri, localName);
    }

	/**
	* Process the literal result element by copying it to the result tree
	*/

    public Instruction compile() throws TransformerConfigurationException {
        // top level elements in the stylesheet are ignored
        if (toplevel) return null;

//        int annotation = getTypeAnnotation(schemaType);

        Instruction inst =
                new FixedElement(
                        translate(resultNameCode),
                        namespaceCodes,
                        attributeSets,
                        schemaType,
                        validation);

        compileChildren(inst);

        Instruction[] atts = null;
        if (numberOfAttributes > 0) {
            atts = new Instruction[attributeNames.length];
            XSLStyleSheet top = getPrincipalStyleSheet();
            for (int i=0; i<atts.length; i++) {
                FixedAttribute att = new FixedAttribute(
                        attributeNames[i],
                        Validation.STRIP,
                        null,
                        -1);
                att.setSelect(attributeValues[i]);
                att.setSourceLocation(top.putModuleNumber(getSystemId()),
                                                    getLineNumber());
                if (attributeChecked[i]) {
                    att.setNoSpecialChars();
                }
                atts[i] = att;
            }

            Instruction[] directChildren = inst.getChildren();
            Instruction[] allChildren = new Instruction[atts.length + directChildren.length];
            System.arraycopy(atts, 0, allChildren, 0, atts.length);
            System.arraycopy(directChildren, 0, allChildren, atts.length, directChildren.length);
            inst.setChildren(allChildren);
        }
        return inst;
    }

    /**
    * Make a top-level literal result element into a stylesheet. This implements
    * the "Literal Result Element As Stylesheet" facility.
    */

    public DocumentImpl makeStyleSheet(PreparedStyleSheet pss,
                                       StyleNodeFactory nodeFactory)
            throws TransformerConfigurationException {

        // the implementation grafts the LRE node onto a containing xsl:template and
        // xsl:stylesheet

		NamePool pool = getNamePool();
        String xslPrefix = getPrefixForURI(NamespaceConstant.XSLT);
        if (xslPrefix==null) {
            String message;
            if (getLocalPart().equals("stylesheet") || getLocalPart().equals("transform")) {
                if (getPrefixForURI(NamespaceConstant.MICROSOFT_XSL)!=null) {
                    message = "Saxon is not able to process Microsoft's WD-xsl dialect";
                } else {
                    message = "Namespace for stylesheet element should be " + NamespaceConstant.XSLT;
                }
            } else {
                message = "The supplied file does not appear to be a stylesheet";
            }
            TransformerConfigurationException err =
               new TransformerConfigurationException (message);
            try {pss.reportError(err);} catch(TransformerException err2) {}
            throw err;

        }

        // check there is an xsl:version attribute (it's mandatory), and copy
        // it to the new xsl:stylesheet element

        String version = getAttributeValue(StandardNames.XSL_VERSION);
        if (version==null) {
            TransformerConfigurationException err =
               new TransformerConfigurationException (
                "Literal Result Element As Stylesheet: xsl:version attribute is missing");
            try {pss.reportError(err);} catch(TransformerException err2) {}
            throw err;
        }

        try {
            TreeBuilder builder = new TreeBuilder();
            builder.setDocumentLocator(null);
            builder.setNamePool(pool);
            builder.setNodeFactory(nodeFactory);
            builder.setSystemId(this.getSystemId());

            builder.startDocument();

            int st = StandardNames.XSL_STYLESHEET;
            builder.startElement(st, 0, 0);
            builder.namespace(NamespaceConstant.XSLT_CODE, 0);
            builder.attribute(pool.allocate("", "", "version"), 0, version, 0);
            builder.startContent();

            int te = StandardNames.XSL_TEMPLATE;
            builder.startElement(te, 0, 0);
            builder.attribute(pool.allocate("", "", "match"), 0, "/", 0);
            builder.startContent();

            builder.graftElement(this);

            builder.endElement();
            builder.endElement();
            builder.endDocument();

            return (DocumentImpl)builder.getCurrentDocument();
        } catch (TransformerException err) {
            throw new TransformerConfigurationException(err);
        }

    }


}
//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
