package net.sf.saxon.style;
import net.sf.saxon.expr.ContextItemExpression;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.TypeChecker;
import net.sf.saxon.expr.RoleLocator;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.sort.CodepointCollator;
import net.sf.saxon.sort.SortKeyDefinition;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.value.EmptySequence;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.type.Type;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerConfigurationException;
import java.text.Collator;
import java.util.Comparator;
import java.util.Locale;

/**
* An xsl:sort element in the stylesheet. <br>
*/

public class XSLSort extends StyleElement {

    private SortKeyDefinition sortKeyDefinition;
    private Expression select;
    private Expression order;
    private Expression dataType = null;
    private Expression caseOrder;
    private Expression lang;
    private String collationName;

    public void prepareAttributes() throws TransformerConfigurationException {

		AttributeCollection atts = getAttributeList();

		String selectAtt = null;
        String orderAtt = null;
        String dataTypeAtt = null;
        String caseOrderAtt = null;
        String langAtt = null;
        String collationAtt = null;

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			String f = getNamePool().getClarkName(nc);
			if (f==StandardNames.SELECT) {
        		selectAtt = atts.getValue(a);
        	} else if (f==StandardNames.ORDER) {
        		orderAtt = atts.getValue(a).trim();
        	} else if (f==StandardNames.DATA_TYPE) {
        		dataTypeAtt = atts.getValue(a).trim();
        	} else if (f==StandardNames.CASE_ORDER) {
        		caseOrderAtt = atts.getValue(a).trim();
        	} else if (f==StandardNames.LANG) {
        		langAtt = atts.getValue(a).trim();
        	} else if (f==StandardNames.COLLATION) {
        		collationAtt = atts.getValue(a).trim();
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (selectAtt==null) {
            select = new ContextItemExpression();
        } else {
            select = makeExpression(selectAtt);
        }

        if (orderAtt == null) {
            order = new StringValue("ascending");
        } else {
            order = makeAttributeValueTemplate(orderAtt);
        }

        if (dataTypeAtt == null) {
            dataType = EmptySequence.getInstance();
        } else {
            dataType = makeAttributeValueTemplate(dataTypeAtt);
        }

        if (caseOrderAtt == null) {
            caseOrder = new StringValue("#default");
        } else {
            caseOrder = makeAttributeValueTemplate(caseOrderAtt);
        }

        if (langAtt == null) {
            lang = new StringValue(Locale.getDefault().getLanguage());
        } else {
            lang = makeAttributeValueTemplate(langAtt);
        }

        collationName = collationAtt;
        // TODO: allow the collation to be specified as an Attribute Value Template


    }

    public void validate() throws TransformerConfigurationException {
        checkEmpty();
        NodeInfo parent = (NodeInfo)getParentNode();
        if (!((parent instanceof XSLApplyTemplates) ||
                 (parent instanceof XSLForEach)
                 || (parent instanceof XSLForEachGroup)
                 || (parent instanceof XSLPerformSort)
//                 || (parent instanceof XSLSortKey)
                 )) {
            compileError("xsl:sort must be child of xsl:apply-templates, xsl:for-each[-group], or xsl:perform-sort");
        }

        // Get the named or default collation


        Comparator collator = null;
        if (collationName != null) {
            if (collationName.equals(CodepointCollator.URI)) {
                collator = CodepointCollator.getInstance();
            } else {
                collator = getPrincipalStyleSheet().findCollation(collationName);
            }
            if (collator==null) {
                compileError("No collation of this name has been defined");
                collator = Collator.getInstance();     // for recovery paths
            }
        }

        select      = typeCheck("select", select);
        order       = typeCheck("order", order);
        caseOrder   = typeCheck("case-order", caseOrder);
        lang        = typeCheck("lang", lang);
        dataType    = typeCheck("data-type", dataType);

        try {
            RoleLocator role =
                new RoleLocator(RoleLocator.INSTRUCTION, "xsl:sort/select", 0);
            select = TypeChecker.staticTypeCheck(select,
                            SequenceType.ATOMIC_SEQUENCE,
                            false, role);
        } catch (XPathException err) {
            compileError(err);
        }

        sortKeyDefinition = new SortKeyDefinition();
        sortKeyDefinition.setOrder(order);
        sortKeyDefinition.setCaseOrder(caseOrder);
        sortKeyDefinition.setLanguage(lang);
        sortKeyDefinition.setSortKey(select);
        sortKeyDefinition.setDataTypeExpression(dataType);
        sortKeyDefinition.setCollation(collator);


        try {
            // Simplify the sort key definition - this is especially important in the case where
            // all aspects of the sort key are known statically.
            sortKeyDefinition = sortKeyDefinition.simplify();
            //sortKeyDefinition.analyze(staticContext);
        } catch (XPathException err) {
            throw new TransformerConfigurationException(err);
        }
    }

    public Instruction compile() throws TransformerConfigurationException {
        // not an executable instruction
        return null;
    }

    public SortKeyDefinition getSortKeyDefinition() {
        return sortKeyDefinition;
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
