package net.sf.saxon.trans;
import net.sf.saxon.Controller;
import net.sf.saxon.style.StandardNames;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.ContextItemExpression;
import net.sf.saxon.expr.Atomizer;
import net.sf.saxon.om.*;
import net.sf.saxon.pattern.NodeTest;
import net.sf.saxon.pattern.Pattern;
import net.sf.saxon.pattern.ContentTypeTest;
import net.sf.saxon.pattern.UnionPattern;
import net.sf.saxon.sort.LocalOrderComparer;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.DoubleValue;
import net.sf.saxon.type.Type;
import net.sf.saxon.type.SchemaType;
import net.sf.saxon.type.BuiltInSchemaFactory;
import net.sf.saxon.xpath.XPathException;

import java.io.Serializable;
import java.lang.ref.WeakReference;
import java.text.Collator;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.WeakHashMap;
import java.util.List;

/**
  * KeyManager manages the set of key definitions in a stylesheet, and the indexes
  * associated with these key definitions. It handles xsl:sort-key as well as xsl:key
  * definitions.
  *
  * <p>The memory management in this class is subtle, with extensive use of weak references.
  * The idea is that an index should continue to exist in memory so long as both the compiled
  * stylesheet and the source document exist in memory: if either is removed, the index should
  * go too. The document itself holds no reference to the index. The compiled stylesheet (which
  * owns the KeyManager) holds a weak reference to the index. The index, of course, holds strong
  * references to the nodes in the document. The Controller holds a strong reference to the
  * list of indexes used for each document, so that indexes remain in memory for the duration
  * of a transformation even if the documents themselves are garbage collected.</p>
  *
  * <p>Potentially there is a need for more than one index for a given key name, depending
  * on the primitive type of the value provided to the key() function. An index is built
  * corresponding to the type of the requested value; if subsequently the key() function is
  * called with the same name and a different type of value, then a new index is built.</p>
  *
  * @author Michael H. Kay (michael.h.kay@ntlworld.com)
  */

public class KeyManager implements Serializable {

    private HashMap keyList;         // one entry for each named key; the entry contains
                                     // a list of key definitions with that name
    private transient WeakHashMap docIndexes;
                                     // one entry for each document that is in memory;
                                     // the entry contains a HashMap mapping the fingerprint of
                                     // the key name plus the primitive item type
                                     // to the HashMap that is the actual index
                                     // of key/value pairs.

    /**
    * create a KeyManager and initialise variables
    */

    public KeyManager() {
        keyList = new HashMap();
        docIndexes = new WeakHashMap();
        // Create a key definition for the idref() function
        registerIdrefKey();
    }

    /**
     *
     */

    private void registerIdrefKey() {
        SchemaType idref = BuiltInSchemaFactory.getSchemaType(StandardNames.XS_IDREF);
        SchemaType idrefs = BuiltInSchemaFactory.getSchemaType(StandardNames.XS_IDREFS);
        Pattern idrefAtt = new ContentTypeTest(Type.ATTRIBUTE, idref);
        Pattern idrefsAtt = new ContentTypeTest(Type.ATTRIBUTE, idrefs);
        Pattern idrefElem = new ContentTypeTest(Type.ELEMENT, idref);
        Pattern idrefsElem = new ContentTypeTest(Type.ELEMENT, idrefs);
        Pattern att = new UnionPattern(idrefAtt, idrefsAtt);
        Pattern elem = new UnionPattern(idrefElem, idrefsElem);
        Pattern all = new UnionPattern(att, elem);
        Expression use = new Atomizer(new ContextItemExpression());
        KeyDefinition key = new KeyDefinition(all, use, null);
        setKeyDefinition(StandardNames.XS_IDREFS, key);
    }

    /**
    * Register a key definition. Note that multiple key definitions with the same name are
    * allowed
    * @param fingerprint Integer representing the name of the key
    * @param keydef The details of the key's definition
    */

    public void setKeyDefinition(int fingerprint, KeyDefinition keydef) {
        Integer keykey = new Integer(fingerprint);
        ArrayList v = (ArrayList)keyList.get(keykey);
        if (v==null) {
            v = new ArrayList();
            keyList.put(keykey, v);
        }
        v.add(keydef);
    }

    /**
    * Get all the key definitions that match a particular fingerprint
    * @param fingerprint The fingerprint of the name of the required key
    * @return The key definition of the named key if there is one, or null otherwise.
    */

    public List getKeyDefinitions(int fingerprint) {
        return (List)keyList.get(new Integer(fingerprint));
    }

    /**
    * Build the index for a particular document for a named key
    * @param fingerprint The fingerprint of the name of the required key
    * @param doc The source document in question
    * @param controller The controller
    * @return the index in question, as a HashMap mapping a key value onto a ArrayList of nodes
    */

    private synchronized HashMap buildIndex(int fingerprint,
                                            int itemType,
                                           DocumentInfo doc,
                                           Controller controller) throws XPathException {

        List definitions = getKeyDefinitions(fingerprint);
        if (definitions==null) {
            throw new XPathException.Dynamic("Key " +
            		controller.getNamePool().getDisplayName(fingerprint) +
            							" has not been defined");
        }

        HashMap index = new HashMap();

        for (int k=0; k<definitions.size(); k++) {
            constructIndex( doc,
                            index,
                            (KeyDefinition)definitions.get(k),
                            itemType,
                            controller,
                            k==0);
        }

        return index;

    }

    /**
    * Process one key definition to add entries to an index
    */

    private void constructIndex(    DocumentInfo doc,
                                    HashMap index,
                                    KeyDefinition keydef,
                                    int itemType,
                                    Controller controller,
                                    boolean isFirst) throws XPathException {

        Pattern match = keydef.getMatch();
        Expression use = keydef.getUse();
        Collator collator = keydef.getCollation();
        NodeInfo sourceRoot = doc;
        NodeInfo curr = sourceRoot;
        SequenceIterator save = controller.getCurrentIterator();
        Object[] savedContext = controller.saveContext();
        XPathContext xc = controller.newXPathContext();

        int nodeType = match.getNodeKind();

        if (nodeType==Type.ATTRIBUTE || nodeType==Type.NODE || nodeType==Type.DOCUMENT) {
            // If the match pattern allows attributes to appear, we must visit them.
            // We also take this path in the ridiculous event that the pattern can match
            // document nodes.
            SequenceIterator all = doc.iterateAxis(Axis.DESCENDANT_OR_SELF);
            while(true) {
                curr = (NodeInfo)all.next();
                if (curr==null) {
                    break;
                }
                if (curr.getNodeKind()==Type.ELEMENT) {
                    SequenceIterator atts =
                        curr.iterateAxis(Axis.ATTRIBUTE);
                    while (true) {
                        NodeInfo att = (NodeInfo)atts.next();
                        if (att == null) {
                            break;
                        }
                        if (match.matches(att, controller)) {
                            processKeyNode(att, use, itemType,
                                            collator, index, xc, isFirst);
                        }
                    }
                    if (nodeType==Type.NODE) {
                        // index the element as well as its attributes
                        if (match.matches(curr, controller)) {
                            processKeyNode(curr, use, itemType,
                                        collator, index, xc, isFirst);
                        }
                    }
                } else {
                    if (match.matches(curr, controller)) {
                        processKeyNode(curr, use, itemType,
                                     collator, index, xc, isFirst);
                    }
                }
            }

        } else {
            SequenceIterator all =
                doc.iterateAxis( Axis.DESCENDANT,
                                    match.getNodeTest());
            // If the match is a nodetest, we avoid testing it again
            while(true) {
                curr = (NodeInfo)all.next();
                if (curr == null) {
                    break;
                }
                if (match instanceof NodeTest || match.matches(curr, controller)) {
                    processKeyNode(curr, use, itemType,
                                collator, index, xc, isFirst);
                }
            }
        }
        // reset the Controller's context
        controller.setCurrentIterator(save);
        controller.restoreContext(savedContext);
    }

    /**
    * Process one matching node, adding entries to the index if appropriate
    */

    private void processKeyNode(    NodeInfo curr,
                                    Expression use,
                                    int itemType,
                                    Collator collation,
                                    HashMap index,
                                    XPathContext xc,
                                    boolean isFirst) throws XPathException {


        // Make the node we are testing the context node and the current node,
        // with context position and context size set to 1

        AxisIterator si = SingletonIterator.makeIterator(curr);
        si.next();    // need to position iterator at first node

        xc.setCurrentIterator(si);
        xc.getController().setCurrentIterator(si);

        // Evaluate the "use" expression against this context node

        SequenceIterator useval = use.iterate(xc);
        while (true) {
            AtomicValue item = (AtomicValue)useval.next();
            if (item == null) {
                break;
            }
            Object val;

            if (itemType==Type.STRING || itemType==Type.UNTYPED_ATOMIC) {
                if (collation==null) {
                    val = item.getStringValue();
                } else {
                    val = collation.getCollationKey(item.getStringValue());
                }
//            } else if (itemType==Type.FLOAT && Float.isNaN(((FloatValue)item).getValue())) {
//                val = "NaN";
//            } else if (itemType==Type.DOUBLE && Double.isNaN(((DoubleValue)item).getValue())) {
//                val = "NaN";
            } else {
                try {
                    val = item.convert(itemType);
                } catch (XPathException err) {
                    // ignore values that can't be converted to the required type
                    break;
                }
            }

            // Ignore NaN values
            if (item instanceof DoubleValue && Double.isNaN(((DoubleValue)item).getValue())) {
                break;
            }

            ArrayList nodes = (ArrayList)index.get(val);
            if (nodes==null) {
                // this is the first node with this key value
                nodes = new ArrayList();
                index.put(val, nodes);
                nodes.add(curr);
            } else {
                // this is not the first node with this key value.
                // add the node to the list of nodes for this key,
                // unless it's already there
                if (isFirst) {
                    // if this is the first index definition that we're processing,
                    // then this node must be after all existing nodes in document
                    // order, or the same node as the last existing node
                    if (nodes.get(nodes.size()-1)!=curr) {
                        nodes.add(curr);
                    }
                } else {
                    // otherwise, we need to insert the node at the correct
                    // position in document order.
                    LocalOrderComparer comparer = LocalOrderComparer.getInstance();
                    for (int i=0; i<nodes.size(); i++) {
                        int d = comparer.compare(curr, (NodeInfo)nodes.get(i));
                        if (d<=0) {
                            if (d==0) {
                                // node already in list; do nothing
                            } else {
                                // add the node at this position
                                nodes.add(i, curr);
                            }
                            return;
                        }
                        // else continue round the loop
                    }
                    // if we're still here, add the new node at the end
                    nodes.add(curr);
                }
            }
        }

    }

    /**
    * Get the nodes with a given key value
    * @param fingerprint The fingerprint of the name of the required key
    * @param doc The source document in question
    * @param value The required key value
    * @param controller The controller, needed only the first time when the key is being built
    * @return an enumeration of nodes, always in document order
    */

    public SequenceIterator selectByKey(
                                int fingerprint,
                                DocumentInfo doc,
                                AtomicValue value,
                                Controller controller) throws XPathException {

        int itemType = value.getItemType().getPrimitiveType();
        if (itemType == StandardNames.XS_INTEGER ||
                itemType == StandardNames.XS_DECIMAL ||
                itemType == StandardNames.XS_FLOAT) {
            itemType = StandardNames.XS_DOUBLE;
            value = value.convert(itemType);
        }

        Object indexObject = getIndex(doc, fingerprint, itemType);
        if (indexObject instanceof String) {
            // index is under construction
            throw new XPathException.Dynamic("Key definition is circular");
        }
        HashMap index = (HashMap)indexObject;
        if (index==null) {
            putIndex(doc, fingerprint, itemType, "Under Construction", controller);
            index = buildIndex(fingerprint, itemType, doc, controller);
            putIndex(doc, fingerprint, itemType, index, controller);
        }

        KeyDefinition definition = (KeyDefinition)getKeyDefinitions(fingerprint).get(0);
               // the itemType and collation will be the same for all keys with the same name
        Collator collation = definition.getCollation();

        Object val;
        if (itemType==Type.STRING || itemType==Type.UNTYPED_ATOMIC) {
            if (collation==null) {
                val = value.getStringValue();
            } else {
                val = collation.getCollationKey(value.getStringValue());
            }
//        } else if (itemType==Type.FLOAT || itemType==Type.DOUBLE ) {
//            // special case: ensure NaN=NaN for indexing purposes
//            val = value.convert(itemType);
//            if (Double.isNaN(((NumericValue)val).asDouble())) {
//                val = "NaN";
//            }
//        } else {
//            val = value.convert(itemType);

        } else {
            val = value;
        }

        ArrayList nodes = (ArrayList)index.get(val);
        if (nodes==null) {
            return EmptyIterator.getInstance();
        } else {
            return new ArrayListIterator(nodes);
        }
    }

    /**
    * Save the index associated with a particular key, a particular item type,
    * and a particular document. This
    * needs to be done in such a way that the index is discarded by the garbage collector
    * if the document is discarded. We therefore use a WeakHashMap indexed on the DocumentInfo,
    * which returns HashMap giving the index for each key fingerprint. This index is itself another
    * HashMap.
    * The methods need to be synchronized because several concurrent transformations (which share
    * the same KeyManager) may be creating indexes for the same or different documents at the same
    * time.
    */

    private synchronized void putIndex(DocumentInfo doc, int keyFingerprint, int itemType, Object index, Controller controller) {
        if (docIndexes==null) {
            // it's transient, so it will be null when reloading a compiled stylesheet
            docIndexes = new WeakHashMap();
        }
        WeakReference indexRef = (WeakReference)docIndexes.get(doc);
        HashMap indexList;
        if (indexRef==null || indexRef.get()==null) {
            indexList = new HashMap();
            // ensure there is a firm reference to the indexList for the duration of a transformation
            controller.setUserData(doc, "key-index-list", indexList);
            docIndexes.put(doc, new WeakReference(indexList));
        } else {
            indexList = (HashMap)indexRef.get();
        }
        indexList.put(new Long(((long)keyFingerprint)<<32 | itemType), index);
    }

    /**
    * Get the index associated with a particular key, a particular source document,
     * and a particular primitive item type
    */

    private synchronized Object getIndex(DocumentInfo doc, int keyFingerprint, int itemType) {
        if (docIndexes==null) {
            // it's transient, so it will be null when reloading a compiled stylesheet
            docIndexes = new WeakHashMap();
        }
        WeakReference ref = (WeakReference)docIndexes.get(doc);
        if (ref==null) return null;
        HashMap indexList = (HashMap)ref.get();
        if (indexList==null) return null;
        return indexList.get(new Long(((long)keyFingerprint)<<32 | itemType));
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
