/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio, Inc.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio, Inc. Exolab is a registered
 *    trademark of Intalio, Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * INTALIO, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999-2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: AtomicType.java,v 1.1 2004/03/09 10:10:35 mhkay Exp $
 */

package net.sf.saxon.type;

import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.SingletonIterator;
import net.sf.saxon.om.NamespaceResolver;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.style.StandardNames;

/**
 * An object belonging to this class represents an atomic type: either a built-in
 * atomic type, or a user-defined atomic type. There is at most one AtomicType
 * object for each distinct type in the schema (so equality can safely be tested
 * using the == operator).
 * @author <a href="mailto:berry@intalio.com">Arnaud Berry</a>
 * @version $Revision:
**/
public class AtomicType extends SimpleType implements ItemType
{

    private boolean builtIn = false;

    public AtomicType() {}

    /**
     * Mark this as a built-in type
     */

    public void setIsBuiltIn(boolean yesOrNo) {
        builtIn = yesOrNo;
    }

    /**
     * Test whether this is a built-in type
     */

    public boolean isBuiltIn() {
        return builtIn;
    }

    /**
     * Test whether a given item conforms to this type
     * @param item The item to be tested
     * @return true if the item is an instance of this type; false otherwise
    */

    public boolean matchesItem(Item item) {
        // MHK addition
        if (item instanceof AtomicValue) {
            AtomicValue value = (AtomicValue)item;
            AtomicType type = (AtomicType)value.getItemType();
            if (type.getFingerprint()==this.getFingerprint()) {
                // note, with compiled stylesheets one can have two objects representing
                // the same type, so comparing identity is not safe
                return true;
            }
            return Type.isSubType(type, this);
        } else {
            return false;
        }
    }

    /**
     * Check whether a given input string is valid according to this SimpleType
     * @throws ValidationException if the value is not valid
     */

    public void validateContent(CharSequence value, NamespaceResolver namespaceResolver)
    throws ValidationException {
        int f = getFingerprint();
        if (f==StandardNames.XS_STRING ||
                f==StandardNames.XS_ANY_SIMPLE_TYPE ||
                f==StandardNames.XDT_UNTYPED_ATOMIC) {
            return;
        }
        if (isNamespaceSensitive()) {
            if (namespaceResolver == null) {
                throw new ValidationException("Cannot validate a QName without a namespace context");
            }
            try {
                new StringValue(value).convert(this, namespaceResolver);
            } catch (XPathException err) {
                throw new ValidationException(err.getMessage());
            }
        }
        try {
            new StringValue(value).convert(this);
        } catch (XPathException err) {
            throw new ValidationException(err.getMessage());
        }
    }

    /**
     * Get the typed value corresponding to a given string value, assuming it is
     * valid against this type
     * @param value the string value
     * @return an iterator over the atomic sequence comprising the typed value
     */

    public SequenceIterator getTypedValue(CharSequence value)
            throws ValidationException {
        try {
            return SingletonIterator.makeIterator(new StringValue(value).convert(this));
        } catch (XPathException err) {
            throw new ValidationException(err.getMessage());
        }
    }

    /**
     * Get the type from which this item type is derived by restriction. This
     * is the supertype in the XPath type heirarchy, as distinct from the Schema
     * base type: this means that the supertype of xs:boolean is xdt:anyAtomicType,
     * whose supertype is item() (rather than xs:anySimpleType).
     * @return the supertype, or null if this type is item()
     */

    public ItemType getSuperType() {
        if (this == Type.ATOMIC_TYPE) {
            return AnyItemType.getInstance();
        } else {
            return (ItemType)getBaseType();
        }
    }

    /**
     * Get the primitive type from which this type is derived. For the definition
     * of primitive types, see {@link Type#isPrimitiveType}
     * @return the type code of the primitive type
     */
    public int getPrimitiveType() {
        int x = getFingerprint();
        if (Type.isPrimitiveType(x)) {
            return x;
        } else {
            ItemType s = getSuperType();
            if (s instanceof AtomicType) {
                return s.getPrimitiveType();
            } else {
                return this.getFingerprint();
            }
        }
    }

    /**
     * Produce string representation for use in diagnostic output
     */

    public String toString() {
        int f = getFingerprint();
        if (f < 0) {
            throw new IllegalArgumentException("fingerprint < 0");
        }
        if (f < 1023) {
            return StandardNames.getDisplayName(f);
        } else {
            return getNamePool().getDisplayName(f);
        }
    }

    /**
     * Gets the pattern facet
     * returns null if there is none
    **/
    public String getPattern()
    {
        Facet patternFacet= getFacet(Facet.PATTERN);
        if (patternFacet == null) return null;
        return patternFacet.getValue();
    }

    /**
     *  Tells if the minimum is inclusive or not
     *  Has no meaning if there is no minInclusive/Exclusive facet
    **/
    public boolean isMinInclusive()
    {
        return (getFacet(Facet.MIN_INCLUSIVE) != null);
    }

    /**
     *  Tells if the maximum is inclusive or not
     *  Has no meaning if there is no maxInclusive/Exclusive facet
    **/
    public boolean isMaxInclusive()
    {
        return (getFacet(Facet.MAX_INCLUSIVE) != null);
    }

    /**
     * Gets the MIN_INCLUSIVE facet or the MIN_EXCLUSIVE facet
     * (both are never defined at the same time)
     * returns null if there is no minimum
    **/
    public Facet getMin()
    {
        Facet minInclusiveFacet= getFacet(Facet.MIN_INCLUSIVE);
        if (minInclusiveFacet != null) return minInclusiveFacet;

        Facet minExclusiveFacet= getFacet(Facet.MIN_EXCLUSIVE);
        if (minExclusiveFacet != null) return minExclusiveFacet;

        return null;
    }

    /**
     * Gets the MAX_INCLUSIVE facet or the MAX_EXCLUSIVE facet
     * (both are never defined at the same time)
     * returns null if there is no maximum
    **/
    public Facet getMax()
    {
        Facet maxInclusiveFacet= getFacet(Facet.MAX_INCLUSIVE);
        if (maxInclusiveFacet != null) return maxInclusiveFacet;

        Facet maxExclusiveFacet= getFacet(Facet.MAX_EXCLUSIVE);
        if (maxExclusiveFacet != null) return maxExclusiveFacet;

        return null;
    }

    public boolean isSameType(ItemType other) {
        return (other instanceof AtomicType) &&
                ((AtomicType)other).getFingerprint() == this.getFingerprint();
    }
}




