/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio, Inc.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio, Inc. Exolab is a registered
 *    trademark of Intalio, Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * INTALIO, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Intalio, Inc. All Rights Reserved.
 *
 * $Id: Facet.java,v 1.1 2004/03/09 10:10:35 mhkay Exp $
 */

package net.sf.saxon.type;

import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.Tokenizer;
import net.sf.saxon.om.NamespaceConstant;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.xpath.StandaloneContext;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.Err;

import javax.xml.transform.ErrorListener;
import java.io.Serializable;
import java.util.regex.PatternSyntaxException;

/**
 * Represents the base type for XML Schema Facets
 * @author <a href="mailto:kvisco@intalio.com">Keith Visco</a>
 * @version $Revision: 1.1 $ $Date: 2004/03/09 10:10:35 $
**/
public class Facet implements Serializable {

    public static final String ENUMERATION      = "enumeration";
    public static final String LENGTH           = "length";
    public static final String PATTERN          = "pattern";
    public static final String MAX_EXCLUSIVE    = "maxExclusive";
    public static final String MAX_INCLUSIVE    = "maxInclusive";
    public static final String MIN_EXCLUSIVE    = "minExclusive";
    public static final String MIN_INCLUSIVE    = "minInclusive";
    public static final String MAX_LENGTH       = "maxLength";
    public static final String MIN_LENGTH       = "minLength";
    public static final String WHITESPACE       = "whiteSpace";
    public static final String TOTALDIGITS      = "totalDigits";
    public static final String FRACTIONDIGITS   = "fractionDigits";

    /**
     * The name of this Facet
    **/
    private String _name    = null;

    /**
     * The character value of this Facet
    **/
    private String  _value  = null;

    /**
     * XPath expression that tests the context item to determine if it conforms
     * to the facet: result is true or false
     */

    private Expression _test = null;

    protected Facet() {}

    /**
     * Factory method: creates a new Facet with the given name
     * @param name the name of the Facet
     * @param value the value of the Facet
    **/
    public static Facet makeFacet(String name, String value) throws SchemaException {
        //assert name != null;
        //assert name.length() > 0;
        if (name.equals(PATTERN)) {
            PatternFacet lf = new PatternFacet();
            lf.setValue(value);
            return lf;
        } else if (name.equals(LENGTH)) {
            LengthFacet lf = new LengthFacet();
            lf.setValue(value);
            return lf;
        } else if (name.equals(MIN_LENGTH)) {
            MinLengthFacet lf = new MinLengthFacet();
            lf.setValue(value);
            return lf;
        } else if (name.equals(MAX_LENGTH)) {
            MaxLengthFacet lf = new MaxLengthFacet();
            lf.setValue(value);
            return lf;
        } else if (name.equals(MIN_INCLUSIVE)) {
            ValueRangeFacet vrf = new ValueRangeFacet(ValueRangeFacet.MIN_INCLUSIVE);
            vrf.setValue(value);
            return vrf;
        } else if (name.equals(MAX_INCLUSIVE)) {
            ValueRangeFacet vrf = new ValueRangeFacet(ValueRangeFacet.MAX_INCLUSIVE);
            vrf.setValue(value);
            return vrf;
        } else if (name.equals(MIN_EXCLUSIVE)) {
            ValueRangeFacet vrf = new ValueRangeFacet(ValueRangeFacet.MIN_EXCLUSIVE);
            vrf.setValue(value);
            return vrf;
        } else if (name.equals(MAX_EXCLUSIVE)) {
            ValueRangeFacet vrf = new ValueRangeFacet(ValueRangeFacet.MAX_EXCLUSIVE);
            vrf.setValue(value);
            return vrf;
        } else {
            // TODO: use subclasses of Facet for each kind of facet.
            // TODO: merge enumeration elements into a single facet.
            Facet f = new Facet();
            f._name = name;
            f._value= value;
            try {
                f._test = f.compileFacet();
            } catch (XPathException err) {
                throw new AssertionError("Invalid expression for facet: " + err.getMessage());
            }
            return f;
        }
        // TODO: in general, facets are not being properly checked. There is a need for more
        // local checks (we now do some on length, minLength and maxLength), for cross-checks
        // between different facets of the same type, and for cross-checks between the facets
        // of a type and the facets of its base type.
    }

    /**
     * Compile an XPath expression that can be used to check whether a given
     * value conforms to the facet. Restricted to AtomicValues currently.
     */

    private Expression compileFacet() throws XPathException, SchemaException {
        StandaloneContext env = new StandaloneContext();
        env.declareNamespace("xdt", NamespaceConstant.XDT);
        Expression exp;
        // Replace any apostrophes in the value by a pair of apostrophes
        String eval = escapeQuotes(_value);
        // Generate code to treat the supplied value of the facet as untyped
        // atomic. This causes it to be compared according to the type of the
        // other operand.
        String sval = "xdt:untypedAtomic('" + eval + "')";
        if (_name.equals("enumeration")) {
            exp = ExpressionTool.make(". eq " + sval, env, 0, Tokenizer.EOF);
            // TODO: consider effect of collations on the enumeration facet
//        } else if (_name.equals("length")) {
//            exp = ExpressionTool.make("string-length(.) eq " + sval, env);
//        } else if (_name.equals("pattern")) {
//            try {
//                java.util.regex.Pattern.compile(_value);
//            } catch (PatternSyntaxException err) {
//                throw new SchemaException("Error in regex " + _value + ": " + err.getMessage());
//            }
//            exp = ExpressionTool.make("matches(., '^" + eval + "$')", env, 0, Tokenizer.EOF);
//        } else if (_name.equals("maxExclusive")) {
//            exp = ExpressionTool.make(". < " + sval, env);
//        } else if (_name.equals("maxInclusive")) {
//            exp = ExpressionTool.make(". <= " + sval, env);
//        } else if (_name.equals("minExclusive")) {
//            exp = ExpressionTool.make(". > " + sval, env);
//        } else if (_name.equals("minInclusive")) {
//            exp = ExpressionTool.make(". >= " + sval, env);
//        } else if (_name.equals("maxLength")) {
//            exp = ExpressionTool.make("string-length(.) le " + sval, env);
//        } else if (_name.equals("minLength")) {
//            exp = ExpressionTool.make("string-length(.) ge " + sval, env);
        } else if (_name.equals("whiteSpace")) {
            exp = ExpressionTool.make("true()", env, 0, Tokenizer.EOF);
        } else if (_name.equals("totalDigits")) {
            exp = ExpressionTool.make("string-length(translate(., '+-.', '')) le " + sval, env, 0, Tokenizer.EOF);
        } else if (_name.equals("fractionDigits")) {
            exp = ExpressionTool.make("if (contains(., '.')) " +
                                      "then string-length(substring-after(., '.')) le " + sval +
                                      " else true()", env, 0, Tokenizer.EOF);
        } else {
            exp = ExpressionTool.make("true()", env, 0, Tokenizer.EOF);
        }
        exp = exp.simplify();
        exp = exp.analyze(env);
        return exp;
    }

    /**
     * Escape single quotes in a value to a pair of single quotes
     */

    private static String escapeQuotes(String value) {
        int apos = value.indexOf('\'');
        if (apos < 0) {
            return value;
        } else {
            return value.substring(0, apos) + "''" +
                    (apos == value.length()-1 ?
                    "" :
                    escapeQuotes(value.substring(apos+1)));
        }
    }
    /**
     * Returns the name of this Facet
     * @return the name of this Facet
    **/
    public String getName() {
        return _name;
    } //-- getName


    /**
     * Returns the character (String) representation of this facet
     * @return the value of this facet
    **/
    public String getValue() {
        return this._value;
    } //-- getValue


    /**
     * Returns true if this Facet can occur more than once, such
     * as the "enumeration" facet.
     * @return true if this Facet can occur more than once.
    **/
    public boolean isMultivalued() {
        return _name.equals(Facet.ENUMERATION);
    }


    /**
     * Returns a long representation of the value of this facet
     * @return a long representation of the value of this facet
    **/
    public long toLong() throws NumberFormatException {
        return Long.parseLong(_value);
    } //-- toInt


    /**
     * Returns an short representation of the value of this facet
     * @return an short representation of the value of this facet
    **/
    public short toShort() throws NumberFormatException {
        return Short.parseShort(_value);
    } //-- toShort


    /**
     * Returns a double representation of the value of this facet
     * @return a double representation of the value of this facet
     */
    public float toFloat() throws NumberFormatException {
        if (_value.equals("INF")) return Float.POSITIVE_INFINITY;
        if (_value.equals("-INF")) return Float.NEGATIVE_INFINITY;
        return Float.valueOf(_value).floatValue();
    } //-- toInt

    /**
     * Returns a double representation of the value of this facet
     * @return a double representation of the value of this facet
    **/
    public double toDouble() throws NumberFormatException {
        return Double.valueOf(_value).doubleValue();
    } //-- toInt


    /**
     * Returns a byte representation of the value of this facet
     * @return a byte representation of the value of this facet
    **/
    public byte toByte() throws NumberFormatException {
        return Byte.parseByte(_value);
    } //-- toInt

    /**
     * Check that this facet is legal when used on a type derived by restriction
     * @param base the type from which the restricted type is derived
     * @throws SchemaException if the facet is not legal
     */

    public void checkFacetRestriction(SimpleType base) throws SchemaException {
        if (_name.equals(ENUMERATION)) {
            try {
                base.validateContent(_value, null);
            } catch (ValidationException err) {
                throw new SchemaException("The enumeration value " + Err.wrap(_value) +
                                          " is not valid for the base type " +
                                          base.getDescription());
            }
        }

    }

    /**
     * Checks the validity of this Schema defintion.
     * @exception net.sf.saxon.type.SchemaException when this Schema definition
     * is invalid.
    **/
    public boolean validate(ErrorListener errorListener) throws SchemaException
    {
        return true;
         //-- do nothing for now
    } //-- validate

    /**
     * Test whether an atomic value conforms to this facet
     * @param value  the value to be tested
     * @return true if the value conforms; false if it doesn't conform, or if the test
     * fails
     */

    public boolean testAtomicValue(AtomicValue value) {
        try {
            XPathContext c = new XPathContext(value);
            return _test.effectiveBooleanValue(c);
        } catch (XPathException err) {
            return false;
        }
    }

} //-- Facet

