package net.sf.saxon.type;

import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.type.Type;
import net.sf.saxon.type.Facet;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.Err;

/**
 * Represents the length facet. Sublclasses represent the minLength and maxLength facets.
 */
public class LengthFacet extends Facet {

    int requiredLength;

    public String getName() {
        return "length";
    }

    public String getValue() {
        return "" + requiredLength;
    }

    public long toLong() {
        return requiredLength;
    }

    public void setValue(String value) throws SchemaException {
        try {
            requiredLength = Integer.parseInt(value);
        } catch (NumberFormatException e) {
            throw new SchemaException("Value of " + getName() + " facet "
                                      + Err.wrap(value) + " is not an integer");
        }
        if (requiredLength < 0) {
            throw new SchemaException("Value of " + getName() + " facet "
                                      + Err.wrap(value) + " must be a positive integer");
        }
    }

    /**
     * Check that this facet is legal when used on a type derived by restriction
     * @param base the type from which the restricted type is derived
     * @throws SchemaException if the facet is not legal
     */

    public void checkFacetRestriction(SimpleType base) throws SchemaException {
        Facet baseFacet = base.getFacet(Facet.LENGTH);
        if (baseFacet != null) {
            if (requiredLength != ((LengthFacet)baseFacet).requiredLength) {
                throw new SchemaException("In a type derived by restriction, " +
                                          "the length facet must not be changed");
            }
        }
        baseFacet = base.getFacet(Facet.MIN_LENGTH);
        if (baseFacet != null) {
            if (requiredLength < baseFacet.toLong()) {
                throw new SchemaException("In a type derived by restriction, " +
                                          "the length facet is shorter than the minLength of the base type");
            }
        }
        baseFacet = base.getFacet(Facet.MAX_LENGTH);
        if (baseFacet != null) {
            if (requiredLength > baseFacet.toLong()) {
                throw new SchemaException("In a type derived by restriction, " +
                                          "the length facet is longer than the maxLength of the base type");
            }
        }
    }


    /**
     * Test whether an atomic value conforms to this facet
     * @param value  the value to be tested
     * @return true if the value conforms; false if it doesn't conform, or if the test
     * fails
     */

    public boolean testAtomicValue(AtomicValue value) {
        if (!(value instanceof StringValue)) {
            try {
                value = value.convert(Type.STRING);
                // TODO: is this correct for all data types, e.g. hexBinary?
            } catch (XPathException e) {
                return false;
            }
        }
        return testLength(((StringValue)value).getLength());
    }

    /**
     * Test whether a supplied length conforms to this required length
     */

    public boolean testLength(int actualLength) {
        return actualLength == requiredLength;
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
