package net.sf.saxon.type;

import net.sf.saxon.expr.MappingFunction;
import net.sf.saxon.expr.MappingIterator;
import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NamespaceResolver;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.Err;

import javax.xml.transform.TransformerException;
import java.util.Iterator;

/**
 * Represents a SimpleType that is a list of a given
 * SimpleType, known as its itemType. A ListType may be a restriction
 * of another ListType; the itemType must either be an AtomicType or
 * a UnionType whose members are not list types.
 *
**/

public class ListType extends SimpleType implements MappingFunction {

    /**
     *  The SimpleType of the items in the list.
     */

    private SimpleType itemType = null;

    /**
     * Create a new ListType.
     */

    public ListType(NamePool pool) {
        super();
        super.setNamePool(pool);
    }

   /**
     * Returns true if this type is derived by list, or if it is derived by restriction
     * from a list type, or if it is a union that contains a list as one of its members
     */

    public boolean isListType() {
        return true;
    }

    /**
     * Returns the simpleType of the items in this ListType.
     * @return the simpleType of the items in this ListType.
    */

    public SimpleType getItemType() {
        return itemType;
    }

    /**
     * Sets the itemType for this ListType (the type of
     * item that instances of this list type contain).
     * @param type the SimpleType for this ListType.
    */

    public void setItemType(SimpleType type) throws SchemaException {
        itemType = type;
    }

    /**
     * Check the validity of this SimpleType defintion.
     * @throws TransformerException when this SimpleType definition
     * is invalid.
     */

//    public boolean validate(Schema schema)
//            throws TransformerException {
//
//        boolean result = true;
//        // check for circularities and for previous validation
//        switch (validationPhase) {
//            case VALIDATED:
//            case INVALID:
//                return true;
//            case VALIDATING:
//                schema.error("The definition of list type " +
//                                getLocalName() + " is circular", this);
//                return false;
//            case UNVALIDATED:
//                break;
//        }
//        validationPhase = VALIDATING;
//
//        if (!super.validate(schema)) {
//            return false;
//        };
//
//        // If the type is derived by restriction, check that the base type is itself
//        // a list, or a union containing a list as a member
//
//        if (getBaseType().getFingerprint() != StandardNames.XS_ANY_SIMPLE_TYPE &&
//                !((SimpleType)getBaseType() instanceof ListType)) {
//            schema.error("The list type " + getDescription() +
//                                " has a base type that is not a list", this);
//
//        }
//
//        SimpleType itemType = getItemType();
//        if (itemType == null) {
//            itemType = ((ListType)getBaseType()).getItemType();
//        } else {
//            result &= itemType.validate(schema);
//        }
//
//        // Check that the item type is not itself a list, or a union containing
//        // a list as a member
//
//        if (itemType.isListType()) {
//            schema.error("The list type " + getDescription() +
//                                " has another list type as its item type", this);
//
//        }
//        validationPhase = VALIDATED;
//        return result;
//    }

    /**
     * Check whether a given input string is valid according to this SimpleType
     */

    public void validateContent(CharSequence value, NamespaceResolver namespaceResolver) throws ValidationException {
        SimpleType base = getItemType();
        SequenceIterator iter = new StringTokenIterator(value.toString());
        int count = 0;
        try {
            while (true) {
                StringValue val = (StringValue)iter.next();
                if (val == null) break;
                base.validateContent(val.getStringValue(), namespaceResolver);
                count++;
            }
            Iterator fi = getFacets();
            while (fi.hasNext()) {
                Facet f = (Facet)fi.next();
                if (f instanceof LengthFacet) {
                    if (!((LengthFacet)f).testLength(count)) {
                        throw new ValidationException(
                                "Length of list (" + count + ") violates " +
                                f.getName() + " facet " + Err.wrap(f.getValue()));

                    }
                }
                // TODO: other facets are allowed, eg pattern?
            }
        } catch (XPathException err) {
            throw new ValidationException(err);
        }
    }

    /**
     * Get the typed value of a given input string. This method assumes that the input value
     * is valid according to this SimpleType
     */

    public SequenceIterator getTypedValue(CharSequence value) throws ValidationException {
        SequenceIterator iter = new StringTokenIterator(value.toString());
        try {
            return new MappingIterator(iter, this, null, getItemType());
        } catch (XPathException e) {
            throw new ValidationException(e);
        }
    }

    /**
    * Implement the mapping function
    */

    public Object map(Item item, XPathContext context, Object info) throws XPathException {
        return ((AtomicValue)item).convert((AtomicType)info);
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay, Saxonica.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//


