/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio, Inc.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio, Inc. Exolab is a registered
 *    trademark of Intalio, Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * INTALIO, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999-2000 (C) Intalio Inc. All Rights Reserved.
 *
 * $Id: SchemaType.java,v 1.1 2004/03/09 10:10:35 mhkay Exp $
 */

package net.sf.saxon.type;

import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NamespaceConstant;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.SourceLocator;
import java.io.Serializable;

/**
 * The base type for XML Schema types, such as complex types
 * and simple types.
 * <BR />
 * @author <a href="mailto:kvisco@intalio.com">Keith Visco</a>
 * @version $Revision: 1.1 $ $Date: 2004/03/09 10:10:35 $
**/
public abstract class SchemaType implements Serializable, SourceLocator
        //extends Structure
{
    /**
     * Flags used to implement the "final" and "block" attributes
     */
    protected int finalProhibitions = 0;
    public static final int DERIVE_BY_RESTRICTION = 1;
    public static final int DERIVE_BY_EXTENSION = 2;
    public static final int DERIVE_BY_UNION = 4;
    public static final int DERIVE_BY_LIST = 8;
    public static final int DERIVE_BY_SUBSTITUTION = 16;
    public static final int ALL_DERIVATIONS = 31;

    /**
     * The local name of this type
     */
    private String localName = null;

    /**
     * The Saxon fingerprint of the name of the base type. Used for
     * resolving forwards references.
     */

    private int baseTypeFingerprint = -1;

   /**
     * The base datatype reference
    **/
    private SchemaType baseType = null;

    /**
     * The fingerprint of the element or attribute declaration containing
     * this anonymous type definition
     */

    private int containingDeclarationName = -1;

    /**
     * Flag to indicate whether the containing declaration of an anonymous type
     * is an element or attribute
     */

    private boolean containingDeclarationIsElement = true;

    /**
     * The name pool containing this type
    **/
    private NamePool namePool = null;

    /**
     * The derivation method (if any), for example DERIVED_BY_RESTRICTION
     */
    private int derivation = 0;

    /**
     * The fingerprint of this type in the Saxon NamePool
     */

    private int fingerprint = -1;

    /**
     * Loctation information
     */

    private String systemId = null;
    private int lineNumber = -1;

    /**
     * Flag used to check for cycles during validation
     */

    protected int validationPhase = 0;
    protected final static int UNVALIDATED = 0;
    protected final static int VALIDATING = 1;
    protected final static int VALIDATED = 2;
    protected final static int INVALID = 3;

    /**
     * Default constructor
    **/
    public SchemaType() {}

    /**
     * Get the fingerprint of the name of this type
     * @return the fingerprint. Return -1 for an anonymous type.
     */

    public int getFingerprint() {
        return fingerprint;
    }

    /**
     * Set the fingerprint of the name of this type
     * @param fingerprint the fingerprint allocated in the namepool for this schema
     */

    public void setFingerprint(int fingerprint) {
        this.fingerprint = fingerprint;
    }

    /**
     * Get the fingerprint of the name of the base type of this type
     * @return the fingerprint. Return -1 for an anonymous type.
     */
    public int getBaseTypeFingerprint() {
        return baseTypeFingerprint;
    }

    /**
     * Set the fingerprint of the name of the base type of this type
     * @param fingerprint the fingerprint allocated in the namepool for this schema
     */

    public void setBaseTypeFingerprint(int fingerprint) {
        this.baseTypeFingerprint = fingerprint;
    }

    /**
     * Returns the local name of this type (null if none was defined)
     */
    public String getLocalName() {
        return localName;
    }

    /**
     * Sets the local name of this type
     * @param localName local name of the type
    **/
    public void setLocalName(String localName) {
        // MHK: we do now allow the name to be changed, as part of <redefine>, and
        // it is the caller's responsibility to update all indexes.
        this.localName = localName;
    }

    public String getDisplayName() {
        return namePool.getDisplayName(fingerprint);
    }

    /**
     * Returns true if this SchemaType is a ComplexType
     * @return true if this SchemaType is a ComplexType
    **/
    public final boolean isComplexType() {
        return !isSimpleType();
    }

    /**
     * Returns true if this SchemaType is a SimpleType
     * @return true if this SchemaType is a SimpleType
    **/
    public abstract boolean isSimpleType();

	/**
	 * Returns the value of the 'block' attribute for this type, as a bit-signnificant
     * integer with fields such as DERIVE_BY_LIST and DERIVE_BY_EXTENSION
     * @return the value of the 'block' attribute for this type
	 */

	public int getBlock() {
		return 0;
	}

    /**
     * Get the NamePool in which this type is defined
     */

    public NamePool getNamePool() {
        return namePool;
    }

    /**
     * Set the NamePool in which the type is defined
     * @param pool the NamePool
     */

    public void setNamePool(NamePool pool) {
        namePool = pool;
    }

    /**
     * Returns the base type that this type inherits from.
     * If this type is a Simpletype that is a built in primitive type then null is returned.
     * @return the parent type.
    **/

    public SchemaType getBaseType() {
        return baseType;
    }

    /**
     * Sets the base type for this datatype
     * @param baseType the base type which this datatype inherits from
    **/
    public void setBaseType(SchemaType baseType) throws SchemaException {
        this.baseType = baseType;
    }

    /**
     * Gets the code of the derivation method used to derive this type from its
     * parent. Return zero for primitive types.
     * @return a code, for example DERIVE_BY_RESTRICTION
     */
    public int getDerivationMethod() {
        return derivation;
    }

    /**
     * Determines whether derivation (of a particular kind)
     * from this type is allowed, based on the "final" property
     * @param derivation the kind of derivation, e.g. DERIVE_BY_LIST
     * @return true if this kind of derivation is allowed
    **/
    public boolean allowsDerivation(int derivation) {
        return (finalProhibitions & derivation) == 0;
    }

    /**
     * Sets the derivation method name
     */
    public void setDerivationMethodName(String method) throws SchemaException {
        if (method.equals("restriction")) {
            derivation = DERIVE_BY_RESTRICTION;
        } else if (method.equals("union")) {
            derivation = DERIVE_BY_UNION;
        } else if (method.equals("list")) {
            derivation = DERIVE_BY_LIST;
        } else if (method.equals("extension")) {
            derivation = DERIVE_BY_EXTENSION;
        } else {
            throw new SchemaException("Unknown derivation method: " + method);
        }
    }

    /**
     * Sets the derivation method code (without validating it)
     */
    public void setDerivationMethod(int method) {
        derivation = method;
    }

    /**
     * Sets the value of the 'final' property, indicating which
     * types of derivation are not allowed.
     *
     * @param value the value of the final property, as a bit-significant integer
     * @exception net.sf.saxon.type.SchemaException when the value is not a valid value.
    **/

    public void setFinal(int value) throws SchemaException {
        finalProhibitions = value;
    }

    /**
     * Sets the value of the 'final' property, indicating which
     * types of derivation are not allowed.
     *
     * @param finalValue the bit-significant code representing the final property.
    **/

    public void setFinalProhibitions(int finalValue) {
        finalProhibitions = finalValue;
    }

    /**
     * Check that this type is validly derived from a given type
     */

    public void checkDerivation(SchemaType type, int block) throws SchemaException {
        int derivations = 0;
        SchemaType t = this;
        while (true) {
            if (t == null) {
                if (type instanceof AnyType) {
                    return;
                } else {
                    throw new SchemaException(
                        "The requested type " +
                        getDescription() +
                        " is not derived from the declared type " +
                        type.getDescription());
                }
            }
            if ((derivations & t.getBlock()) != 0) {
                throw new SchemaException(
                        "Derivation of the requested type " +
                        getDescription() +
                        " is blocked by the base type " +
                        t.getDescription());
            }
            if (t == type) {
                break;
            }
            derivations |= t.getDerivationMethod();
            t = t.getBaseType();
        }
        if ((derivations & block) != 0) {
            throw new SchemaException(
                    "Derivation of the requested type " +
                    getDescription() +
                    " is blocked by the element declaration");
        }
    }


    /**
     * Get the typed value of a node that is annotated with this schema type
     */

    public abstract SequenceIterator getTypedValue(NodeInfo node)
            throws XPathException;


    /**
     * Set the name of the containing declaration (for diagnostics)
     */

    public void setContainingDeclaration(int fingerprint, boolean isElement) {
        containingDeclarationName = fingerprint;
        containingDeclarationIsElement = isElement;
    }

    /**
     * Set location information for diagnostics
     */

    public void setLocator(SourceLocator locator) {
        systemId = locator.getSystemId();
        lineNumber = locator.getLineNumber();
    }

    public void setSystemId(String systemId) {
        this.systemId = systemId;
    }

    public void setLineNumber(int lineNumber) {
        this.lineNumber = lineNumber;
    }

    /**
     * Get the system ID of the location where the type was defined
     */

    public String getSystemId() {
        return systemId;
    }

    /**
     * Get the line number of the location where the type was defined
     */

    public int getLineNumber() {
        return lineNumber;
    }

    /**
     * Get the public ID of the location where the type was defined.
     * This is always null.
     */

    public String getPublicId() {
        return null;
    }

    /**
     * Get the column number of the location where the type was defined.
     * This is always -1, meaning unknown.
     */

    public int getColumnNumber() {
        return -1;
    }

    /**
     * Get a description of this type for use in diagnostics
     */

    public String getDescription() {
        int fp = getFingerprint();
        if (fp != -1 && getNamePool().getURI(fp) != NamespaceConstant.ANONYMOUS) {
            return getNamePool().getDisplayName(fp);
        }
        if (containingDeclarationName != -1) {
            return "of " +
                    (containingDeclarationIsElement ? "element " : "attribute ") +
                    getNamePool().getDisplayName(containingDeclarationName);
        }
        return "defined at line " + getLineNumber() + " of " +
                getSystemId();
    }

    public int getContainingDeclarationName() {
        return containingDeclarationName;
    }

    public boolean containingDeclarationIsElement() {
        return containingDeclarationIsElement;
    }
    /**
     * Determine whether this type is a valid restriction of another type
     */

    public boolean isValidRestriction(SchemaType base, SchemaMarker schema)
    throws SchemaException {
        // TODO: implement this
        return true;
    }

}
