/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Intalio, Inc.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Intalio, Inc. Exolab is a registered
 *    trademark of Intalio, Inc.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY INTALIO, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * INTALIO, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999-2000 (C) Intalio, Inc. All Rights Reserved.
 *
 * $Id: SimpleType.java,v 1.1 2004/03/09 10:10:35 mhkay Exp $
 *
 * Almost totally rewritten 2004 by Michael Kay for Saxon
 */

package net.sf.saxon.type;

import net.sf.saxon.type.ValidationException;
import net.sf.saxon.type.SchemaType;
import net.sf.saxon.type.Facet;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.NamespaceResolver;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.style.StandardNames;

import javax.xml.transform.TransformerException;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import java.util.Collections;

/**
 * An XML Schema SimpleType
 * @author <a href="mailto:kvisco@intalio.com">Keith Visco</a>
 * @version $Revision: 1.1 $ $Date: 2004/03/09 10:10:35 $
**/

public abstract class SimpleType extends SchemaType
{

    /**
     * The value of the final attribute used for
     * blocking all types of derivation
    **/
    public static final String FINAL_ALL         = "#all";

    /**
     * The value of the final attribute used for
     * blocking list derivation
    **/
    public static final String FINAL_LIST        = "list";

    /**
     * The value of the final attribute used for
     * blocking union derivation
    **/
    public static final String FINAL_UNION       = "union";

    /**
     * The value of the final attribute used for
     * blocking restriction derivation
    **/
    public static final String FINAL_RESTRICTION = "restriction";

    /**
     * The constraining facets of this type
    **/
    private List facets     = null;

    /**
     * Default constructor
     */
    public SimpleType() {
        super();
        this.facets  = new ArrayList();
    }

    /**
     * Returns true if this SchemaType is a SimpleType
     * @return true
    **/
    public final boolean isSimpleType() {
        return true;
    }

    /**
     * Returns true if this type is derived by list, or if it is derived by restriction
     * from a list type, or if it is a union that contains a list as one of its members
     */

    public boolean isListType() {
        return false;
    }

    /**
     * Set all the facets of this simple type
     */

    public void setFacets(List facets) {
        this.facets = facets;
    }

    /**
     * Adds the given Facet to this Simpletype.
     * @param facet the Facet to add to this Simpletype
     */

    public void addFacet(Facet facet) {
        facets.add(facet);
    }

    /**
     * Returns the first facet associated with the given name
     * @return the first facet associated with the given name
    **/
    public Facet getFacet(String name) {
        Iterator facets= getFacets(name);

        if (facets == null) return null;

        return (Facet)facets.next();
    } //-- getFacet

    /**
     * Returns the facets associated with the given name
     * @return the facets associated with the given name
    **/

    public List getFacetList() {
        return facets;
    }

    /**
     * Returns the facets associated with the given name
     * @return the facets associated with the given name
    **/
    public Iterator getFacets(String name) {
        FacetListIterator fle = null;
        SimpleType datatype = (SimpleType)getBaseType();
        if (datatype != null && getDerivationMethod() != DERIVE_BY_LIST) {
            fle = (FacetListIterator)datatype.getFacets(name);
        }
        fle = new FacetListIterator(facets, fle);
        fle.setMask(name);
        return fle;
    } //-- getFacets

    /**
     * Returns an Enumeration of all the Facets (including inherited)
     * facets for this type.
     * @return an Enumeration of all the Facets for this type
    **/
    public FacetListIterator getFacets() {
        FacetListIterator fle = null;
        if (getDerivationMethod() == DERIVE_BY_LIST) {
            return new FacetListIterator(facets, null);
        }
        SimpleType datatype = (SimpleType)getBaseType();
        if (datatype != null && datatype != this) {
            fle = datatype.getFacets();
        }
        return new FacetListIterator(facets, fle);
    }


    /**
     * Returns the facets of this type (without the parent's facets)
     */
    public Iterator getLocalFacets()    {
        if (facets == null) {
            return Collections.EMPTY_LIST.iterator();
        }
        return facets.iterator();
    }

    /**
     * Returns the built in type this type is derived from.
     */
    public SimpleType getBuiltInBaseType() {
        SimpleType base = this;
        while ((base != null) && ( base.getFingerprint() > 1023 )) {
            base = (SimpleType)base.getBaseType();
        }
        return base;
    }

    /**
     * Test whether this simple type is namespace-sensitive, that is, whether
     * it is derived from xs:QName or xs:NOTATION
     */

    public boolean isNamespaceSensitive() {
        SimpleType base = this;
        int fp = base.getFingerprint();
        while ((base != null) && ( fp > 1023 )) {
            base = (SimpleType)base.getBaseType();
            fp = base.getFingerprint();
            if (fp == StandardNames.XS_QNAME || fp == StandardNames.XS_NOTATION) {
                return true;
            }
        }
        return false;
    }

    /**
     * Checks the validity of this SimpleType defintion.
     *
     * @throws javax.xml.transform.TransformerException when this SimpleType definition
     * is invalid.
    **/
    public boolean validate(SchemaMarker schema)
            throws TransformerException {

        // Check that the base type is a simple type

        if (getFingerprint() < 1023) {
            // built-in type
            return true;
        }

        if (getBaseType() == null) {
            setBaseType(BuiltInSchemaFactory.getSchemaType(StandardNames.XS_ANY_SIMPLE_TYPE));
        } else if (getBaseType() instanceof SimpleType) {
            ((SimpleType)getBaseType()).validate(schema);
        } else if (!getBaseType().isSimpleType()) {
            schema.error("The base type of the simple type " +
                                        getDescription() +
                                        " is not a simple type", this);
            return false;
        }
        return true;
    }

    /**
     * Check whether a given input string is valid according to this SimpleType
     */

    public abstract void validateContent(CharSequence value, NamespaceResolver namespaceResolver)
            throws ValidationException;

    /**
     * Get the typed value of a node that is annotated with this schema type
     */

    public final SequenceIterator getTypedValue(NodeInfo node)
    throws XPathException {
        try {
            return getTypedValue(node.getStringValue());
        } catch (ValidationException err) {
            throw new XPathException.Dynamic("Value doesn't match its type annotation. " + err.getMessage());
        }
    }


    /**
     * Get the typed value corresponding to a given string value, assuming it is
     * valid against this type
     * @param value the string value
     * @return an iterator over the atomic sequence comprising the typed value
     */

    public abstract SequenceIterator getTypedValue(CharSequence value)
            throws ValidationException;


}
