package net.sf.saxon.type;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.pattern.AnyNodeTest;
import net.sf.saxon.pattern.NoNodeTest;
import net.sf.saxon.pattern.NodeTest;
import net.sf.saxon.style.StandardNames;
import net.sf.saxon.value.AtomicValue;

import java.io.Serializable;

/**
 * This class contains static information about types and methods for constructing type codes.
 * The class is never instantiated.
 *
 */

public abstract class Type implements Serializable {

    // Note that the integer codes representing node item types are the same as
    // the codes allocated in the DOM interface, while the codes for built-in
    // atomic type are fingerprints allocated in StandardNames. These two sets of
    // codes must not overlap!

    // The constants that are defined as synonyms of constants in StandardNames
    // are purely for convenience (or for legacy), they will eventually be phased out.

    /**
     * Type representing an element node - element()
     */

    public static final short ELEMENT = 1;
    /**
     * Item type representing an attribute node - attribute()
     */
    public static final short ATTRIBUTE = 2;
    /**
     * Item type representing a text node - text()
     */
    public static final short TEXT = 3;
    /**
     * Item type representing a processing-instruction node
     */
    public static final short PROCESSING_INSTRUCTION = 7;
    /**
     * Item type representing a comment node
     */
    public static final short COMMENT = 8;
    /**
     * Item type representing a document node
     */
    public static final short DOCUMENT = 9;
    /**
     * Item type representing a namespace node
     */
    public static final short NAMESPACE = 13;

    /**
     * An item type that matches any node
     */

    public static final short NODE = 0;

    public static final ItemType NODE_TYPE = AnyNodeTest.getInstance();

    /**
     * An item type that matches any item
     */

    public static final short ITEM = 88;

    public static final ItemType ITEM_TYPE = AnyItemType.getInstance();



    /**
     * A type that matches nothing
     */

    public static final short MAX_NODE_TYPE = 13;
    /**
     * Item type that matches no items (corresponds to SequenceType empty())
     */
    public static final short EMPTY = 15;    // a test for this type will never be satisfied

    /**
     * Test whether a given type is (some subtype of) node()
     *
     * @param type The type to be tested
     * @return true if the item type is node() or a subtype of node()
     */

    public static boolean isNodeType(ItemType type) {
        return type instanceof NodeTest;
    }

    /**
     * Constant denoting any atomic type (the union of all primitive types and types
     * derived from primitive types by restriction or by union)
     */

    //public final static int ATOMIC          = 90;
    public final static int ATOMIC          = StandardNames.XDT_ANY_ATOMIC_TYPE;

    /**
     * Constant denoting any numeric type (the union of float, double, and decimal)
     */

    //public final static int NUMBER          = 91;
    public final static int NUMBER          = StandardNames.XDT_NUMERIC;

    /**
     * Constants representing primitive data types defined in Schema Part 2
     */

    public final static int STRING = StandardNames.XS_STRING;
    /**
     * Item type representing the type xs:boolean
     */
    public final static int BOOLEAN = StandardNames.XS_BOOLEAN;
    /**
     * Item type representing the type xs:decimal
     */
    public final static int DECIMAL = StandardNames.XS_DECIMAL;
    /**
     * Item type representing the type xs:float
     */
    public final static int FLOAT = StandardNames.XS_FLOAT;
    /**
     * Item type representing the type xs:double
     */
    public final static int DOUBLE = StandardNames.XS_DOUBLE;
    /**
     * Item type representing the type xs:duration
     */
    public final static int DURATION = StandardNames.XS_DURATION;
    /**
     * Item type representing the type xs:dateTime
     */
    public final static int DATE_TIME = StandardNames.XS_DATE_TIME;
    /**
     * Item type representing the type xs:time
     */
    public final static int TIME = StandardNames.XS_TIME;
    /**
     * Item type representing the type xs:date
     */
    public final static int DATE = StandardNames.XS_DATE;
    /**
     * Item type representing the type xs:gYearMonth
     */
    public final static int G_YEAR_MONTH = StandardNames.XS_G_YEAR_MONTH;
    /**
     * Item type representing the type xs:gYear
     */
    public final static int G_YEAR = StandardNames.XS_G_YEAR;
    /**
     * Item type representing the type xs:monthDay
     */
    public final static int G_MONTH_DAY = StandardNames.XS_G_MONTH_DAY;
    /**
     * Item type representing the type xs:gDay
     */
    public final static int G_DAY = StandardNames.XS_G_DAY;
    /**
     * Item type representing the type xs:gMonth
     */
    public final static int G_MONTH = StandardNames.XS_G_MONTH;
    /**
     * Item type representing the type xs:hexBinary
     */
    public final static int HEX_BINARY = StandardNames.XS_HEX_BINARY;
    /**
     * Item type representing the type xs:base64Binary
     */
    public final static int BASE64_BINARY   = StandardNames.XS_BASE64_BINARY;
    /**
     * Item type representing the type xs:anyURI
     */
    public final static int ANY_URI = StandardNames.XS_ANY_URI;
    /**
     * Item type representing the type xs:QName
     */
    public final static int QNAME = StandardNames.XS_QNAME;
    /**
     * Item type representing the type xs:NOTATION
     *
     */
    public final static int NOTATION = StandardNames.XS_NOTATION;

    /**
     * Item type representing the type xdt:untypedAtomic
     * (the type of the content of a schema-less node)
     */

    public final static int UNTYPED_ATOMIC = StandardNames.XDT_UNTYPED_ATOMIC;

    public final static int ANY_SIMPLE_TYPE = StandardNames.XS_ANY_SIMPLE_TYPE;

    /**
     * Constant representing the type of an external object (for use by extension functions)
     */

    public final static int OBJECT = StandardNames.SAXON_JAVA_LANG_OBJECT;

    /**
     * Item type representing the type xs:integer
     */

    public final static int INTEGER = StandardNames.XS_INTEGER;
    /**
     * Item type representing the type xs:nonPositiveInteger
     */
    public final static int NON_POSITIVE_INTEGER = StandardNames.XS_NON_POSITIVE_INTEGER;
    /**
     * Item type representing the type xs:negativeInteger
     */
    public final static int NEGATIVE_INTEGER = StandardNames.XS_NEGATIVE_INTEGER;
    /**
     * Item type representing the type xs:long
     */
    public final static int LONG = StandardNames.XS_LONG;
    /**
     * Item type representing the type xs:int
     */
    public final static int INT = StandardNames.XS_INT;
    /**
     * Item type representing the type xs:short
     */
    public final static int SHORT = StandardNames.XS_SHORT;
    /**
     * Item type representing the type xs:byte
     */
    public final static int BYTE = StandardNames.XS_BYTE;
    /**
     * Item type representing the type xs:nonNegativeInteger
     */
    public final static int NON_NEGATIVE_INTEGER = StandardNames.XS_NON_NEGATIVE_INTEGER;
    /**
     * Item type representing the type xs:positiveInteger
     */
    public final static int POSITIVE_INTEGER = StandardNames.XS_POSITIVE_INTEGER;
    /**
     * Item type representing the type xs:unsignedLong
     */
    public final static int UNSIGNED_LONG = StandardNames.XS_UNSIGNED_LONG;
    /**
     * Item type representing the type xs:unsignedInt
     */
    public final static int UNSIGNED_INT = StandardNames.XS_UNSIGNED_INT;
    /**
     * Item type representing the type xs:unsignedShort
     */
    public final static int UNSIGNED_SHORT = StandardNames.XS_UNSIGNED_SHORT;
    /**
     * Item type representing the type xs:unsignedByte
     */
    public final static int UNSIGNED_BYTE = StandardNames.XS_UNSIGNED_BYTE;

    /**
     * Item type representing the type xs:normalizedString
     */
    public final static int NORMALIZED_STRING = StandardNames.XS_NORMALIZED_STRING;
    /**
     * Item type representing the type xs:token
     */
    public final static int TOKEN = StandardNames.XS_TOKEN;
    /**
     * Item type representing the type xs:language
     */
    public final static int LANGUAGE = StandardNames.XS_LANGUAGE;
    /**
     * Item type representing the type xs:NMTOKEN
     */
    public final static int NMTOKEN = StandardNames.XS_NMTOKEN;
    /**
     * Content type representing the complex type xs:NMTOKENS
     */
    public final static int NMTOKENS = StandardNames.XS_NMTOKENS;      // NB: list type
    /**
     * Item type representing the type xs:NAME
     */
    public final static int NAME = StandardNames.XS_NAME;
    /**
     * Item type representing the type xs:NCNAME
     */
    public final static int NCNAME = StandardNames.XS_NCNAME;
    /**
     * Item type representing the type xs:ID
     */
    public final static int ID = StandardNames.XS_ID;
    /**
     * Item type representing the type xs:IDREF
     */
    public final static int IDREF = StandardNames.XS_IDREF;
    /**
     * Content type representing the complex type xs:IDREFS
     */
    public final static int IDREFS = StandardNames.XS_IDREFS;      // NB: list type
    /**
     *
     * Item type representing the type xs:ENTITY
     */
    public final static int ENTITY = StandardNames.XS_ENTITY;

    /**
     *
     * Item type representing the type xdt:yearMonthDuration
     */
    public final static int YEAR_MONTH_DURATION = StandardNames.XDT_YEAR_MONTH_DURATION;
    /**
     *
     * Item type representing the type xdt:dayTimeDuration
     */
    public final static int DAY_TIME_DURATION = StandardNames.XDT_DAY_TIME_DURATION;

    public final static AtomicType UNTYPED_ATOMIC_TYPE =  (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XDT_UNTYPED_ATOMIC);

    public final static AtomicType ATOMIC_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XDT_ANY_ATOMIC_TYPE);

    public final static AtomicType YEAR_MONTH_DURATION_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XDT_YEAR_MONTH_DURATION);

    public final static AtomicType DAY_TIME_DURATION_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XDT_DAY_TIME_DURATION);

    public final static AtomicType STRING_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_STRING);

    public final static AtomicType BOOLEAN_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_BOOLEAN);

    public final static AtomicType DECIMAL_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_DECIMAL);

    public final static AtomicType FLOAT_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_FLOAT);

    public final static AtomicType DOUBLE_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_DOUBLE);

    public final static AtomicType DURATION_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_DURATION);

    public final static AtomicType DATE_TIME_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_DATE_TIME);

    public final static AtomicType TIME_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_TIME);

    public final static AtomicType DATE_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_DATE);

    public final static AtomicType G_YEAR_MONTH_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_G_YEAR_MONTH);

    public final static AtomicType G_YEAR_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_G_YEAR);

    public final static AtomicType G_MONTH_DAY_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_G_MONTH_DAY);

    public final static AtomicType G_DAY_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_G_DAY);

    public final static AtomicType G_MONTH_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_G_MONTH);

    public final static AtomicType HEX_BINARY_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_HEX_BINARY);

    public final static AtomicType BASE64_BINARY_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_BASE64_BINARY);

    public final static AtomicType ANY_URI_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_ANY_URI);

    public final static AtomicType QNAME_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_QNAME);

    public final static AtomicType NOTATION_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_NOTATION);

    public final static AtomicType INTEGER_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_INTEGER);

    public final static AtomicType ID_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XS_ID);

    public final static AtomicType NUMBER_TYPE = (AtomicType)
        BuiltInSchemaFactory.getSchemaType(StandardNames.XDT_NUMERIC);
            // TODO: actually, the number type is anonymous?


    //public final static AtomicType OBJECT_TYPE = ExternalObjectType.getInstance();


    /**
     * Output (for diagnostics) a representation of the type of an item. This
     * does not have to be the most specific type
     */

    public final static String displayTypeName(Item item) {
        if (item instanceof NodeInfo) {
            NodeInfo node = (NodeInfo)item;
            switch (node.getNodeKind()) {
                case DOCUMENT:
                    return "document-node()";
                case ELEMENT:
                    NamePool pool = node.getNamePool();
                    int annotation = node.getTypeAnnotation();
                    return "element(" +
                            ((NodeInfo)item).getDisplayName() + ", " +
                            (annotation == -1 ?
                                "xdt:untyped)" :
                                pool.getDisplayName(annotation) + ")");
                case ATTRIBUTE:
                    NamePool pool2 = node.getNamePool();
                    int annotation2 = node.getTypeAnnotation();
                    return "attribute(" +
                            ((NodeInfo)item).getDisplayName() +
                            (annotation2 == -1 ?
                                ")" :
                                "," + pool2.getDisplayName(annotation2) + ")");
                case TEXT:      return "text()";
                case COMMENT:   return "commen()";
                case PROCESSING_INSTRUCTION:
                                return "processing-instruction()";
                case NAMESPACE: return "namespace()";
                default:        return "";
            }
        } else {
            return ((AtomicValue)item).getItemType().toString();
        }
    }

    /**
     * Get the SimpleType object for a built-in simple type code
     * @return the SimpleType, or null if not found
     */

    public static ItemType getBuiltInItemType(String namespace, String localName) {
        return (ItemType)BuiltInSchemaFactory.getSchemaType(
                StandardNames.getFingerprint(namespace, localName));
    }

    /**
     * Determine whether type A is type B or one of its subtypes, recursively
     *
     * @param subtype identifies the first type
     * @param supertype identifies the second type
     * @return true if the first type is the second type or a (direct or
     *     indirect) subtype of the second type
     */

    public static boolean isSubType(ItemType subtype, ItemType supertype) {
        // System.err.println("isSubtype(" + (subtype) + ", " + (supertype) + ")");
        // System.err.println("isSubtype(" + getTypeName(subtype) + ", " + getTypeName(supertype) + ")");
        if (subtype==null) return false;
        if (subtype.isSameType(supertype)) return true;
        if (supertype instanceof AnyItemType) return true;
        if (subtype instanceof AnyItemType) return false;
        if (subtype instanceof NoNodeTest) return true;
                    // an empty sequence can satisfy any required item type
        return isSubType(subtype.getSuperType(), supertype);
        // TODO: if the supertype is an AtomicType, it contains a list of all its
        // known descendant types: use this!
    }

    /**
     * Determine whether type A is numerically promotable to type B, for example
     * xs:integer is promotable to xs:double
     */

    public static boolean isPromotable(ItemType sourceType, ItemType targetType) {
        if (isSubType(sourceType, DECIMAL_TYPE)) {
            return (targetType==FLOAT_TYPE || targetType==DOUBLE_TYPE);
        }
        if (isSubType(sourceType, FLOAT_TYPE)) {
            return (targetType==DOUBLE_TYPE);
        }
        return false;
    }

    /**
     * Get a type that is a common supertype of two given types
     *
     * @param t1 the first item type
     * @param t2 the second item type
     * @return the item type that is a supertype of both
     *     the supplied item types
     */

    public final static ItemType getCommonSuperType(ItemType t1, ItemType t2) {
        if (t1 instanceof NoNodeTest) {
            return t2;
        }
        if (t2 instanceof NoNodeTest) {
            return t1;
        }
        if (t1==t2) {
            return t1;
        } else if (Type.isSubType(t1, t2)) {
            return t2;
        } else if (Type.isSubType(t2, t1)) {
            return t1;
        } else {
            return getCommonSuperType(t2.getSuperType(), t1);
            // eventually we will hit a type that is a supertype of t2. We reverse
            // the arguments so we go up each branch of the tree alternately.
            // If we hit the root of the tree, one of the earlier conditions will be satisfied,
            // so the recursion will stop.
        }
    }

    /**
     * Determine whether this type is a primitive type. The primitive types are
     * the 19 primitive types of XML Schema, plus xs:integer; the 7 node kinds;
     * xdt:untypedAtomic; and all supertypes of these (item(), node(), xdt:anyAtomicType,
     * xdt:number, ...)
     * @param code the item type code to be tested
     * @return true if the type is considered primitive under the above rules
     */
    public static boolean isPrimitiveType(int code) {
        return code >= 0 && (code <= INTEGER || code == NUMBER ||
                code == UNTYPED_ATOMIC || code == ATOMIC ||
                code == StandardNames.XS_ANY_SIMPLE_TYPE);
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
