package net.sf.saxon.type;

import net.sf.saxon.type.AtomicType;
import net.sf.saxon.type.Facet;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.Err;

/**
 * Represents any of the facets minInclusive, maxInclusive, minExclusive, maxExclusive
 */
public class ValueRangeFacet extends Facet {

    public static final int MIN_INCLUSIVE = 0;
    public static final int MAX_INCLUSIVE = 1;
    public static final int MIN_EXCLUSIVE = 2;
    public static final int MAX_EXCLUSIVE = 3;

    int facet;
    StringValue requiredValue;

    public ValueRangeFacet(int facetType) {
        facet = facetType;
    }

    public String getName() {
        switch (facet) {
        case MIN_INCLUSIVE:
            return Facet.MIN_INCLUSIVE;
        case MAX_INCLUSIVE:
            return Facet.MIN_INCLUSIVE;
        case MIN_EXCLUSIVE:
            return Facet.MIN_EXCLUSIVE;
        case MAX_EXCLUSIVE:
            return Facet.MAX_EXCLUSIVE;
        default:
            //assert false;
            return null;
        }
    }

    public String getValue() {
        return requiredValue.getStringValue();
    }

    public void setValue(String value) throws SchemaException {
        requiredValue = new StringValue(value);
    }

    /**
      * Check that this facet is legal when used on a type derived by restriction
      * @param base the type from which the restricted type is derived
      * @throws SchemaException if the facet is not legal
      */

     public void checkFacetRestriction(SimpleType base) throws SchemaException {
        if (!(base instanceof AtomicType)) {
            throw new SchemaException("The " + getName() +
                    " facet is not available on the non-atomic type " +
                    base.getDescription());
        }

        int primitive = ((AtomicType)base).getPrimitiveType();
        boolean OK = (primitive == Type.INTEGER ||
                primitive == Type.DOUBLE ||
                primitive == Type.FLOAT ||
                primitive == Type.DECIMAL ||
                primitive == Type.DURATION ||
                primitive == Type.DATE ||
                primitive == Type.DATE_TIME ||
                primitive == Type.TIME ||
                primitive == Type.G_YEAR ||
                primitive == Type.G_YEAR_MONTH ||
                primitive == Type.G_MONTH ||
                primitive == Type.G_MONTH_DAY ||
                primitive == Type.G_DAY ||
                primitive == Type.DAY_TIME_DURATION ||
                primitive == Type.YEAR_MONTH_DURATION);
        if (!OK) {
            throw new SchemaException("The " + getName() +
                    " facet is not allowed on the unordered type " +
                    base.getDescription());
        }
        try {
            AtomicValue req = requiredValue.convert((AtomicType)base);
        } catch (XPathException err) {
            throw new SchemaException("The value " + Err.wrap(requiredValue.getStringValue()) +
                                      " of the " + getName() +
                                      " facet is not valid for the type " +
                                      base.getDescription());
        }
     }


    /**
     * Test whether an atomic value conforms to this facet
     * @param value  the value to be tested
     * @return true if the value conforms; false if it doesn't conform, or if the test
     * fails
     */

    public boolean testAtomicValue(AtomicValue value) {
        try {
            // TODO: only do this conversion once, as soon as the primitive type is known
            AtomicValue req = requiredValue.convert((AtomicType)value.getItemType());
            if (!(value instanceof Comparable)) {
                return false;
            }
            int c = ((Comparable)value).compareTo(req);
            switch (facet) {
                case MIN_INCLUSIVE:
                    return c >= 0;
                case MAX_INCLUSIVE:
                    return c <= 0;
                case MIN_EXCLUSIVE:
                    return c > 0;
                case MAX_EXCLUSIVE:
                    return c < 0;
                default:
                    //assert false;
                    return false;
            }
        } catch (XPathException e) {
            return false;
        }
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
