package net.sf.saxon.value;
import net.sf.saxon.type.AtomicType;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Facet;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.Err;

import java.util.Iterator;


/**
 * A AtomicUserDefinedValue is an atomic value of a user-defined type
 */

public class AtomicUserDefinedValue extends AtomicValue implements Item {

    // the value in the value-space of a primitive type
    AtomicValue primitiveValue;

    // the user-defined type to which this value belongs
    AtomicType typeLabel;

    private AtomicUserDefinedValue() {}

    /**
     * Create a value of a user-defined atomic type
     * @param primValue the value in the value space of the primitive type
     * @param type the required atomic type
     * @param throwError true if an exception is to be thrown when the value is
     * invalid (if false, the method returns null instead)
     * @throws XPathException if the value is invalid
     */

    public static AtomicUserDefinedValue makeValue(
            AtomicValue primValue, AtomicType type, boolean throwError)
            throws XPathException {
        AtomicUserDefinedValue val = new AtomicUserDefinedValue();
        val.primitiveValue = primValue;
        val.typeLabel = type;
        // TODO: no need to revalidate when atomizing an already-validated node
        XPathException err = val.validate();
        if (err != null) {
            if (throwError) {
                throw err;
            } else {
                return null;
            }
        }
        return val;
    }

    /**
     * Get the primitive value (the value in the value space)
     */

    public AtomicValue getPrimitiveValue() {
        return primitiveValue;
    }
    /**
     * Validate that the value conforms to all the facets for the given type
     * @return null if the value is OK, otherwise an XPathException. Note that
     * the exception is returned rather than being thrown. This is designed to
     * make the "castable" operator more efficient.
     */

    private XPathException validate() {
        Iterator iter = typeLabel.getFacets();
        boolean enumerationMatched = false;
        boolean enumerationFound = false;
        while (iter.hasNext()) {
            Facet f = (Facet)iter.next();
            boolean isEnumeration = (f.getName().equals(Facet.ENUMERATION));
            if (isEnumeration && enumerationMatched) {
                continue;
                // we've matched an enumeration value, don't check any others
                // TODO: compile the enumeration facets into a single XPath expression
            }
            boolean matches = f.testAtomicValue(primitiveValue);
            if (isEnumeration) {
                enumerationFound = true;
                if (matches) {
                    enumerationMatched = true;
                }
            } else if (!matches) {
                return new XPathException.Dynamic(
                        "Value " + displayValue() + " violates the " +
                        f.getName() + " facet " + Err.wrap(f.getValue()) + " of the type "
                        + typeLabel.getDescription());
            }
        }
        if (enumerationFound && !enumerationMatched) {
            return new XPathException.Dynamic(
                    "Value " + displayValue() + " is not one of the" +
                    " permitted values for the type " + typeLabel.getDescription());
        }
        return null;
    }

    private String displayValue() {
        String v;
        try {
            v = primitiveValue.getStringValue();
        } catch (XPathException err) {
            v = primitiveValue.toString();
        }
        if (v.length() > 30) {
            v = v.substring(0, 30) + "...";
        }
        return v;
    }

    /**
     * Convert the value to a given type. The result of the conversion will be an
     * atomic value of the required type.
     *
     * @exception XPathException if conversion is not allowed for this
     *     required type, or if the particular value cannot be converted
     * @param requiredType type code of the required atomic type
     * @return the result of the conversion, if conversion was possible. This
     *      will always be an instance of the class corresponding to the type
     *      of value requested
     */

    public AtomicValue convert(int requiredType) throws XPathException {
         return primitiveValue.convert(requiredType);
    }

    /**
     * Evaluate the value (this simply returns the value unchanged)
     *
     * @exception XPathException
     * @param context the evaluation context (not used in this implementation)
     * @return the value, unchanged
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        return this;
    }

    /**
     * Evaluate as a string
     */

    public String getStringValue() throws XPathException {
        return primitiveValue.getStringValue();
    }

    /**
     * Get the effective boolean value of the value
     *
     * @param context the evaluation context (not used in this implementation)
     * @return true, unless the value is boolean false, numeric zero, or
     *     zero-length string
     */
    public boolean effectiveBooleanValue(XPathContext context) {
        return primitiveValue.effectiveBooleanValue(context);
    }

    public int hashCode() {
        return primitiveValue.hashCode();
    }

    public Object convertToJava(Class target) throws XPathException {
        return primitiveValue.convertToJava(target);
    }

    public int conversionPreference(Class required) {
        return primitiveValue.conversionPreference(required);
    }

    public ItemType getItemType() {
        return typeLabel;
    }

    public void display(int level, NamePool pool) {
        // TODO: display the type as well
        primitiveValue.display(level, pool);
    }

    public boolean equals(Object obj) {
        return primitiveValue.equals(obj);
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//

