package net.sf.saxon.value;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.*;
import net.sf.saxon.type.AtomicType;
import net.sf.saxon.type.Facet;
import net.sf.saxon.type.LengthFacet;
import net.sf.saxon.xpath.XPathException;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;


/**
 * A AtomicValue is a value that isn't a sequence and isn't a node.
 * More strictly, it is any sequence of length one whose content is not a node.
 */

public abstract class AtomicValue extends Value implements Item {

    /**
     * Determine the static cardinality
     *
     * @return code identifying the cardinality
     * @see net.sf.saxon.value.Cardinality
     */

    public final int getCardinality() {
        return StaticProperty.EXACTLY_ONE;
    }

     /**
     * Convert the value to a given type. The result of the conversion will be an
     * atomic value of the required type. This method works only where the target
     * type is a built-in type.
     *
     * @exception XPathException if conversion is not allowed for this
     *     required type, or if the particular value cannot be converted
     * @param requiredType type code of the required atomic type
     * @return the result of the conversion, if conversion was possible. This
     *      will always be an instance of the class corresponding to the type
     *      of value requested
     */

    public abstract AtomicValue convert(int requiredType) throws XPathException;

    /**
     * Convert the value to a given type. The result of the conversion will be
     * an atomic value of the required type. This method works where the target
     * type is a built-in atomic type and also where it is a user-defined atomic
     * type.
     */

    public final AtomicValue convert(AtomicType targetType) throws XPathException {
        if (targetType.isBuiltIn()) {
            return convert(targetType.getFingerprint());
        } else {
            AtomicValue v = convert(targetType.getPrimitiveType());
            return AtomicUserDefinedValue.makeValue(v, targetType, true);
        }
    }

    /**
     * Convert the value to a namespace-sensitive type, that is, a QName or NOTATION.
     * The result of the conversion will be
     * an atomic value of the required type. This method works where the target
     * type is a built-in atomic type and also where it is a user-defined atomic
     * type.
     */

    public final AtomicValue convert(AtomicType targetType, NamespaceResolver resolver)
    throws XPathException {
        // First convert to a string and test the pattern and length facet(s)
        String s = getStringValue();
        StringValue sv = new StringValue(s);
        Iterator iter = targetType.getFacets();
        while (iter.hasNext()) {
            Facet f = (Facet)iter.next();
            if (f instanceof LengthFacet || f.getName().equals(Facet.PATTERN)) {
                f.testAtomicValue(sv);
            }
        }
        // TODO: handle enumeration facets for a QName or NOTATION
        try {
            String parts[] = Name.getQNameParts(s);
            String uri = resolver.getURIForPrefix(parts[0], true, targetType.getNamePool());
            return new QNameValue(uri, parts[2]);
        } catch (QNameException e) {
            throw new XPathException.Dynamic("Invalid QName value. " + e.getMessage());
        }

    }

    /**
     * Evaluate the value (this simply returns the value unchanged)
     *
     * @exception XPathException
     * @param context the evaluation context (not used in this implementation)
     * @return the value, unchanged
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        return this;
    }

    /**
     * Iterate over the (single) item in the sequence
     *
     * @param context the evaluation context (not used in this implementation)
     * @return a SequenceIterator that iterates over the single item in this
     *     value
     */

    public final SequenceIterator iterate(XPathContext context) {
        return SingletonIterator.makeIterator(this);
    }

    /**
     * Evaluate as a string
     */

    public final String evaluateAsString(XPathContext context) throws XPathException {
        return getStringValue();
    }
    /**
     * Get the typed value of this item
     *
     * @return the typed value of the expression (which is this value)
     */

    public final SequenceIterator getTypedValue() {
        return SingletonIterator.makeIterator(this);
    }

    /**
     * Get the primitive value (the value in the value space). This returns an
     * AtomicValue of a class that would be used to represent the primitive value.
     * In effect this means that for built-in types, it returns the value itself,
     * but for user-defined type, it returns the primitive value minus the type
     * annotation
     */

    public AtomicValue getPrimitiveValue() {
        // overridden where necessary
        return this;
    }

    /**
     * Get the effective boolean value of the value
     *
     * @param context the evaluation context (not used in this implementation)
     * @return true, unless the value is boolean false, numeric zero, or
     *     zero-length string
     */
    public boolean effectiveBooleanValue(XPathContext context) {
        return true;
        // unless otherwise specified in a subclass
    }

    /**
     * Method to extract components of a value. Implemented by some subclasses,
     * but defined at this level for convenience
     */

    public AtomicValue getComponent(int component) throws XPathException {
        throw new UnsupportedOperationException("Data type does not support component extraction");
    }

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference. This method is overridden in subclasses, but
     * it contains common functionality to supply a scalar value to a method
     * expecting a list or iterator.
    */

    public int conversionPreference(Class required) {
        if (required.isAssignableFrom(SequenceIterator.class)) return 101;
        if (required.isAssignableFrom(SequenceValue.class)) return 102;
        if (required.isAssignableFrom(List.class)) return 103;
        return Integer.MAX_VALUE;
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {
        if (target.isAssignableFrom(SequenceIterator.class)) {
            return SingletonIterator.makeIterator(this);
        }
        if (target.isAssignableFrom(SequenceValue.class)) {
            Item[] a = {this};
            return new SequenceExtent(a);
        }
        if (target.isAssignableFrom(List.class)) {
            List list = new ArrayList();
            list.add(this);
            return list;
        };
        return null;
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//

