package net.sf.saxon.value;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.style.StandardNames;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Type;
import net.sf.saxon.xpath.XPathException;

/**
 * A boolean XPath value
 */

public final class BooleanValue extends AtomicValue implements Comparable {
    private boolean value;

    /**
     * The boolean value TRUE
     */
    public final static BooleanValue TRUE = new BooleanValue(true);
    /**
     * The boolean value FALSE
     */
    public final static BooleanValue FALSE = new BooleanValue(false);

    /**
     * Private Constructor: create a boolean value
     *
     * @param value the initial value, true or false
     */

    private BooleanValue(boolean value) {
        this.value = value;
    }

    /**
     * Factory method: get a BooleanValue
     *
     * @param value true or false, to determine which boolean value is
     *     required
     * @return the BooleanValue requested
     */

    public static BooleanValue get(boolean value) {
        return (value ? TRUE : FALSE);
    }

    /**
     * Get the value
     *
     * @return true or false, the actual boolean value of this BooleanValue
     */

    public boolean getValue() {
        return value;
    }

    /**
     * Get the effective boolean value of this expression
     *
     * @param context dynamic evaluation context, not used in this
     *     implementation
     * @return the boolean value
     */
    public boolean effectiveBooleanValue(XPathContext context) {
        return value;
    }

    /**
     * Convert to target data type
     *
     * @exception XPathException if the conversion is not possible
     * @param requiredType an integer identifying the required atomic type
     * @return an AtomicValue, a value of the required type
     */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.BOOLEAN:
        case Type.ATOMIC:
        case Type.ITEM:
            return this;
        case Type.NUMBER:
        case Type.INTEGER:
            return new IntegerValue(value ? 1 : 0);
        case Type.DECIMAL:
        case Type.FLOAT:
        case Type.DOUBLE:
            return new IntegerValue(value ? 1 : 0).convert(requiredType);
        case Type.STRING:
            return new StringValue(getStringValue());
        case Type.UNTYPED_ATOMIC:
            return new UntypedAtomicValue(getStringValue());
        default:
            throw new XPathException.Dynamic("Cannot convert boolean to " +
                                    StandardNames.getDisplayName(requiredType));
        }
    }

    /**
     * Convert to string
     *
     * @return "true" or "false"
     */

    public String getStringValue() {
        return (value ? "true" : "false");
    }

    /**
     * Determine the data type of the expression
     *
     * @return Type.BOOLEAN,
     */

    public ItemType getItemType() {
        return Type.BOOLEAN_TYPE;
    }

    /**
     * Get conversion preference for this value to a Java class.
     *
     * @param required the Java class to which the value is to be converted
     * @return An integer giving the conversion preference. A low result
     *     indicates higher preference.
     */

    public int conversionPreference(Class required) {

        if (required==Object.class) return 50;

        if (required.isAssignableFrom(BooleanValue.class)) return 0;

        if (required==boolean.class) return 1;
        if (required==Boolean.class) return 2;
        if (required==String.class) return 3;
        if (required==byte.class) return 4;
        if (required==Byte.class) return 5;
        if (required==char.class) return 6;
        if (required==Character.class) return 7;
        if (required==double.class) return 8;
        if (required==Double.class) return 9;
        if (required==float.class) return 10;
        if (required==Float.class) return 11;
        if (required==int.class) return 12;
        if (required==Integer.class) return 13;
        if (required==long.class) return 14;
        if (required==Long.class) return 15;
        if (required==short.class) return 16;
        if (required==Short.class || required==CharSequence.class) return 17;

        return super.conversionPreference(required);
    }

    /**
     * Convert to Java object (for passing to external functions)
     *
     * @param target the Java class to which conversion is required
     * @exception XPathException if conversion is not possible or fails
     * @return An object of the specified Java class
     */

    public Object convertToJava(Class target) throws XPathException {
        if (target==Object.class) {
            return new Boolean(value);
        } else if (target.isAssignableFrom(BooleanValue.class)) {
            return this;
        } else if (target==boolean.class) {
            return new Boolean(value);
        } else if (target==Boolean.class) {
            return new Boolean(value);
        } else if (target==String.class || target==CharSequence.class) {
            return getStringValue();
        } else if (target==double.class) {
            return new Double((double)(value ? 1 : 0));
        } else if (target==Double.class) {
            return new Double((double)(value ? 1 : 0));
        } else if (target==float.class) {
            return new Float((float)(value ? 1 : 0));
        } else if (target==Float.class) {
            return new Float((float)(value ? 1 : 0));
        } else if (target==long.class) {
            return new Long((long)(value ? 1 : 0));
        } else if (target==Long.class) {
            return new Long((long)(value ? 1 : 0));
        } else if (target==int.class) {
            return new Integer(value ? 1 : 0);
        } else if (target==Integer.class) {
            return new Integer(value ? 1 : 0);
        } else if (target==short.class) {
            return new Short((short)(value ? 1 : 0));
        } else if (target==Short.class) {
            return new Short((short)(value ? 1 : 0));
        } else if (target==byte.class) {
            return new Byte((byte)(value ? 1 : 0));
        } else if (target==Byte.class) {
            return new Byte((byte)(value ? 1 : 0));
        } else if (target==char.class) {
            return new Character(value ? '1' : '0');
        } else if (target==Character.class) {
            return new Character(value ? '1' : '0');
        } else {
            Object o = super.convertToJava(target);
            if (o == null) {
                throw new XPathException.Dynamic("Conversion of boolean to " + target.getName() +
                        " is not supported");
            }
            return o;
        }
    }

    /**
     * Compare the value to another boolean value
     *
     * @throws ClassCastException if the other value is not a BooleanValue
     *     (the parameter is declared as Object to satisfy the Comparable
     *     interface)
     * @param other The other boolean value
     * @return -1 if this one is the lower, 0 if they are equal, +1 if this
     *     one is the higher. False is considered to be less than true.
     */

    public int compareTo(Object other) {
        if (!(other instanceof BooleanValue)) {
            throw new ClassCastException("Boolean values are not comparable to " + other.getClass());
        }
        if (this.value == ((BooleanValue)other).value) return 0;
        if (this.value) return +1;
        return -1;
    }

    /**
     * Determine whether two boolean values are equal
     *
     * @param other the value to be compared to this value
     * @return true if the other value is a boolean value and is equal to this
     *      value
     */
    public boolean equals(Object other) {
        if (!(other instanceof BooleanValue)) {
            throw new ClassCastException("Boolean values are not comparable to " + other.getClass());
        }
        return (this.value == ((BooleanValue)other).value);
    }

    /**
     * Get a hash code for comparing two BooleanValues
     *
     * @return the hash code
     */
    public int hashCode() {
        return (value ? 0 : 1);
    }

    /**
     * Diagnostic print of expression structure
     *
     * @param level indentation level of output
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "boolean (" + getStringValue() + ")" );
    }

    /**
     * Diagnostic display of this value as a string
     *
     * @return a string representation of this value: "true()" or "false()"
     */
    public String toString() {
        return getStringValue() + "()";
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//

