package net.sf.saxon.value;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.Tokenizer;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Type;
import net.sf.saxon.style.StandardNames;

import java.lang.Math;

/**
* A numeric (single precision floating point) value
*/

public final class FloatValue extends NumericValue {
    private float value;

    /**
    * Constructor supplying a float
    * @param value the value of the float
    */

    public FloatValue(float value) {
        this.value = value;
    }

    /**
    * Get the value
    */

    public float getValue() {
        return value;
    }

    public double asDouble() {
        return (double)value;
    }

    /**
     * Get the hashCode. This must conform to the rules for other NumericValue hashcodes
     * @see NumericValue#hashCode
     */

    public int hashCode() {
        if (value > Integer.MIN_VALUE && value < Integer.MAX_VALUE) {
            return (int)value;
        } else {
            return new Double(this.asDouble()).hashCode();
        }
    }

    /**
     * Test whether the value is the double/float value NaN
     */

    public boolean isNaN() {
        return Float.isNaN(value);
    }

    public boolean effectiveBooleanValue(XPathContext context) {
        return (value!=0.0 && !Float.isNaN(value));
        // NB, this returns false in the case of NaN
    }

    /**
    * Convert to target data type
    * @param requiredType an integer identifying the required atomic type
    * @return an AtomicValue, a value of the required type
    * @throws XPathException if the conversion is not possible
    */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.BOOLEAN:
            return BooleanValue.get(value!=0.0 && !Float.isNaN(value));
        case Type.FLOAT:
        case Type.NUMBER:
        case Type.ATOMIC:
        case Type.ITEM:
            return this;
        case Type.INTEGER:
            if (Float.isNaN(value)) {
                throw new XPathException.Dynamic("Cannot convert float NaN to an integer");
            }
            return new IntegerValue((long)value);
        case Type.DECIMAL:
            return new DecimalValue(value);
        case Type.DOUBLE:
            return new DoubleValue((double)value);
        case Type.STRING:
            return new StringValue(getStringValue());
        case Type.UNTYPED_ATOMIC:
            return new UntypedAtomicValue(getStringValue());
        default:
            throw new XPathException.Dynamic("Cannot convert float to " +
                                     StandardNames.getDisplayName(requiredType));
        }
    }


    /**
    * Get the value as a String
    * @return a String representation of the value
    */

    public String getStringValue() {
        // TODO: check whether the Java and XMLSchema specs give the same answer
        if (value == Float.POSITIVE_INFINITY) {
            return "INF";
        } else if (value == Float.NEGATIVE_INFINITY) {
            return "-INF";
        } else {
            return "" + value;
        }
    }

    /**
    * Determine the data type of the expression
    * @return Type.DOUBLE
    */

    public ItemType getItemType() {
        return Type.FLOAT_TYPE;
    }

    /**
    * Negate the value
    */

    public NumericValue negate() {
        return new FloatValue(-value);
    }

    /**
    * Implement the XPath floor() function
    */

    public NumericValue floor() {
        return new FloatValue((float)Math.floor(value));
    }

    /**
    * Implement the XPath ceiling() function
    */

    public NumericValue ceiling() {
        return new FloatValue((float)Math.ceil(value));
    }

    /**
    * Implement the XPath round() function
    */

    public NumericValue round() {
        if (Float.isNaN(value)) return this;
        if (Float.isInfinite(value)) return this;
        if (value==0.0) return this;    // handles the negative zero case
        if (value > -0.5 && value < 0.0) return new DoubleValue(-0.0);
        if (value > Integer.MIN_VALUE && value < Integer.MAX_VALUE) {
            return new FloatValue((float)Math.round(value));
        }

        // if the float is larger than the maximum int, then
        // it can't have any significant digits after the decimal
        // point, so return it unchanged

        return this;
    }

    /**
    * Implement the XPath round-to-half-even() function
    */

    public NumericValue roundToHalfEven(int scale) {
        try {
            return (FloatValue)new DoubleValue((double)value).roundToHalfEven(scale).convert(Type.FLOAT);
        } catch (XPathException err) {
            // can't happen
            System.err.println(err);
            return this;
        }
    }

    /**
    * Determine whether the value is a whole number, that is, whether it compares
    * equal to some integer
    */

    public boolean isWholeNumber() {
        return value == Math.floor(value);
    }

    /**
    * Evaluate a binary arithmetic operator.
    */

    public NumericValue arithmetic(int operator, NumericValue other) throws XPathException {
        if (other instanceof FloatValue) {
            switch(operator) {
                case Tokenizer.PLUS:
                    return new FloatValue(value + ((FloatValue)other).value);
                case Tokenizer.MINUS:
                    return new FloatValue(value - ((FloatValue)other).value);
                case Tokenizer.MULT:
                    return new FloatValue(value * ((FloatValue)other).value);
                case Tokenizer.DIV:
                    return new FloatValue(value / ((FloatValue)other).value);
                case Tokenizer.MOD:
                    return new FloatValue(value % ((FloatValue)other).value);
                default:
                    throw new UnsupportedOperationException("Unknown operator");
            }
        } else if (other instanceof DoubleValue) {
            return ((DoubleValue)convert(Type.DOUBLE)).arithmetic(operator, other);
        } else {
            return arithmetic(operator, (FloatValue)other.convert(Type.FLOAT));
        }
    }

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {

        if (required==Object.class) return 50;
        if (required.isAssignableFrom(FloatValue.class)) return 0;

        if (required==boolean.class) return 21;
        if (required==Boolean.class) return 22;
        if (required==byte.class) return 18;
        if (required==Byte.class) return 19;
        if (required==char.class) return 16;
        if (required==Character.class) return 17;
        if (required==double.class) return 3;
        if (required==Double.class) return 4;
        if (required==float.class) return 1;
        if (required==Float.class) return 2;
        if (required==int.class) return 12;
        if (required==Integer.class) return 13;
        if (required==long.class) return 10;
        if (required==Long.class) return 11;
        if (required==short.class) return 14;
        if (required==Short.class) return 15;
        if (required==String.class || required==CharSequence.class) return 25;

        return super.conversionPreference(required);
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {
        if (target==Object.class) {
            return new Double(value);
        } else if (target.isAssignableFrom(DoubleValue.class)) {
            return this;
        } else if (target==boolean.class || target==Boolean.class) {
            return new Boolean(value!=0.0 && !Float.isNaN(value));
        } else if (target==String.class || target==CharSequence.class) {
            return getStringValue();
        } else if (target==double.class || target==Double.class) {
            return new Double((double)value);
        } else if (target==float.class ||target==Float.class ) {
            return new Float(value);
        } else if (target==long.class || target==Long.class) {
            return new Long((long)value);
        } else if (target==int.class || target==Integer.class) {
            return new Integer((int)value);
        } else if (target==short.class || target==Short.class) {
            return new Short((short)value);
        } else if (target==byte.class || target==Byte.class) {
            return new Byte((byte)value);
        } else if (target==char.class || target==Character.class) {
            return new Character((char)value);
        } else {
            Object o = super.convertToJava(target);
            if (o == null) {
                throw new XPathException.Dynamic("Conversion of float to " + target.getName() +
                        " is not supported");
            }
            return o;
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "float (" + getStringValue() + ")" );
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (xt) are Copyright (C) (James Clark). All Rights Reserved.
//
// Contributor(s): none.
//

