package net.sf.saxon.value;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.Tokenizer;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.AtomicType;
import net.sf.saxon.type.Type;
import net.sf.saxon.type.BuiltInSchemaFactory;
import net.sf.saxon.style.StandardNames;

/**
 * An integer value: note this is a subtype of decimal in XML Schema, not a primitive type.
 * This class also supports the built-in subtypes of xs:integer.
 * Actually supports a value in the range permitted by a Java "long"
 */

// TODO: this implementation can't properly handle xs:unsignedLong

public final class IntegerValue extends NumericValue {

    /**
     * IntegerValue representing the value -1
     */
    public final static IntegerValue MINUS_ONE  = new IntegerValue(-1);
    /**
     * IntegerValue representing the value zero
     */
    public final static IntegerValue ZERO       = new IntegerValue(0);
    /**
     * IntegerValue representing the value +1
     */
    public final static IntegerValue PLUS_ONE   = new IntegerValue(+1);

    private long value;
    private ItemType type;

    /**
     * Constructor supplying a double
     *
     * @param value the value of the IntegerValue
     */

    public IntegerValue(long value) {
        this.value = value;
        this.type = Type.INTEGER_TYPE;
    }

    /**
     * Constructor for a subtype, supplying an integer
     *
     * @param val The supplied value, as an integer
     * @param type the required item type, a subtype of xs:integer
     * @exception XPathException if the supplied value is out of range for the
     *      target type
     */

    public IntegerValue(long val, ItemType type) throws XPathException  {
        this.value = val;
        this.type = type;
        checkRange(value, type);
    }

    /**
     * Check that a value is in range for the specified subtype of xs:integer
     *
     * @param value the value to be checked
     * @param type the required item type, a subtype of xs:integer
     * @exception XPathException if the value is out of range for the required
     *      subtype
     */

    private static void checkRange(long value, ItemType type) throws XPathException {
        for (int i=0; i<ranges.length; i+=3) {
            if (ranges[i] == ((AtomicType)type).getFingerprint()) {
                if (value < ranges[i+1] || value > ranges[i+2]) {
                    throw new XPathException.Dynamic("Value is out of range for type " + type.toString());
                }
                return;
            }
        }
        throw new XPathException.Dynamic("No range information for type " + type.toString());
    }

    /**
     * Static data identifying the min and max values for each built-in subtype of xs:integer
     */
    static long[] ranges = {
        Type.INTEGER                 , Long.MIN_VALUE,      Long.MAX_VALUE,
        Type.NON_POSITIVE_INTEGER    , Long.MIN_VALUE,      0,
        Type.NEGATIVE_INTEGER        , Long.MIN_VALUE,      -1,
        Type.LONG                    , Long.MIN_VALUE,      Long.MAX_VALUE,
        Type.INT                     , Integer.MIN_VALUE,   Integer.MAX_VALUE,
        Type.SHORT                   , Short.MIN_VALUE,     Short.MAX_VALUE,
        Type.BYTE                    , Byte.MIN_VALUE,      Byte.MAX_VALUE,
        Type.NON_NEGATIVE_INTEGER    , 0,                   Long.MAX_VALUE,
        Type.POSITIVE_INTEGER        , 1,                   Long.MAX_VALUE,
        Type.UNSIGNED_LONG           , 0,                   Long.MAX_VALUE,     // not long enough!
        Type.UNSIGNED_INT            , 0,                   4294967295L,
        Type.UNSIGNED_SHORT          , 0,                   65535,
        Type.UNSIGNED_BYTE           , 0,                   255 };


    /**
     * Get the hashCode. This must conform to the rules for other NumericValue hashcodes
     * @see NumericValue#hashCode
     */

    public int hashCode() {
        if (value > Integer.MIN_VALUE && value < Integer.MAX_VALUE) {
            return (int)value;
        } else {
            return new Double(this.asDouble()).hashCode();
        }
    }

    /**
     * Get the value
     *
     * @return the value of the xs:integer, as a Java long
     */

    public long getValue() {
        return value;
    }

    public long asLong() {
        return value;
    }

    /**
     * Return the effective boolean value of this integer
     *
     * @param context The dynamic evaluation context; ignored in this
     *     implementation of the method
     * @return false if the integer is zero, otherwise true
     */
    public boolean effectiveBooleanValue(XPathContext context) {
        return value != 0;
    }

    /**
     * Compare the value to another numeric value
     *
     * @param other the numeric value to be compared to this value
     * @return -1 if this value is less than the other, 0 if they are equal,
     *     +1 if this value is greater
     */

    public int compareTo(Object other) {
        if (other instanceof IntegerValue) {
            long val2 = ((IntegerValue)other).value;
            if (value == val2) return 0;
            if (value <  val2) return -1;
            return 1;
        } else {
            return super.compareTo(other);
        }
    }

    /**
     * Convert to target data type
     *
     * @exception XPathException if the conversion is not possible
     * @param requiredType an integer identifying the required atomic type
     * @return an AtomicValue, a value of the required type
     */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.BOOLEAN:
            return BooleanValue.get(value!=0);

        case Type.NUMBER:
        case Type.INTEGER:
        case Type.LONG:
        case Type.ATOMIC:
        case Type.ITEM:
            return this;

        case Type.NON_POSITIVE_INTEGER:
        case Type.NEGATIVE_INTEGER:
        case Type.INT:
        case Type.SHORT:
        case Type.BYTE:
        case Type.NON_NEGATIVE_INTEGER:
        case Type.POSITIVE_INTEGER:
        case Type.UNSIGNED_LONG:
        case Type.UNSIGNED_INT:
        case Type.UNSIGNED_SHORT:
        case Type.UNSIGNED_BYTE:
            return new IntegerValue(value,
                                    (AtomicType)BuiltInSchemaFactory.getSchemaType(requiredType));

        case Type.DOUBLE:
            return new DoubleValue((double)value);

        case Type.FLOAT:
            return new FloatValue((float)value);

        case Type.DECIMAL:
            return new DecimalValue(value);

        case Type.STRING:
            return new StringValue(getStringValue());

        case Type.UNTYPED_ATOMIC:
            return new UntypedAtomicValue(getStringValue());

        default:
            throw new XPathException.Dynamic("Cannot convert integer to " +
                                     StandardNames.getDisplayName(requiredType));
        }
    }

    /**
     * Get the value as a String
     *
     * @return a String representation of the value
     */

    public String getStringValue() {
        return value+"";
    }

    /**
     * Negate the value
     *
     * @return the result of inverting the sign of the value
     */

    public NumericValue negate() {
        return new IntegerValue(-value);
    }

    /**
     * Implement the XPath floor() function
     *
     * @return the integer value, unchanged
     */

    public NumericValue floor() {
        return this;
    }

    /**
     * Implement the XPath ceiling() function
     *
     * @return the integer value, unchanged
     */

    public NumericValue ceiling() {
        return this;
    }

    /**
     * Implement the XPath round() function
     *
     * @return the integer value, unchanged
     */

    public NumericValue round() {
        return this;
    }

    /**
     * Implement the XPath round-to-half-even() function
     *
     * @param scale number of digits required after the decimal point; the
     *     value -2 (for example) means round to a multiple of 100
     * @return if the scale is >=0, return this value unchanged. Otherwise
     *     round it to a multiple of 10**-scale
     */

    public NumericValue roundToHalfEven(int scale) {
        long absolute = Math.abs(value);
        if (scale >= 0) {
            return this;
        } else {
            long factor = 1;
            for (long i=1; i<=-scale; i++) {
                factor *= 10;
            }
            long modulus = absolute % factor;
            long rval = absolute - modulus;
            long d = modulus * 2;
            if (d > factor) {
                rval += factor;
            } else if (d<factor) {
                // no-op
            } else {
                // round to even
                if (rval % (2*factor) == 0) {
                    // no-op
                } else {
                    rval += factor;
                }
            }
            if (value<0) rval = -rval;
            return new IntegerValue(rval);
        }
    }

    /**
     * Determine whether the value is a whole number, that is, whether it compares
     * equal to some integer
     *
     * @return always true for this implementation
     */

    public boolean isWholeNumber() {
        return true;
    }

    /**
     * Evaluate a binary arithmetic operator.
     *
     * @param operator the operator to be applied, identified by a constant in
     *      the Tokenizer class
     * @param other the other operand of the arithmetic expression
     * @exception XPathException if an arithmetic failure occurs, e.g. divide
     *     by zero
     * @return the result of performing the arithmetic operation
     */

    public NumericValue arithmetic(int operator, NumericValue other) throws XPathException {
        // TODO: handle overflow
        if (other instanceof IntegerValue) {
            switch(operator) {
                case Tokenizer.PLUS:
                    return new IntegerValue(value + ((IntegerValue)other).value);
                case Tokenizer.MINUS:
                    return new IntegerValue(value - ((IntegerValue)other).value);
                case Tokenizer.MULT:
                    return new IntegerValue(value * ((IntegerValue)other).value);
                case Tokenizer.IDIV:
                    try {
                        return new IntegerValue(value / ((IntegerValue)other).value);
                    } catch (ArithmeticException err) {
                        throw new XPathException.Dynamic("Integer division failure", err);
                    }
                case Tokenizer.DIV:
                    // the result of dividing two integers is a decimal; but if
                    // one divides exactly by the other, we implement it as an integer
                    long quotient = ((IntegerValue)other).value;
                    if (quotient != 0 && value % quotient == 0) {
                        return new IntegerValue(value / quotient);
                    }
                    return new DecimalValue(value).arithmetic(Tokenizer.DIV,
                                                              new DecimalValue(quotient));
                case Tokenizer.MOD:
                    return new IntegerValue(value % ((IntegerValue)other).value);
                default:
                    throw new UnsupportedOperationException("Unknown operator");
            }
        } else {
            NumericValue v = (NumericValue)convert(other.getItemType().getPrimitiveType());
            return v.arithmetic(operator, other);
        }
    }


    /**
     * Determine the data type of the expression
     *
     * @return the actual data type
     */

    public ItemType getItemType() {
        return type;
    }


    /**
     * Get conversion preference for this value to a Java class.
     *
     * @param required the Java class to which conversion is required
     * @return the conversion preference. A low result indicates higher
     *     preference.
     */

    // Note: this table gives java Long preference over Integer, even if the
    // XML Schema type is xs:int

    public int conversionPreference(Class required) {

        if (required==Object.class) return 50;
        if (required.isAssignableFrom(IntegerValue.class)) return 0;

        if (required==boolean.class) return 14;
        if (required==Boolean.class) return 15;
        if (required==byte.class) return 12;
        if (required==Byte.class) return 13;
        if (required==char.class) return 10;
        if (required==Character.class) return 11;
        if (required==double.class) return 2;
        if (required==Double.class) return 3;
        if (required==float.class) return 4;
        if (required==Float.class) return 5;
        if (required==int.class) return 6;
        if (required==Integer.class) return 7;
        if (required==long.class) return 0;
        if (required==Long.class) return 1;
        if (required==short.class) return 8;
        if (required==Short.class) return 9;
        if (required==String.class || required==CharSequence.class) return 16;

        return super.conversionPreference(required);
    }

    /**
     * Convert to Java object (for passing to external functions)
     *
     * @param target The Java class to which conversion is required
     * @exception XPathException if conversion is not possible, or fails
     * @return the Java object that results from the conversion; always an
     *     instance of the target class
     */

    public Object convertToJava(Class target) throws XPathException {
        if (target==Object.class) {
            return new Long(value);
        } else if (target.isAssignableFrom(IntegerValue.class)) {
            return this;
        } else if (target==boolean.class) {
            BooleanValue bval = (BooleanValue)convert(Type.BOOLEAN);
            return new Boolean(bval.getValue());
        } else if (target==Boolean.class) {
            BooleanValue bval = (BooleanValue)convert(Type.BOOLEAN);
            return new Boolean(bval.getValue());
        } else if (target==String.class || target==CharSequence.class) {
            return getStringValue();
        } else if (target==double.class || target==Double.class) {
            return new Double(value);
        } else if (target==float.class || target==Float.class) {
            return new Float(value);
        } else if (target==long.class || target==Long.class) {
            return new Long(value);
        } else if (target==int.class || target==Integer.class) {
            return new Integer((int)value);
        } else if (target==short.class || target==Short.class) {
            return new Short((short)value);
        } else if (target==byte.class || target==Byte.class) {
            return new Byte((byte)value);
        } else if (target==char.class || target==Character.class) {
            return new Character((char)value);
        } else {
            Object o = super.convertToJava(target);
            if (o == null) {
                throw new XPathException.Dynamic("Conversion of integer to " + target.getName() +
                        " is not supported");
            }
            return o;
        }
    }

    /**
     * Diagnostic print of expression structure
     *
     * @param level indentation level of the output
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + type.toString() + " (" + getStringValue() + ")" );
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file except the asStringXT() and zeros() methods (not currently used).
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (xt) are Copyright (C) (James Clark). All Rights Reserved.
//
// Contributor(s): none.
//

