package net.sf.saxon.value;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.ComputedExpression;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.Item;
import net.sf.saxon.dom.DOMNodeList;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.VirtualNode;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.AnyItemType;

import java.util.List;
import java.util.ArrayList;

import org.w3c.dom.*;

/**
* A SequenceValue object represents a sequence whose members
* are all AtomicValues or Nodes.
*/


public abstract class SequenceValue extends Value {

    /**
    * Determine the data type of the items in the expression, if possible
    * @return AnyItemType (not known)
    */

    public ItemType getItemType() {
        return AnyItemType.getInstance();
    }

    /**
    * Determine the cardinality
    */

    public int getCardinality() {
        try {
            SequenceIterator iter = iterate(null);
            Item next = iter.next();
            if (next == null) {
                return StaticProperty.EMPTY;
            } else {
                if (iter.next() != null) {
                    return StaticProperty.ALLOWS_ONE_OR_MORE;
                } else {
                    return StaticProperty.EXACTLY_ONE;
                }
            }
        } catch (XPathException err) {
            // can't actually happen
            return StaticProperty.ALLOWS_ZERO_OR_MORE;
        }
    }

     /**
    * Get the n'th item in the sequence (starting from 0). This is defined for all
    * SequenceValues, but its real benefits come for a SequenceValue stored extensionally
    */

    public Item itemAt(int n) throws XPathException {
        if (n<0) return null;
        int i=0;        // indexing is zero-based
        SequenceIterator iter = iterate(null);
        while (true) {
            Item item = iter.next();
            if (item == null) {
                return null;
            }
            if (i++ == n) {
                return item;
            }
        }
    }

    /**
     * Evaluate as a singleton item (or empty sequence)
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        return iterate(context).next();
    }

    /**
     * Evaluate an expression as a String. This function must only be called in contexts
     * where it is known that the expression will return a single string (or where an empty sequence
     * is to be treated as a zero-length string). Implementations should not attempt to convert
     * the result to a string, other than converting () to "". This method is used mainly to
     * evaluate expressions produced by compiling an attribute value template.
     *
     * @exception XPathException if any dynamic error occurs evaluating the
     *     expression
     * @exception ClassCastException if the result type of the
     *     expression is not xs:string?
     * @param context The context in which the expression is to be evaluated
     * @return the value of the expression, evaluated in the current context.
     *     The expression must return a string or (); if the value of the
     *     expression is (), this method returns "".
     */

    public String evaluateAsString(XPathContext context) throws XPathException {
        AtomicValue value = (AtomicValue)evaluateItem(context);
        if (value==null) return "";
        return value.getStringValue();
    }


    /**
     * Get the effective boolean value of the expression. This returns false if the value
     * is the empty sequence, a zero-length string, a number equal to zero, or the boolean
     * false. Otherwise it returns true.
     *
     * @param context The context in which the expression is to be evaluated
     * @exception XPathException if any dynamic error occurs evaluating the
     *     expression
     * @return the effective boolean value
     */

    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
        SequenceIterator it = iterate(context);
        Item first = it.next();
        if (first == null) {
            return false;
        }
        if (first instanceof NodeInfo) {
            return true;
        } else {
            if (first instanceof BooleanValue) {
                return ((BooleanValue)first).getValue() || (it.next() != null);
            } else if (first instanceof StringValue) {
                return (!first.getStringValue().equals("")) || (it.next() != null);
            } else if (first instanceof NumericValue) {
                // first==first is a test for NaN
                return (it.next() != null) || (!(first.equals(DoubleValue.ZERO)) && first.equals(first));
            } else {
                return true;
            }
        }
    }


    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {
        if (required==Object.class) return 50;
        // the preferences for a method expecting a list are chosen to be lower
        // (more preferred) than any method expecting a scalar.
        if (required.isAssignableFrom(SequenceIterator.class)) return -10;
        if (required.isAssignableFrom(SequenceValue.class)) return -9;
        if (required.isAssignableFrom(List.class)) return -8;
        if (required.isAssignableFrom(NodeList.class)) return -7;

        // If an atomic type is required, base the decision on the type of the
        // first item in the sequence.

        try {
            SequenceIterator iter = iterate(null);
            Item item = iter.next();
            if (item == null) {
                return Integer.MAX_VALUE;
            } else if (item instanceof NodeInfo) {
                return nodeToJavaPreference((NodeInfo)item, required);
            } else {
                return ((AtomicValue)item).conversionPreference(required);
            }
        } catch (XPathException err) {}
        return Integer.MAX_VALUE;
    }

    public static int nodeToJavaPreference(NodeInfo node, Class required) {
        if (required==Object.class) return 50;
        if (NodeInfo.class.isAssignableFrom(required)) return 6;
        if (Node.class.isAssignableFrom(required)) return 7;

        try {
            SequenceIterator iter = node.getTypedValue();
            AtomicValue first = (AtomicValue)iter.next();
            if (first==null) {
                return Integer.MAX_VALUE;
            } else {
                return first.conversionPreference(required);
            }
        } catch (XPathException err) {
            return Integer.MAX_VALUE;
        }
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {

        if (target==Object.class) {
            return iterate(null);

        } else if (target.isAssignableFrom(SequenceValue.class)) {
            return this;

        } else if (target.isAssignableFrom(SequenceIterator.class)) {
            return iterate(null);

        } else if (target.isAssignableFrom(List.class)) {
            ArrayList list = new ArrayList();
            SequenceIterator iter = iterate(null);
            while (true) {
                Item it = iter.next();
                if (it == null) {
                    return list;
                }
                list.add(it);
            }
        } else if (target.isAssignableFrom(NodeList.class)) {
            if (this instanceof SequenceExtent) {
                return new DOMNodeList((SequenceExtent)this);
            } else {
                SequenceExtent ext = new SequenceExtent(iterate(null));
                return new DOMNodeList(ext);
            }
        } else {
            // try converting the first item in the sequence
            SequenceIterator iter = iterate(null);
            while (true) {
                Item next = iter.next();
                if (next == null) {
                    break;
                }
                if (next instanceof NodeInfo) {
                    if (target.isAssignableFrom(next.getClass())) {
                        // should cover both NodeInfo and DOM Node
                        return next;
                    }
                    Object n = next;
                    while (n instanceof VirtualNode) {
                        // If we've got a wrapper around a DOM or JDOM node, and the user wants a DOM
                        // or JDOM node, we unwrap it
                        Object vn = ((VirtualNode)n).getUnderlyingNode();
                        if (target.isAssignableFrom(vn.getClass())) {
                            return vn;
                        } else {
                            n = vn;
                        }
                    }
                    AtomicValue first = (AtomicValue)next.getTypedValue().next();
                    if (first==null) {
                        return null;
                    }
                    return first.convertToJava(target);
                    // TODO: we are passing the first item only. Review this.
                } else {
                    return ((AtomicValue)next).convertToJava(target);
                }
            }
            // value is an empty sequence: pass it as a Java null
            return null;
        }

    }

    /**
    * Diagnostic display of the expression
    */

    public void display(int level, NamePool pool) {
        try {
            System.err.println(ExpressionTool.indent(level) + "sequence of " +
                               getItemType().toString() + " (");
            SequenceIterator iter = iterate(null);
            while (true) {
                Item it = iter.next();
                if (it == null) {
                    break;
                }
                ((ComputedExpression)it).display(level+1, pool);
            }
            System.err.println(ExpressionTool.indent(level) + ")");
        } catch (Exception err) {}
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
