package net.sf.saxon.value;
import net.sf.saxon.expr.ComputedExpression;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.PromotionOffer;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.Tokenizer;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.Err;

import java.io.Serializable;

/**
* A value is the result of an expression but it is also an expression in its own right.
* Note that every value can be regarded as a sequence - in many cases, a sequence of
* length one.
*/

public abstract class Value implements Expression, Serializable {

    /**
     * Static method to make a Value from a given Item (which may be either an AtomicValue
     * or a NodeInfo
     * @param item      The supplied item, or null, indicating the empty sequence.
     * @return          The supplied item, if it is a value, or a SingletonNode that
     *                  wraps the item, if it is a node. If the supplied value was null,
     *                  return an EmptySequence
     */

    public static Value asValue(Item item) {
        if (item == null) {
            return EmptySequence.getInstance();
        } else if (item instanceof AtomicValue) {
            return (AtomicValue)item;
        } else {
            return new SingletonNode((NodeInfo)item);
        }
    }

    /**
     * Static method to make an Item from a Value
     * @param value the value to be converted
     * @param context the context. It is probably safe to set this to null.
     * @return null if the value is an empty sequence; or the only item in the value
     * if it is a singleton sequence
     * @throws XPathException if the Value contains multiple items
     */

    public static Item asItem(Value value, XPathContext context) throws XPathException {
        if (value instanceof EmptySequence) {
            return null;
        } else if (value instanceof SingletonNode) {
            return ((SingletonNode)value).getNode();
        } else if (value instanceof AtomicValue) {
            return (AtomicValue)value;
        } else if (value instanceof Closure) {
            return value.evaluateItem(context);
        } else {
            SequenceIterator iter = value.iterate(context);
            //if (iter.hasNext()) {
            Item item = iter.next();
            if (item == null) {
                return null;
            } else if (iter.next() != null) {
                throw new AssertionError("Attempting to access a sequence as an item");
            } else {
                return item;
            }
        }
    }

    /**
    * Static method to convert strings to numbers. Might as well go here as anywhere else.
    * @param s the String to be converted
    * @return a double representing the value of the String
     * @throws NumberFormatException if the value cannot be converted
    */

    public static double stringToNumber(CharSequence s) throws NumberFormatException {
        String n = trimWhitespace(s).toString();
        if (n.equals("INF")) {
            return Double.POSITIVE_INFINITY;
        } else if (n.equals("-INF")) {
            return Double.NEGATIVE_INFINITY;
        } else if (n.equals("NaN")) {
            return Double.NaN;
        } else {
            return new Double(n).doubleValue();
        }
    }

    /**
    * Static method to convert strings to integers.
    * @param s the String to be converted
    * @return a long representing the value of the String; if it cannot be converted,
    * throw an exception
    */

    public static long stringToInteger(CharSequence s) throws XPathException {
        try {
            CharSequence t = trimWhitespace(s);
            if (t.charAt(0)=='+') {
                t = t.subSequence(1, t.length());
            }
            return new Long(t.toString()).longValue();
        } catch (NumberFormatException err) {
            //err.printStackTrace();
            throw new XPathException.Dynamic("Cannot convert string " + Err.wrap(s) + " to an integer");
        } catch (StringIndexOutOfBoundsException err2) {
            throw new XPathException.Dynamic("Cannot convert zero-length string to an integer");
        }
    }

    /**
    * Normalize whitespace as defined in XML Schema
    */

    public static CharSequence normalizeWhitespace(CharSequence in) {
        StringBuffer sb = new StringBuffer(in.length());
        for (int i=0; i<in.length(); i++) {
            char c = in.charAt(i);
            switch (c) {
                case '\n':
                case '\r':
                case '\t':
                    sb.append(' ');
                    break;
                default:
                    sb.append(c);
                    break;
            }
        }
        return sb;
    }

    /**
    * Collapse whitespace as defined in XML Schema
    */

    public static CharSequence collapseWhitespace(CharSequence in) {
        if (in.length()==0) {
            return in;
        }

        StringBuffer sb = new StringBuffer(in.length());
        boolean inWhitespace = true;
        int i = 0;
        for (; i<in.length(); i++) {
            char c = in.charAt(i);
            switch (c) {
                case '\n':
                case '\r':
                case '\t':
                case ' ':
                    if (inWhitespace) {
                        // remove the whitespace
                    } else {
                        sb.append(' ');
                        inWhitespace = true;
                    }
                    break;
                default:
                    sb.append(c);
                    inWhitespace = false;
                    break;
            }
        }
        if (sb.charAt(sb.length()-1)==' ') {
            sb.deleteCharAt(sb.length()-1);
        }
        return sb;
    }

    /**
     * Remove leading and trailing whitespace. This has the same effect as collapseWhitespace,
     * but is cheaper, for use by data types that do not allow internal whitespace.
     * @param in the input string whose whitespace is to be removed
     * @return the result of removing excess whitespace
     */
    public static CharSequence trimWhitespace(CharSequence in) {
        if (in.length()==0) {
            return in;
        }
        int first = 0;
        int last = in.length()-1;
        while (in.charAt(first) <= 0x20) {
            if (first++ >= last) {
                return "";
            }
        }
        while (in.charAt(last) <= 0x20) {
            last--;
        }
        return in.subSequence(first, last+1);
    }

    /**
    * Simplify an expression
    * @return the simplified expression
    */

    public final Expression simplify() {
        return this;
    }

    /**
    * TypeCheck an expression
    * @return the simplified expression
    */

    public final Expression analyze(StaticContext env) {
        return this;
    }

    /**
     * Get the sub-expressions of this expression. Always returns an empty array
     */

    public final Expression[] getSubExpressions() {
        return ComputedExpression.NO_ARGUMENTS;
    }

    public int getSpecialProperties() {
        return 0;
    }

    /**
     * Offer promotion for this subexpression. Values (constant expressions)
     * are never promoted
     *
     * @param offer details of the offer, for example the offer to move
     *     expressions that don't depend on the context to an outer level in
     *     the containing expression
     * @exception net.sf.saxon.xpath.XPathException if any error is detected
     * @return if promotion is not possible, returns the expression
     *     unchanged. If promotion has occurred, returns the expression
     *     suitably modified.
     */

     public Expression promote(PromotionOffer offer) throws XPathException {
        return this;
    }

    /**
    * Determine which aspects of the context the expression depends on. The result is
    * a bitwise-or'ed value composed from constants such as StaticProperty.VARIABLES and
    * StaticProperty.CURRENT_NODE
    */

    public final int getDependencies() {
        return 0;
    }

	/**
	* Return the inverse of a relational operator, so that "a op b" can be
	* rewritten as "b inverse(op) a"
	*/

    public final static int inverse(int operator) {
        switch(operator) {
            case Tokenizer.EQUALS:
            case Tokenizer.NE:
            case Tokenizer.FEQ:
            case Tokenizer.FNE:
                return operator;
            case Tokenizer.LT:
                return Tokenizer.GT;
            case Tokenizer.LE:
                return Tokenizer.GE;
            case Tokenizer.GT:
                return Tokenizer.LT;
            case Tokenizer.GE:
                return Tokenizer.LE;
            case Tokenizer.FLT:
                return Tokenizer.FGT;
            case Tokenizer.FLE:
                return Tokenizer.FGE;
            case Tokenizer.FGT:
                return Tokenizer.FLT;
            case Tokenizer.FGE:
                return Tokenizer.FLE;
            default:
                return operator;
        }
    }

    /**
    * Convert to Java object (for passing to external functions)
    * @param target The class required by the external function
    * @return an object of the target class
    */

    public abstract Object convertToJava(Class target) throws XPathException;

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public abstract int conversionPreference(Class required);

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
